/*
 *  Copyright 2009 Lari Natri <lari.natri@gmail.com>.
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Lesser General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */


package net.aavalla.utils.general;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;

/**
 * This immutable class represents basic statistical numbers computed from
 * a list of numbers.
 * @author Lari Natri
 */
public final class BasicStatNumbers<N extends Number & Comparable<? super N>> {

    /** Count */
    public final int count;
    /** Sum */
    public final double sum;
    /** Smallest */
    public final N min;
    /** Largest */
    public final N max;
    /** Median */
    public final double median;
    /** Mean */
    public final double mean;
    /** Standard Deviation */
    public final double sd;

    /**
     * Constructs a BasicStatNumbers object and fills its fields from the
     * collection of numbers given as parameter
     * @param numberCollection numbers to calculate stats from
     * @throws IllegalArgumentException if collection is empty
     */
    public BasicStatNumbers(final Collection<N> numberCollection)
    {
        if (numberCollection.isEmpty())
            throw new IllegalArgumentException("Cannot create statistics " +
                    "from an empty list of numbers.");
        ArrayList<N> sortedNums = new ArrayList<N>(numberCollection);
        Collections.sort(sortedNums);

        this.count = sortedNums.size();
        boolean even = (count % 2 == 0);
        int medianIdxFirst = count/2;
        if (even) --medianIdxFirst;
        double i_sum = 0;
        double i_median = -1;
        min = sortedNums.get(0);
        max = sortedNums.get(count-1);
        int i = 0;

        for (N n : sortedNums) {
            if (i==medianIdxFirst) {
                i_median = n.doubleValue();
            } else if (even && i==medianIdxFirst+1) {
                i_median = (i_median + n.doubleValue()) / 2d;
            }
            i_sum+=n.doubleValue();
            ++i;
        }

        this.median = i_median;
        this.sum = i_sum;
        this.mean = sum / ((double)count);
        double i_sd = 0;

        for (Number n : sortedNums) {
            i_sd += (n.doubleValue()-mean)*(n.doubleValue()-mean);
        }

        this.sd = Math.sqrt(i_sd / ((double)count-1));
    }

}
