/*
 *  Copyright 2009 Lari Natri <lari.natri@gmail.com>.
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Lesser General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */



package net.aavalla.utils.general;

import java.security.MessageDigest;

/**
 * Defines static methods to easily create cryptographic digests. Wraps around
 * <code>java.security.MessageDigest</code>.
 * @see java.security.MessageDigest
 * @author Lari Natri
 */
public final class CryptoUtils {

    public final static String MD5(final String text)
    {
        return getDigest(text, "MD5", 128); // NOI18N
    }

    public final static String SHA1(final String text)
    {
        return getDigest(text, "SHA-1", 160); // NOI18N
    }

    public final static String SHA256(final String text)
    {
        return getDigest(text, "SHA-256", 256); // NOI18N
    }

    public final static String SHA512(final String text)
    {
        return getDigest(text, "SHA-512", 512); // NOI18N
    }
    
    /**
     * Returns a hash of a given text generated by given algorithm
     * @param text source string to hash
     * @param algorithm algorithm to use
     * @param algDigestBitSize Digest size of the given algorithm in bits
     * @return Digest as a hex string representation, of length algDigestBitSize/4
     */
    private final static String getDigest(final String text, final String algorithm, final int algDigestBitSize)
    {
        byte[] hash = new byte[algDigestBitSize/4];
        try {
            MessageDigest md = MessageDigest.getInstance(algorithm);
            md.update(text.getBytes("iso-8859-1"), 0, text.length()); // NOI18N
            hash = md.digest();
        } catch (Exception ex) {
            ex.printStackTrace();
        }
        return convertToHex(hash);
    }

    /**
     * Converts byte array to a string of hex numbers, letters are lower-case
     * @param data source data
     */
    private final static String convertToHex(final byte[] data) {
        StringBuffer buf = new StringBuffer();
        for (int i = 0; i < data.length; i++) {
            int halfbyte = (data[i] >>> 4) & 0x0F;
            int two_halfs = 0;
            do {
                if ((0 <= halfbyte) && (halfbyte <= 9)) {
                    buf.append((char)('0' + halfbyte)); // NOI18N
                } else {
                    buf.append((char)('a' + (halfbyte - 10))); // NOI18N
                }
                halfbyte = data[i] & 0x0F;
            } while(two_halfs++ < 1);
        }
        return buf.toString();
    }


}
