/*
 *  Copyright 2009 Lari Natri <lari.natri@gmail.com>.
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Lesser General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */


package net.aavalla.utils.general;

import java.io.BufferedInputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.lang.reflect.Method;
import java.net.URL;
import java.net.URLClassLoader;
import java.util.jar.JarEntry;
import java.util.jar.JarInputStream;

/**
 *
 * @author Lari Natri
 */
public class DynamicClassLoader {

    // Parameters
    private static final Class[] parameters = new Class[]{URL.class};

    /**
     * Add file to CLASSPATH
     * @param s File name
     * @throws IOException  IOException
     */
    public static void addFile(String s) throws IOException {
        File f = new File(s);
        addFile(f);
    }

    /**
     * Add file to CLASSPATH
     * @param f  File object
     * @throws IOException IOException
     */
    public static void addFile(File f) throws IOException {
        addURL(f.toURI().toURL());
    }

    /**
     * Add URL to CLASSPATH
     * @param u URL
     * @throws IOException IOException
     */
    @SuppressWarnings("unchecked")
    public static void addURL(URL u) throws IOException {
        URLClassLoader systemClassLoader = (URLClassLoader) ClassLoader.getSystemClassLoader();
        URL urls[] = systemClassLoader.getURLs();
        for (int i = 0; i < urls.length; i++) {
            if (urls[i].toString().equalsIgnoreCase(u.toString())) {
                System.out.println("DEBUG: URL " + u + " is already in the CLASSPATH.");
                return;
            }
        }
        Class sysclass = URLClassLoader.class;
        try {
            Method method = sysclass.getDeclaredMethod("addURL", parameters);
            method.setAccessible(true);
            method.invoke(systemClassLoader, new Object[]{u});
        } catch (Throwable t) {
            t.printStackTrace();
            throw new IOException("Could not add URL to system classloader.");
        }
    }

    /**
     * Tries to add a .jar file to classpath.
     * @param paths paths to search for the jar
     * @param jarFileName filename without path but with .jar
     * @param testClassName Binary name of a class that should be within the
     *        wanted jar file. If such a class is not found, the jar is not
     *        added to the classpath. Example: compid.utils.DynamicClassLoader
     * @return true if jar was found and added, false otherwise
     */
    public static boolean loadJar(File[] paths, String jarFileName, String testClassName) {
        boolean jarFoundAndAddedToClasspath = false;

        if (testClassName.endsWith(".class")) {
            testClassName = testClassName.substring(0,testClassName.length()-6);
        }
        testClassName.replace('/', '.'); // just for safety
        for (File f : paths) {
            JarInputStream jis = null;
            if (jarFoundAndAddedToClasspath) break;
            try {
                String fullName = f.toString() + File.separator + jarFileName;
                URL url = new URL("file",null,fullName);
                //URLClassLoader urlLoader = new URLClassLoader(new URL[]{url}); // not needed, but leave here for reference

                jis = new JarInputStream(new BufferedInputStream(new FileInputStream(fullName)));
                JarEntry entry = jis.getNextJarEntry();

                while (entry != null) {
                    String entryName = entry.getName();
                    if (entryName.endsWith(".class")) {
                        entryName = entryName.substring(0, entryName.length() - 6);
                        entryName = entryName.replace('/', '.');

                        try {
                            if (entryName.equals(testClassName)) {
                                addURL(url); // adds the JAR to classpath
                                jarFoundAndAddedToClasspath = true;
                                break;
                            }
                            //urlLoader.loadClass(entryName); // not needed but leave here for reference
                        } catch (Throwable e) {
                            System.out.println("\t " + e.getClass().getName() + ": " + e.getMessage());
                        }

                    }
                    entry = jis.getNextJarEntry();
                }

            } catch (Exception ex) {
                continue; // Didn't work: try the next one
            } finally {   // Close the file gracefully
                if (jis!=null) {
                    try {
                        jis.close();
                    } catch (Exception ex) {}
                }
            }
        } // for
        return jarFoundAndAddedToClasspath;
    }

    /**
     * Tries to load a native system library. Given paths are tried first and
     * if they fail, normal system call is used.
     * @param libName Library name
     * @param paths paths to check.
     * @return True, if loading was succesful, false otherwise.
     */
    public static boolean loadNativeLibrary(String libName, File[] paths) {
        if (libName.startsWith("lib") && libName.endsWith(".so")) {
            libName = libName.substring(3, libName.length() - 3);
        } else if (libName.toUpperCase().endsWith(".DLL")) {
            libName = libName.substring(0, libName.length() - 4);
        }
        String[] libFileNames = {"lib" + libName + ".so", // for Unix
            libName + ".DLL"};     // for Windows

        boolean success = false;
        if (paths != null) { // try paths first
            for (File path : paths) {
                if (success) {
                    break;
                }
                for (String libFileName : libFileNames) {
                    String tryFile = path.toString() + File.separator + libFileName;
                    try {
                        System.load(tryFile);
                        success = true;
                        System.out.println("DEBUG: Success loading native library: " + tryFile);
                        break;
                    } catch (UnsatisfiedLinkError ex) {
                        System.out.println("DEBUG: Native library not loaded from " + tryFile + " ("+ex.getMessage()+")");
                    }
                }
            }
        }

        if (!success) {
            try {
                System.loadLibrary(libName);
                success = true;
                System.out.println("DEBUG: Success loading native system library " + libName);
            } catch (UnsatisfiedLinkError ex) {
                System.out.println("DEBUG: Couldn't find system library " + libName);
            }
        }

        return success;

    }
}

