/*
 *  Copyright 2009 Lari Natri <lari.natri@gmail.com>.
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Lesser General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */



package net.aavalla.utils.general;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.IOException;

/**
 *
 * @author Lari Natri
 */
public class FileUtils {

    /** End-Of-Line style (LF, CR or CRLF) */
    public enum EOLStyle {
        UNKNOWN { @Override public final String EOL() { return ""; } },
        LF      { @Override public final String EOL() { return "\n"; } },
        CR      { @Override public final String EOL() { return "\r"; } },
        CRLF    { @Override public final String EOL() { return "\r\n"; } },
        /** Represents the systems default EOL style */
        SYSTEM  { @Override public final String EOL() { return System.getProperty("line.separator"); } };
        /** The actual character(s) that signal EOL */
        public abstract String EOL();
    }

    private static final int BUFFERSIZE = 4096;
    private static final int MAXBYTESFOREOLRECOGNITION = 16777216; // 16MiB

    /** Returns the EOL style used in file given as parameter */
    public static EOLStyle getFileEOLStyle(File file)
            throws FileNotFoundException, IOException
    {
        BufferedReader br = new BufferedReader(new FileReader(file));
        char[] buf = new char[BUFFERSIZE];
        EOLStyle eolStyle = EOLStyle.UNKNOWN;
        boolean found = false;
        boolean lastWasCR = false;
        int nextReadBytes = BUFFERSIZE;
        int lastReadBytes = 0;
        int totalReadBytes = 0;

        while (nextReadBytes>0 && !found && (lastReadBytes = br.read(buf, 0, nextReadBytes))>0) {
            totalReadBytes+=lastReadBytes;
            if (totalReadBytes+BUFFERSIZE > MAXBYTESFOREOLRECOGNITION) {
                nextReadBytes = MAXBYTESFOREOLRECOGNITION - totalReadBytes;
            }
            if (lastWasCR) { // last char of last buffer was CR
                if (buf[0]=='\n') {
                    eolStyle = EOLStyle.CRLF;
                } else {
                    eolStyle = EOLStyle.CR;
                }
                found = true;
                lastWasCR = false;
            }
            for (int i=0; (i<lastReadBytes && !found); ++i) { // buf
                if (buf[i]=='\n') {
                    eolStyle = EOLStyle.LF;
                    found = true;
                } else if (buf[i]=='\r') {
                    if (i>=lastReadBytes-1) {
                        lastWasCR = true;
                    } else if (buf[i+1]=='\n') {
                        eolStyle = EOLStyle.CRLF;
                        found = true;
                    } else {
                        eolStyle = EOLStyle.CR;
                        found = true;
                    }
                }
            } // for
        } // while

        br.close();

        if (eolStyle.equals(EOLStyle.UNKNOWN) && lastWasCR) {
            eolStyle = EOLStyle.CR;
        }

        return eolStyle;

    }

    /**
     * Returns the filesystem path of given class. If class is located in a
     * .jar file, the location of the .jar is returned. null value is returned
     * if path cannot be determined.
     * @param c
     * @return a Directory where class is located on disk or null
     */
    public static File getClassPath(Class c) {
        try {
            File f = new File(c.getProtectionDomain().getCodeSource().getLocation().toURI().getPath());
            f = new File(f.getAbsolutePath());
            if (f.isFile()) {
                f = new File(f.getParent());
            }
            if (f.isDirectory()) {
                return f;
            } else {
                System.out.println(f);
                return null;
            }
        } catch (Exception ex) {
            return null;
        }

    }

}
