/*
 *  Copyright 2010 Lari Natri <lari.natri@gmail.com>.
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Lesser General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */


package net.aavalla.utils.general;

/**
 * Represents a pair of objects. equals() and hashCode() return the same value
 * for reversed pair. hashCode() is (lazily) initialized only once, so member
 * objects should be immutable (or immutated :)). Pair class is itself
 * immutable. Pair's objects cannot be null.
 *
 * @author Lari Natri
 */
public final class Pair<T> {
    public final T o1;                      // first object of the pair
    public final T o2;                      // second object of the pair
    private int hash;                       // holds the hash code
    private boolean hashCalculated = false; // true if hash already calculated

    /**
     * Constructor.
     * @param o1 first object to store in the pair
     * @param o2 second object to store in the pair
     * @throws NullPointerException Thrown if either or both of the objects
     *                              is null
     */
    public Pair(final T o1, final T o2) throws NullPointerException {
        if (o1==null || o2==null) {
            throw new NullPointerException("Tried to initialize a Pair with " +
                    "null value(s)");
        }
        this.o1 = o1;
        this.o2 = o2;
    }

    /**
     * Returns true if (obj.o1==this.o1 && obj.o2==this.o2) ||
     *                 (obj.o1==this.o2 && obj.o2==this.o1).
     * Uses objects' equals() method for '=='.
     * @param obj other object to compare to
     * @return true if objects are equal, false otherwise
     */
    @Override
    public final boolean equals(final Object obj) {
        if (obj == null || this.getClass() != obj.getClass()) {
            return false;
        }
        final Pair other = (Pair)obj;
        if (other.o1.equals(this.o1) && other.o2.equals(this.o2)){
            return true;
        }
        if (other.o1.equals(this.o2) && other.o2.equals(this.o1)){
            return true;
        }
        return false;
    }

    /**
     * Returns same hash for pairs regardless of their order. Hash code
     * is lazily calculated only once when first needed.
     * @return hash code
     */
    @Override
    public final int hashCode() {
        if (!this.hashCalculated) {
            this.hash = this.preCalculateHash();
        }
        return this.hash;
    }

    /**
     * Returns same hash for pairs regardless of their order
     * @return hash code
     */
    private final int preCalculateHash() {
        return 37 * (this.o1.hashCode()*this.o2.hashCode());
    }
}

