/*
 *  Copyright 2009 Lari Natri <lari.natri@gmail.com>.
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Lesser General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */



package net.aavalla.utils.general;

import java.util.Collection;
import java.util.Iterator;

/**
 * This class represents a read only iterator for any Collection object. It
 * uses the default iterator, but there is one difference: remove() method is
 * overwritten with an empty method.
 * @author Lari Natri
 * @param <T> Type of items stored in the collection
 */
public final class ReadonlyIterator<T> implements Iterator<T> {

    /** The default iterator of given Collection */
    private final Iterator<T> origIter;

    /**
     * Constructor
     * @param collection Collection to iterate
     */
    public ReadonlyIterator(Collection<T> collection) {
        this.origIter = collection.iterator();
    }

    @Override
    public final boolean hasNext() {
        return origIter.hasNext();
    }

    @Override
    public final T next() {
        return origIter.next();
    }


    /**
     * <b>Do not call this mehtod.</b> This method only throws an exception
     * and does nothing else. It does not remove any items.
     * @throws java.lang.UnsupportedOperationException Thrown <b>always</b>
     * when this method is called.
     */
    @Override @Deprecated
    public final void remove() throws UnsupportedOperationException {
        throw new UnsupportedOperationException("Removal of elements is not" +
                "supported in ReadonlyIterator.");
    }


} // end class
