/*
 *  Licensed to the Apache Software Foundation (ASF) under one or more
 *  contributor license agreements.  See the NOTICE file distributed with
 *  this work for additional information regarding copyright ownership.
 *  The ASF licenses this file to You under the Apache License, Version 2.0
 *  (the "License"); you may not use this file except in compliance with
 *  the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */

/*
 * This file is from Apache Harmony software package. It is only slightly
 * modified by Lari Natri to allow retrieving of keys by a template key:
 *   - class name was originally java.util.HashSet
 *   - added imports from java.util
 *   - added function getKey(K keyTemplate)
 */

package net.aavalla.utils.general;

import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.Serializable;
import java.util.AbstractSet;
import java.util.Collection;
import java.util.Iterator;
import java.util.Set;

/**
 * HashSet is an implementation of a Set. All optional operations (adding and
 * removing) are supported. The elements can be any objects.
 */
public class RetrievalHashSet<E> extends AbstractSet<E> implements Set<E>, Cloneable,
        Serializable {

    private static final long serialVersionUID = -5024744406713321676L;

    transient RetrievalHashMap<E, RetrievalHashSet<E>> backingMap;

    /**
     * Constructs a new empty instance of {@code HashSet}.
     */
    public RetrievalHashSet() {
        this(new RetrievalHashMap<E, RetrievalHashSet<E>>());
    }

    /**
     * Constructs a new instance of {@code RetrievalHashSet} with the specified capacity.
     *
     * @param capacity
     *            the initial capacity of this {@code RetrievalHashSet}.
     */
    public RetrievalHashSet(int capacity) {
        this(new RetrievalHashMap<E, RetrievalHashSet<E>>(capacity));
    }

    /**
     * Constructs a new instance of {@code RetrievalHashSet} with the specified capacity
     * and load factor.
     *
     * @param capacity
     *            the initial capacity.
     * @param loadFactor
     *            the initial load factor.
     */
    public RetrievalHashSet(int capacity, float loadFactor) {
        this(new RetrievalHashMap<E, RetrievalHashSet<E>>(capacity, loadFactor));
    }

    /**
     * Constructs a new instance of {@code RetrievalHashSet} containing the unique
     * elements in the specified collection.
     *
     * @param collection
     *            the collection of elements to add.
     */
    public RetrievalHashSet(Collection<? extends E> collection) {
        this(new RetrievalHashMap<E, RetrievalHashSet<E>>(collection.size() < 6 ? 11 : collection
                .size() * 2));
        for (E e : collection) {
            add(e);
        }
    }

    RetrievalHashSet(RetrievalHashMap<E, RetrievalHashSet<E>> backingMap) {
        this.backingMap = backingMap;
    }

    /**
     * Adds the specified object to this {@code RetrievalHashSet} if not already present.
     *
     * @param object
     *            the object to add.
     * @return {@code true} when this {@code RetrievalHashSet} did not already contain
     *         the object, {@code false} otherwise
     */
    @Override
    public boolean add(E object) {
        return backingMap.put(object, this) == null;
    }

    /**
     * Removes all elements from this {@code RetrievalHashSet}, leaving it empty.
     *
     * @see #isEmpty
     * @see #size
     */
    @Override
    public void clear() {
        backingMap.clear();
    }

    /**
     * Returns a new {@code RetrievalHashSet} with the same elements and size as this
     * {@code RetrievalHashSet}.
     *
     * @return a shallow copy of this {@code RetrievalHashSet}.
     * @see java.lang.Cloneable
     */
    @Override
    @SuppressWarnings("unchecked")
    public Object clone() {
        try {
            RetrievalHashSet<E> clone = (RetrievalHashSet<E>) super.clone();
            clone.backingMap = (RetrievalHashMap<E, RetrievalHashSet<E>>) backingMap.clone();
            return clone;
        } catch (CloneNotSupportedException e) {
            return null;
        }
    }

    /**
     * Searches this {@code RetrievalHashSet} for the specified object.
     *
     * @param object
     *            the object to search for.
     * @return {@code true} if {@code object} is an element of this
     *         {@code RetrievalHashSet}, {@code false} otherwise.
     */
    @Override
    public boolean contains(Object object) {
        return backingMap.containsKey(object);
    }

    /**
     * Returns true if this {@code RetrievalHashSet} has no elements, false otherwise.
     *
     * @return {@code true} if this {@code RetrievalHashSet} has no elements,
     *         {@code false} otherwise.
     * @see #size
     */
    @Override
    public boolean isEmpty() {
        return backingMap.isEmpty();
    }

    /**
     * Returns an Iterator on the elements of this {@code RetrievalHashSet}.
     *
     * @return an Iterator on the elements of this {@code RetrievalHashSet}.
     * @see Iterator
     */
    @Override
    public Iterator<E> iterator() {
        return backingMap.keySet().iterator();
    }

    /**
     * Removes the specified object from this {@code RetrievalHashSet}.
     *
     * @param object
     *            the object to remove.
     * @return {@code true} if the object was removed, {@code false} otherwise.
     */
    @Override
    public boolean remove(Object object) {
        return backingMap.remove(object) != null;
    }

    /**
     * Returns the number of elements in this {@code RetrievalHashSet}.
     *
     * @return the number of elements in this {@code RetrievalHashSet}.
     */
    @Override
    public int size() {
        return backingMap.size();
    }

    public E getKey(E templateKey) {
        return backingMap.getKey(templateKey);
    }

    private void writeObject(ObjectOutputStream stream) throws IOException {
        stream.defaultWriteObject();
        stream.writeInt(backingMap.elementData.length);
        stream.writeFloat(backingMap.loadFactor);
        stream.writeInt(backingMap.elementCount);
        for (int i = backingMap.elementData.length; --i >= 0;) {
            RetrievalHashMap.Entry<E, RetrievalHashSet<E>> entry = backingMap.elementData[i];
            while (entry != null) {
                stream.writeObject(entry.key);
                entry = entry.next;
            }
        }
    }

    @SuppressWarnings("unchecked")
    private void readObject(ObjectInputStream stream) throws IOException,
            ClassNotFoundException {
        stream.defaultReadObject();
        int length = stream.readInt();
        float loadFactor = stream.readFloat();
        backingMap = createBackingMap(length, loadFactor);
        int elementCount = stream.readInt();
        for (int i = elementCount; --i >= 0;) {
            E key = (E) stream.readObject();
            backingMap.put(key, this);
        }
    }

    RetrievalHashMap<E, RetrievalHashSet<E>> createBackingMap(int capacity, float loadFactor) {
        return new RetrievalHashMap<E, RetrievalHashSet<E>>(capacity, loadFactor);
    }
}
