/*
 * Copyright 2009 Lari Natri <lari.natri@gmail.com>
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Lesser General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package net.aavalla.utils.general;


/**
 * This class contains static methods for statistical analysis. The class is
 * thread safe.
 *
 * @author Lari Natri
 */
public class Statistics {

    public static enum TailedType {
        TWO,
        ONE_SMALLEXTREME,
        ONE_LARGEEXTREME,
        NOTAIL
    }

    
    private static final int LNFACTSBUFFERSIZE = 64;
    private static final Object lnFactsCacheLock = new Object();
    private static double[] lnFactsCache = null;
    
    
    /**
     * Calculates the factorial of parameter. Result might be imprecise with
     * large values.
     * 
     * @param n number (>=0) to calculate the factorial from
     * @return n!
     * @throws IllegalArgumentException if parameter < 0
     */
    public static long factorial(int n) {
        if (n < 0) {
            throw new IllegalArgumentException("Cannot compute factorial for " +
                                               "a negative number (" + n + ")");
        }
        return Math.round(Math.exp(getLnFactorials(n)[n]));
    }

    /**
     * Fisher exact test.
     * 
     * @param k hits acquired
     * @param N total number of items
     * @param m maximum number of possible hits
     * @param n number of tests done
     * @param tailed type of tailed test to execute
     * @return p-value
     * @throws IllegalArgumentException if parameter values are illegal
     */
    public static double fisherExactTest(int k, int N, int m, int n,
                                         TailedType tailed)
    {
        double p = hyperGeomTest(k,N,m,n); // throws exception if wrong values
        int kTest = 0;
        if (tailed.equals(TailedType.TWO)
                || tailed.equals(TailedType.ONE_SMALLEXTREME)) {
            kTest = k-1;
            while (kTest >= 0 && N+kTest-n-m >=0) {
                p += hyperGeomTestNoChecks(kTest,N,m,n);
                --kTest;
            }
        }
        if (tailed.equals(TailedType.TWO)
                || tailed.equals(TailedType.ONE_LARGEEXTREME)) {
            kTest = k+1;
            while (n-kTest >= 0 && m-kTest >= 0) {
                p += hyperGeomTestNoChecks(kTest,N,m,n);
                ++kTest;
            }
        }
        return p;
    }
    
    /**
     * Hypergeometric test.
     * 
     * @param  k hits acquired
     * @param  N total number of items
     * @param  m maximum number of possible hits
     * @param  n number of tests done
     * @return probability
     * @throws IllegalArgumentException if parameter values are illegal
     */
    public static double hyperGeomTest(int k, int N, int m, int n)
    {
        if (k<0 || N<0 || m<0 || n<0 || k>m || k>n) {
            throw new IllegalArgumentException("Illegal parameters for " +
                                               "hypergeometric test");
        }
        return hyperGeomTestNoChecks(k,N,m,n);
    }
    
    /**
     * Empties the factorial cache to free memory. Should only be called if
     * very big factorials have been calculated and factorials of similar size
     * are not needed in the near future.
     */
    public static void emptyFactorialCache() {
        synchronized(lnFactsCacheLock) {
            lnFactsCache = null;
        }
    }

    /**
     * Hypergeometric test. Doesn't check for parameter correctness.
     * 
     * @param  k hits acquired
     * @param  N total number of items
     * @param  m maximum number of possible hits
     * @param  n number of tests done
     * @return probability
     */    
    private static double hyperGeomTestNoChecks(int k, int N, int m, int n) {
        double[] lnFacts = getLnFactorials(N);
        return Math.exp(lnFacts[m] + lnFacts[N-m] + lnFacts[n]   + lnFacts[N-n]
                      - lnFacts[N] - lnFacts[k]   - lnFacts[m-k] - lnFacts[n-k] 
                      - lnFacts[N+k-n-m]);        
    }
    
    /**
     * Returns an array of doubles where getLnFactorials(0) = 0 and
     * getLnFactorials(i>0) = getLnFactorials(i-1) + ln(i). New items
     * are created in static cache when needed.
     * 
     * @param maxNeededFactorialOf maximum factorial to get
     * @return array of "logarithmic factorials"
     */
    private static final double[] getLnFactorials(int maxNeededFactorialOf) {
        synchronized(lnFactsCacheLock) {
            if (lnFactsCache==null || lnFactsCache.length==0) {
                lnFactsCache = new double[1]; // java init's as 0.0d
            }

            if (maxNeededFactorialOf < 0) {
                return lnFactsCache; // as is
            }

            if (lnFactsCache.length-1 < maxNeededFactorialOf) {
                double[] newLnFacts =
                        new double[LNFACTSBUFFERSIZE
                                 * (maxNeededFactorialOf/LNFACTSBUFFERSIZE)
                                 + LNFACTSBUFFERSIZE];
                System.arraycopy(lnFactsCache, 0, newLnFacts, 0,
                                 lnFactsCache.length);
                newLnFacts[0] = 0d;
                for (int i = lnFactsCache.length; i < newLnFacts.length; ++i) {
                    newLnFacts[i] = newLnFacts[i-1] + Math.log(i);
                }
                lnFactsCache = newLnFacts;
            }
            return lnFactsCache;
        } // synchronized
    }

}
