/*
 * Copyright 2009 Lari Natri <lari.natri@gmail.com>
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Lesser General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */


package net.aavalla.utils.general;

import java.util.regex.Matcher;
import java.util.regex.Pattern;

/**
 * Utilities for manipulating String objects.
 *
 * @author Lari Natri
 */
public class StrUtils {

//    public static final String NICECHARS
//            = "a-zA-Z0-9 _-+.,:;\\!\\@\\(\\)\\[\\]\\{\\}\\#\\%\\*"; // NOI18N
//    public static final String RE_NICECHAR = "[" + NICECHARS + "]"; // NOI18N
//    public static final String RE_NICESTRING = RE_NICECHAR + "*"; // NOI18N
//    public static final String RE_NONNICECHAR = "[^" + NICECHARS + "]"; // NOI18N


    /**
     * Checks whether any substring in a string matches given pattern
     * @param test string to test matching for
     * @param pattern the pattern
     * @return true, if matches are found, false otherwise
     */
    public static boolean containsPattern(String test, Pattern pattern) {
        return pattern.matcher(test).find();
    }


    /**
     * Removes any substrings matching given pattern
     * @param source source string to remove matches from
     * @param unwanted pattern of unwanted substrings
     * @return string with unwanted substrings removed
     */
    public static String removeUnwanted(String source, Pattern unwanted) {
        Matcher m = unwanted.matcher(source);
        StringBuffer sb = new StringBuffer();

        while (m.find()) {
            m.appendReplacement(sb, ""); // NOI18N
        }
        m.appendTail(sb);

        return sb.toString();
    }

    /**
     * Escapes any substrings matching given pattern
     * @param source source string to escape
     * @param unwanted unwanted pattern
     * @param escapePre string to add before any unwanted pattern
     * @param escapePost string to add after any unwanted pattern
     * @return escaped string
     */
    public static String escapeUnwanted(String source, Pattern unwanted,
                                        String escapePre, String escapePost)
    {
        if (escapePre==null) escapePre = "";
        if (escapePost==null) escapePost = "";
        if (escapePre.contains("$") || escapePre.contains("\\") ||
                escapePost.contains("$") || escapePost.contains("\\")) { // NOI18N
            throw new IllegalArgumentException("Escape strings cannot contain $ or \\");
        }


        Matcher m = unwanted.matcher(source);
        StringBuffer sb = new StringBuffer();

        while (m.find()) {
            m.appendReplacement(sb, escapePre + "$g" + escapePost); // NOI18N
        }
        m.appendTail(sb);

        return sb.toString();
    }

    /**
     * Concatenates strings, puts separator between them and returns the result.
     * Null Strings are ignored, but empty ones are "catenated" with separator
     * between them.
     * 
     * @param strings Array of Strings to catenate
     * @param separator Added between concatenated strings
     * @return Concatenated String
     */
    public static String catStrings(Object[] strings, String separator) {
        if (strings==null || strings.length==0) return "";
        if (separator==null) separator = "";
        StringBuilder buf = new StringBuilder();
        for (int i=0; i<strings.length; ++i) {
            if (strings[i]!=null) {
                buf.append((String)strings[i]);
                if (i<strings.length-1) buf.append(separator);
            }
        }
        return buf.toString();
    }
    
    /**
     * Returns a string concatenated with itself a number of times with a
     * separator string added in between. ' characters are doubled:
     * "'" -> "''".
     * 
     * @param string String to concatenate with itself
     * @param separator Added between concatenated strings
     * @param c Number of times
     * @return Concatenated String
     */
    public static String catStringTimes(String string, String separator, int c) {
        string = string.replaceAll("'", "''");
        if (string==null||c<1) return "";
        if (separator==null) separator = "";
        StringBuilder buf = new StringBuilder();
        for (int i=0; i<c; ++i) {
            buf.append(string);
            if (i<c-1) buf.append(separator);
        }
        return buf.toString();        
    }

    /**
     * Slices a string to substrings of given length and return them as an
     * array of strings. Last string might be shorter.
     * @param sourceString Original string
     * @param sliceLength length of a slice
     * @return sliced string
     */
    public static String[] slice(final String sourceString, int sliceLength) {
        if(sliceLength<1) return new String[0];

        int sliceCount = sourceString.length() / sliceLength;
        if (sliceCount*sliceLength < sourceString.length()) ++sliceCount;

        int sourceIndex = 0;
        int endIndex;

        String[] result = new String[sliceCount];
        for (int i = 0; i < sliceCount; ++i) {
            endIndex = sourceIndex + sliceLength;
            if (endIndex > sourceString.length()) {
                endIndex = sourceString.length();
            }
            result[i] = sourceString.substring(sourceIndex,endIndex);
            sourceIndex += result[i].length();
        }

        return result;
    }

    public static String asString(double d) {
        return asString(d,0,8);
    }

    public static String asString(double d, int minDecimals, int maxDecimals) {
        String s = String.valueOf(d);
        int pointIdx = s.indexOf(".");
        if (pointIdx < 0) pointIdx = s.indexOf(",");
        if (pointIdx < 0 && minDecimals > 0) { // no decimal point; ever happen?
            pointIdx = s.length();
            s = s + ".";
        }
        if (pointIdx > 0) {
            int decCount = s.length()-1-pointIdx;
            while (decCount < minDecimals) {
                s = s + "0";
                ++decCount;
            }
            if (decCount > maxDecimals) {
                double p = Math.pow(10, maxDecimals);
                d = d * p;
                double tmp = Math.round(d);
                s = String.valueOf((double)tmp/p);
                decCount = maxDecimals;
            }
        }
        if (minDecimals <= 0 && pointIdx >= 0) {
            String tmp = s.substring(pointIdx+1);
            if (tmp.replaceAll("0", "").equals("")) {
                s = s.substring(0,pointIdx);
            }
        }
        return s;
    }


}
