/*
 * Copyright 2009 Lari Natri <lari.natri@gmail.com>
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Lesser General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */


package net.aavalla.utils.general;

/**
 * This class can be used to time operations.
 * @author Lari Natri
 */
public final class Timer {

    private long startTime = -1;
    private long endTime = -1;
    private long lastPauseStartTime = -1;
    private long previouslyPaused = 0;
    private boolean isPaused = false;

    /** Default constructor */
    public Timer() {}

    /**
     * Resets all aspects of the timer and (re)starts it.
     */
    public final synchronized void start() {
        this.startTime = System.nanoTime();
        this.endTime = -1;
        this.isPaused = false;
        this.lastPauseStartTime = -1;
        this.previouslyPaused = 0;
    }

    /**
     * Stops the timer.
     */
    public final synchronized void stop() {
        final long now = System.nanoTime();
        if (isPaused) {
            this.isPaused = false;
            this.previouslyPaused += now-this.lastPauseStartTime;
        }
        this.endTime = now;
    }

    /**
     * Pauses the timer. If timer isn't started yet, it will be started and
     * immediately put to paused state.
     */
    public final synchronized void pause() {
        if (this.startTime<0) {
            this.start();
            this.lastPauseStartTime = this.startTime;
            this.isPaused = true;
        }
        if (this.isPaused) return;
        this.isPaused = true;
        this.lastPauseStartTime = System.nanoTime();
    }

    /**
     * Unpauses the timer. If timer isn't started yet, this will also start it.
     */
    public final synchronized void unPause() {
        if (this.startTime<0) {
            this.start();
            this.isPaused = false;
            return;
        }
        if (!this.isPaused) return;
        this.previouslyPaused += (System.nanoTime()-this.lastPauseStartTime);
        this.isPaused = false;
    }

    /**
     * Returns the time in nanoseconds spent after this timer has started, not
     * counting the time it's been paused.
     * @return current non-paused execution time, in nanoseconds
     */
    public final synchronized long getExecutionTime() {
        if (startTime<0) return 0;
        if (this.endTime>=0) return (this.endTime-this.startTime-this.previouslyPaused);
        final long now = System.nanoTime();
        long runTime = (now - this.startTime - this.previouslyPaused);
        if (this.isPaused) {
            runTime -= (now - this.lastPauseStartTime);
        }
        return runTime;
    }

    /**
     * Returns the time in nanoseconds spent after this timer has started, not
     * counting the time it's been paused.
     * @return current non-paused execution time, in seconds
     */
    public final synchronized double getExecutionTimeInSeconds() {
        return ((double)this.getExecutionTime())/1000000000d;
    }

    /**
     * Returns the time this timer has spent in paused state. Returns 0, if
     * timer hasn't been started yet.
     * @return time spent in paused mode, in nanoseconds
     */
    public final synchronized long getPausedTime() {
        if (this.startTime<0) return 0;
        if (this.isPaused) {
            return this.previouslyPaused + System.nanoTime()
                    - this.lastPauseStartTime;
        } else {
            return this.previouslyPaused;
        }
    }

    /**
     * Returns the time this timer has spent in paused state. Returns 0, if
     * timer hasn't been started yet.
     * @return time spent in paused mode, in seconds
     */
    public final synchronized double getPausedTimeInSeconds() {
        return ((double)this.getPausedTime())/1000000000d;
    }

    /**
     * Returns true, if the timer is running (even if it is paused).
     */
    public final synchronized boolean isStarted() {
        return (this.startTime>=0 && this.endTime<0);
    }

    /**
     * Returns true, if the timer is currently paused. Returns false, if the
     * timer hasn't been started yet.
     * @return paused-state of the timer
     */
    public final synchronized boolean isPaused() {
        return this.isPaused;
    }

}
