/*
 * Copyright 2010 Lari Natri <lari.natri@gmail.com>
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Lesser General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package net.aavalla.utils.general;

/**
 * Wrapper for an object 'value' implementing Comparable interface.
 * UniqueComparable's compareTo() returns the result of the underlying
 * value object's compareTo() except if it would return 0. In that case
 * UniqueComparable's compareTo() compares the age of the objects. The
 * only time UniqueComparable's compareTo() ever returns 0, is if the
 * UniqueComparable objects are actually the same object.
 *
 * The idea is that values can be ordered and kept unique. This way, for
 * example, an OrderedSet can hold objects with equal values while keeping
 * consistent with equals().
 *
 * Immutable. Thread-safe.
 *
 * @author Lari Natri
 */
public final class UniqueComparable<T extends Comparable> 
        implements Comparable<UniqueComparable<T>>
{
    // lock for synchronization
    private static final Object uniquenessLock = new Object();
    // counter
    private static long uniquenessNumber = Long.MIN_VALUE;

    private final T value;     // value held
    private final long unique; // unique value

    /**
     * Constructor
     * @param value value to hold
     * @throws NullPointerException Thrown, if trying to add null value
     */
    public UniqueComparable(final T value) {
        if (value==null) throw new NullPointerException();
        this.value = value;
        synchronized(uniquenessLock) {
            if (uniquenessNumber==Long.MAX_VALUE) {
                uniquenessNumber = Long.MIN_VALUE;
            }
            this.unique = uniquenessNumber++;
        }
    }

    /**
     * Returns the value held.
     * @return value
     */
    public final T value() {
        return this.value;
    }

    @Override @SuppressWarnings("unchecked")
    public final int compareTo(final UniqueComparable<T> o) {
        if (this==o) {
            return 0;
        }
        if (this.value.compareTo((T)o.value)==0) {
            if (this.unique<o.unique) {
                return 1;
            } else {
                return -1;
            }
        }

        return this.value.compareTo((T)o.value);
    }

}

