/*
 * Copyright 2010 Lari Natri <lari.natri@gmail.com>
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Lesser General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package net.aavalla.utils.general;

import java.lang.reflect.Array;

/**
 *
 * @author Lari Natri
 */
public class VirtualMatrix<T> {

    private final T[][] matrix;
    private final int maxDist;
    private final int midIdx;
    private final int virtualSizeJ;
    private final T outValue;

    @SuppressWarnings("unchecked")
    public VirtualMatrix(int sizeI, int sizeJ, int maxDist, 
                         T sampleValue, T outValue) throws NullPointerException
    {
        int realSizeJ = 1 + 2*maxDist + Math.abs(sizeI-sizeJ);
        this.midIdx = (realSizeJ+sizeI-sizeJ)/2;

        
        this.matrix = (T[][])Array.newInstance(sampleValue.getClass(),
                                               sizeI,realSizeJ); // throws
        this.maxDist = maxDist;
        this.virtualSizeJ = sizeJ;
        this.outValue = outValue;
//        System.out.println("DEBUG: VirtualMatrix size=("+matrix.length+","+matrix[0].length+") for ("+sizeI+","+sizeJ+"); maxDist="+maxDist+", midIdx="+midIdx);
    }

    public final T get(final int i, final int j)
            throws IndexOutOfBoundsException
    {
        if (i < 0 || i >= matrix.length || j < 0 || j >= virtualSizeJ) {
            throw new IndexOutOfBoundsException("i="+i+", j="+j);
        }

        // get real indexes from virtual ones
        final int realI;
        final int realJ;
        if (i == j) {
            realI = i;
            realJ = this.midIdx;
        } else if (i > j) {
            realI = j;
            realJ = this.midIdx + j - i;
        } else { // i < j
            realI = i;
            realJ = this.midIdx - i + j;
        }

        if (realJ < 0 || realJ >= matrix[0].length) {
//            System.out.println("DEBUG: VirtualMatrix: get from outside");
            return this.outValue;
        }

        synchronized(this.matrix) {

            return this.matrix[realI][realJ];
        }
    }

    public final void set(final int i, final int j, final T value)
            throws IndexOutOfBoundsException
    {
        if (i < 0 || i >= matrix.length || j < 0 || j >= virtualSizeJ) {
            throw new IndexOutOfBoundsException("i="+i+", j="+j);
        }

        // get real indexes from virtual ones
        final int realI;
        final int realJ;
        if (i == j) {
            realI = i;
            realJ = this.midIdx;
        } else if (i > j) {
            realI = j;
            realJ = this.midIdx + j - i;
        } else { // i < j
            realI = i;
            realJ = this.midIdx - i + j;
        }

        if (realJ < 0 || realJ >= matrix[0].length) {
//            System.out.println("DEBUG: VirtualMatrix: SET to outside");
            // outside of maximum distance from virtual diagonal: return
            return;
        }

        synchronized(this.matrix) {
            this.matrix[realI][realJ] = value; // set value
        }
    }

    public int getMaxDist() {
        return this.maxDist;
    }

}
