/*
 *  Copyright 2010 Lari Natri <lari.natri@gmail.com>.
 * 
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Lesser General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package net.aavalla.utils.gui.docbrowser;

import java.io.BufferedInputStream;
import java.io.IOException;
import java.lang.reflect.InvocationTargetException;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.HashMap;
import javax.swing.SwingUtilities;
import javax.swing.tree.DefaultMutableTreeNode;
import javax.swing.tree.DefaultTreeModel;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.parsers.SAXParser;
import javax.xml.parsers.SAXParserFactory;
import org.xml.sax.Attributes;
import org.xml.sax.SAXException;
import org.xml.sax.helpers.DefaultHandler;

/**
 *
 * @author Lari Natri <lari.natri@gmail.com>
 */
public class IndexData {
    
    public static final String ENAME_PAGE = "page"; //NOI18N
    public static final String ANAME_FILE = "file"; //NOI18N
    public static final String ANAME_NAME = "name"; //NOI18N
    public static final String ANAME_ID   = "id";   //NOI18N
    public static final String ANAME_URL  = "url";  //NOI18N

    private final DefaultMutableTreeNode treeRoot = new DefaultMutableTreeNode("root"); //NOI18N
    private final DefaultTreeModel treeModel = new DefaultTreeModel(treeRoot);
    private final HashMap<URL,DefaultMutableTreeNode> urlToNodeMap = new HashMap<URL,DefaultMutableTreeNode>();
    private final HashMap<String,URL> idToUrlMap = new HashMap<String,URL>();
    private final URL docRoot;
    private final String indexFileName;
    private Throwable lastError = null;
    private boolean loadingIndex = false;

    public IndexData(URL docRoot, String indexFileName) {
        super();
        this.docRoot = docRoot;
        this.indexFileName = indexFileName;
    }

    public void loadIndexInBackground()
    {
        if (this.indexFileName==null) {
            synchronized(this) {
                this.lastError = new Throwable("Help index not set");
            }
            System.out.println("DEBUG: help index not set");
            return;
        }

        Runnable r = new Runnable() {
            @Override public void run() {
                synchronized(IndexData.this) {
                    IndexData.this.loadingIndex = true;
                }
                Throwable exception = null;
                try {
                    final IndexParser handler = new IndexParser();
                    final SAXParserFactory factory = SAXParserFactory.newInstance();
                    factory.setValidating(false);
                    final URL indexUrl;
                    if (IndexData.this.docRoot==null) {
                        indexUrl = new URL(IndexData.this.indexFileName);
                    } else {
                        indexUrl = new URL(IndexData.this.docRoot.toString()+"/"+IndexData.this.indexFileName); //NOI18N
                    }
                    final SAXParser parser = factory.newSAXParser();
                    final BufferedInputStream bis = new BufferedInputStream(indexUrl.openStream());
                    parser.parse(bis, handler);
                    bis.close();
                    
                    Runnable swingR = new Runnable() {
                        @Override public void run() {
                            synchronized(IndexData.this.treeRoot) {
                                IndexData.this.treeRoot.removeAllChildren();
                                while (handler.newRoot.getChildCount()>0) {
                                    IndexData.this.treeRoot.add((DefaultMutableTreeNode)handler.newRoot.getFirstChild());
                                }
                                idToUrlMap.clear();
                                idToUrlMap.putAll(handler.newIdToUrlMap);
                                urlToNodeMap.clear();
                                urlToNodeMap.putAll(handler.newUrlToNodeMap);
                                IndexData.this.treeModel.reload();
                            }
                        }
                    };
                    if (SwingUtilities.isEventDispatchThread()) {
                        swingR.run();
                    } else try {
                        SwingUtilities.invokeAndWait(swingR);
                    } catch (InterruptedException ex) {
                    } catch (InvocationTargetException ex) {}

                } catch (ParserConfigurationException ex) {
                    exception = ex;
                } catch (IOException ex) {
                    exception = ex;
                } catch (SAXException ex) {
                    exception = ex;
                }
                if (exception!=null) {
                    synchronized(IndexData.this) {
                        IndexData.this.lastError = exception;
                    }
                    exception.printStackTrace();
                }
                synchronized(IndexData.this) {
                    IndexData.this.loadingIndex = false;
                    IndexData.this.notifyAll();
                }
            }
        };
        r.run();
    }

    public synchronized boolean hasErrors() {
        return this.lastError!=null;
    }

    public synchronized Throwable getLastError() {
        return this.lastError;
    }

    public boolean isLoadingIndex() {
        synchronized(this) {
            return this.loadingIndex;
        }
    }
    
    public void waitForIndex() {
        synchronized(this) {
            if (this.loadingIndex) {
                try {
                    this.wait();
                } catch (InterruptedException ex) {}
            }
        }
    }

    DefaultTreeModel getTreeModel() {
        return this.treeModel;
    }

    DefaultMutableTreeNode getTreeRoot() {
        return this.treeRoot;
    }

    URL getDocRootUrl() {
        return this.docRoot;
    }

    String getIndexFileName() {
        return this.indexFileName;
    }

    URL getUrlById(String id) {
        this.waitForIndex();
        synchronized(this.treeRoot) {
            return this.idToUrlMap.get(id);
        }
    }

    DefaultMutableTreeNode getNodeByUrl(URL url) {
        this.waitForIndex();
        synchronized(this.treeRoot) {
            return this.urlToNodeMap.get(url);
        }
    }


    private class IndexParser extends DefaultHandler {

        private final DefaultMutableTreeNode newRoot = new DefaultMutableTreeNode("root");
        private final HashMap<URL,DefaultMutableTreeNode> newUrlToNodeMap = new HashMap<URL,DefaultMutableTreeNode>();
        private final HashMap<String,URL> newIdToUrlMap = new HashMap<String,URL>();
        private DefaultMutableTreeNode curParent = newRoot;

        @Override
        public void endElement(String uri, String localName, String qName) throws SAXException {
            if (qName.toLowerCase().equals(ENAME_PAGE)) {
                curParent = (DefaultMutableTreeNode)curParent.getParent();
            }
        }

        @Override
        public void startElement(String uri, String localName, String qName, Attributes attrs) throws SAXException {
            if (qName.toLowerCase().equals(ENAME_PAGE) && attrs!=null) {
                String file = null;
                String name = null;
                String url = null;
                String id = null;
                for (int i=0; i<attrs.getLength(); ++i) {
                    String attrName = attrs.getQName(i).toLowerCase();
                    if (attrName.equals(ANAME_FILE)) {
                        file = attrs.getValue(i);
                    } else if (attrName.equals(ANAME_NAME)) {
                        name = attrs.getValue(i);
                    } else if (attrName.equals(ANAME_URL)) {
                        url = attrs.getValue(i);
                    } else if (attrName.equals(ANAME_ID)) {
                        id = attrs.getValue(i);
                    }
                }
                URL pageUrl = null;
                try {
                    if (url!=null) {
                        pageUrl = new URL(url);
                    } else if (file!=null) {
                        pageUrl = new URL(IndexData.this.docRoot.toString()+"/"+file);
                    }
                } catch (MalformedURLException ex) {
                    ex.printStackTrace();
                }
                if (pageUrl!=null) {
                    PageInfo page = new PageInfo(pageUrl,(name==null ? pageUrl.toString() : name),id);
                    DefaultMutableTreeNode node = new DefaultMutableTreeNode(page);
                    this.curParent.add(node);
                    this.curParent = node;
                    this.newUrlToNodeMap.put(pageUrl, node);
                    if (id!=null) {
                        this.newIdToUrlMap.put(id, pageUrl);
                    }
                }
            }
        }

    }


}
