/*
 *  Copyright 2010 Lari Natri <lari.natri@gmail.com>.
 * 
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Lesser General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */


package net.aavalla.utils.gui.docbrowser;

import net.aavalla.utils.general.BareBonesBrowserLaunch;
import net.aavalla.utils.general.StrUtils;
import java.io.IOException;
import java.net.URL;
import java.util.Enumeration;
import java.util.LinkedHashSet;
import java.util.LinkedList;
import java.util.Set;
import javax.swing.ImageIcon;
import javax.swing.JFrame;
import javax.swing.JOptionPane;
import javax.swing.SwingUtilities;
import javax.swing.event.HyperlinkEvent;
import javax.swing.event.HyperlinkListener;
import javax.swing.event.TreeSelectionEvent;
import javax.swing.event.TreeSelectionListener;
import javax.swing.tree.DefaultMutableTreeNode;
import javax.swing.tree.DefaultTreeSelectionModel;
import javax.swing.tree.TreeNode;
import javax.swing.tree.TreePath;
import javax.swing.tree.TreeSelectionModel;


/**
 *
 * @author Lari Natri <lari.natri@gmail.com>
 */
public class SimpleDocBrowser extends JFrame
                              implements HyperlinkListener,
                                         TreeSelectionListener
{
    public static final int MAXHISTORYSIZE_DEFAULT = 10;

    private final URL docRoot;
    private final IndexData indexData;
    private final LinkedList<URL> history;
    private final LinkedHashSet<String> extBrowserPrefixes;
    private int maxHistorySize = MAXHISTORYSIZE_DEFAULT;
    private int curHistoryIdx = -1;

    /** Creates new form SimpleDocBrowser */
    public SimpleDocBrowser(URL docRoot, String indexFileName, boolean useIcons, String title) {
        initComponents();

        this.indexData = new IndexData(docRoot,indexFileName);
        this.docRoot = docRoot;
        this.pageTextPane.setEditable(false);
        this.pageTextPane.addHyperlinkListener(this);
        this.indexTree.setModel(this.indexData.getTreeModel());
        this.indexTree.addTreeSelectionListener(this);
        this.indexTree.setRootVisible(false);
        this.indexTree.setCellRenderer(new HelpTreeRenderer());
        if (indexFileName==null || indexFileName.isEmpty()) {
            this.indexTree.setVisible(false);
            this.previousButton.setVisible(false);
            this.nextButton.setVisible(false);
            this.splitPane.setDividerLocation(0);
            this.splitPane.setDividerSize(0);
        } else {
            this.indexData.loadIndexInBackground();
        }
        DefaultTreeSelectionModel selMod = new DefaultTreeSelectionModel();
        selMod.setSelectionMode(TreeSelectionModel.SINGLE_TREE_SELECTION);
        this.indexTree.setSelectionModel(selMod);

        this.history = new LinkedList<URL>();

        this.previousButton.setEnabled(false);
        this.nextButton.setEnabled(false);
        this.backButton.setEnabled(false);
        this.forwardButton.setEnabled(false);

        if (useIcons) {
            SwingUtilities.invokeLater(new Runnable() {
                @Override public void run() {
                    URL iUrl = getClass().getResource("resources/go-previous.png");
                    if (iUrl != null) {
                        backButton.setIcon(new ImageIcon(iUrl));
                    }
                    iUrl = getClass().getResource("resources/go-next.png");
                    if (iUrl != null) {
                        forwardButton.setIcon(new ImageIcon(iUrl));
                    }
                    iUrl = getClass().getResource("resources/arrow-up.png");
                    if (iUrl != null) {
                        previousButton.setIcon(new ImageIcon(iUrl));
                    }
                    iUrl = getClass().getResource("resources/arrow-down.png");
                    if (iUrl != null) {
                        nextButton.setIcon(new ImageIcon(iUrl));
                    }
                    iUrl = getClass().getResource("resources/system-help.png");
                    if (iUrl != null) {
                        SimpleDocBrowser.this.setIconImage(new ImageIcon(iUrl).getImage());
                    }                    
                }
            });
        }
        if (title!=null) {
            this.setTitle(title);
        }

        this.extBrowserPrefixes = new LinkedHashSet<String>(4);
        this.extBrowserPrefixes.add("http:");
        this.extBrowserPrefixes.add("https:");
        this.extBrowserPrefixes.add("mailto:");
        this.extBrowserPrefixes.add("ftp:");

    }

    public SimpleDocBrowser(URL docRoot, boolean useIcons, String title) {
        this(docRoot, null, useIcons, title);
    }

    public SimpleDocBrowser(boolean useIcons, String title) {
        this(null,null,useIcons, title);
    }

    public SimpleDocBrowser(String title) {
        this(null,null,true,title);
    }

    public SimpleDocBrowser() {
        this(null,null,true,null);
    }

    /**
    * @param args the command line arguments
    */
    public static void main(final String args[]) {
        java.awt.EventQueue.invokeLater(new Runnable() {
            @Override public void run() {
                SimpleDocBrowser sdb = new SimpleDocBrowser(getClass().getResource(StrUtils.catStrings(args, "%20")),true,null); //NOI18N
                sdb.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
                sdb.setVisible(true);
            }
        });
    }

    public synchronized Set<String> getPrefixesForExternal() {
        return new LinkedHashSet<String>(this.extBrowserPrefixes);
    }

    public synchronized void clearPrefixesForExternal() {
        this.extBrowserPrefixes.clear();
    }
    
    public synchronized void setPrefixesForExternal(Set<String> prefixes) {
        this.extBrowserPrefixes.addAll(prefixes);
    }

    public synchronized void addPrefixForExternal(String prefix) {
        this.extBrowserPrefixes.add(prefix);
    }


    public synchronized boolean canGoBack() {
        return (curHistoryIdx > 0);
    }

    public synchronized boolean canGoForward() {
        return (curHistoryIdx < history.size()-1);
    }

    public synchronized boolean goBack() {
        if (!this.canGoBack()) {
            System.out.println("DEBUG: DocBrowser: cannot go back");
            return false;
        }

        URL pageUrl = history.get(curHistoryIdx-1);
        curHistoryIdx-=2;
        this.fireUpdateSelectionAndShowPage(pageUrl);
        return true;
    }

    public synchronized boolean goForward() {
        if (!this.canGoForward()) {
            return false;
        }
        URL pageUrl = history.get(curHistoryIdx+1);
        this.fireUpdateSelectionAndShowPage(pageUrl);
        return true;
    }

    public synchronized void showPageById(String pageId) throws IllegalStateException,
                                                                IllegalArgumentException
    {
        if (this.docRoot==null) {
            throw new IllegalStateException("Document root is not set");
        }
        URL url = this.indexData.getUrlById(pageId);
        if (url==null) {
            throw new IllegalArgumentException("Page not found by id");
        }
        this.showPage(url);
    }

    public synchronized void showPage(URL pageUrl) {
        this.fireUpdateSelectionAndShowPage(pageUrl);
    }

    @Override
    public synchronized void hyperlinkUpdate(HyperlinkEvent e) {
        if(e.getEventType() != HyperlinkEvent.EventType.ACTIVATED) {
            return;
        }
        this.fireUpdateSelectionAndShowPage(e.getURL());
    }

    @Override
    public synchronized void valueChanged(TreeSelectionEvent e) {
        if (!e.isAddedPath()) {
            return;
        }
        Object obj = ((DefaultMutableTreeNode)e.getPath().getLastPathComponent()).getUserObject();
        if (obj instanceof PageInfo) {
            URL url = ((PageInfo)obj).getUrl();
            this.updatePageAndButtons(url);
        }
    }

    public void expandAll() {
        this.expandAll(true);
    }

    public void collapseAll() {
        this.expandAll(false);
    }

    private void expandAll(boolean expand) {
        TreeNode root = (TreeNode)this.indexTree.getModel().getRoot();
        this.expandAll(new TreePath(root), expand);
    }
    
    private void expandAll(TreePath parent, boolean expand) {
        TreeNode node = (TreeNode)parent.getLastPathComponent();
        if (node.getChildCount() >= 0) {
            for (Enumeration e=node.children(); e.hasMoreElements(); ) {
                TreeNode n = (TreeNode)e.nextElement();
                TreePath path = parent.pathByAddingChild(n);
                this.expandAll(path, expand);
            }
        }

        if (expand) {
            this.indexTree.expandPath(parent);
        } else {
            this.indexTree.collapsePath(parent);
        }
    }

    private synchronized void fireUpdateSelectionAndShowPage(URL url) {
        DefaultMutableTreeNode node = this.indexData.getNodeByUrl(url);
        if (node==null) {
            this.indexTree.clearSelection();
            this.previousButton.setEnabled(false);
            this.nextButton.setEnabled(false);
            this.updatePageAndButtons(url);
        } else {
            this.indexTree.setSelectionPath(new TreePath(node.getPath()));
        }
    }

    private synchronized void updatePageAndButtons(URL url) {
        try {
            if (url!=null && url.equals(pageTextPane.getPage())) {
                return;
            }
            for (String pre : this.extBrowserPrefixes) {
                if (url.toString().startsWith(pre)) {
                    try {
                        BareBonesBrowserLaunch.openURL(url.toString());
                    } catch (RuntimeException ex) {
                        JOptionPane.showMessageDialog(this, ex.getMessage(),
                            "Could not launch external browser",
                            JOptionPane.ERROR_MESSAGE);
                    }
                    return;
                }
            }
            pageTextPane.setPage(url);
            if (this.history.size()>this.curHistoryIdx+1) {
                if (!this.history.get(this.curHistoryIdx+1).equals(url)) {
                    while (curHistoryIdx+1<history.size()) {
                        history.removeLast();
                    }
                    history.add(url);

                }
            } else {
                history.add(url);
            }
            if (history.size()>maxHistorySize) {
                history.removeFirst();
            } else {
                ++curHistoryIdx;
            }
            this.backButton.setEnabled(this.canGoBack());
            this.forwardButton.setEnabled(this.canGoForward());
            this.previousButton.setEnabled(this.hasPrevious(this.indexTree.getSelectionPath()));
            this.nextButton.setEnabled(this.hasNext(this.indexTree.getSelectionPath()));
        } catch (IOException ex) {
            ex.printStackTrace();
        }
    }


    /** This method is called from within the constructor to
     * initialize the form.
     * WARNING: Do NOT modify this code. The content of this method is
     * always regenerated by the Form Editor.
     */
    @SuppressWarnings("unchecked")
    // <editor-fold defaultstate="collapsed" desc="Generated Code">//GEN-BEGIN:initComponents
    private void initComponents() {

        toolBar = new javax.swing.JToolBar();
        backButton = new javax.swing.JButton();
        forwardButton = new javax.swing.JButton();
        jSeparator1 = new javax.swing.JToolBar.Separator();
        previousButton = new javax.swing.JButton();
        nextButton = new javax.swing.JButton();
        splitPane = new javax.swing.JSplitPane();
        jScrollPaneForTree = new javax.swing.JScrollPane();
        indexTree = new javax.swing.JTree();
        jScrollPaneForPage = new javax.swing.JScrollPane();
        pageTextPane = new javax.swing.JTextPane();
        closeButton = new javax.swing.JButton();

        setTitle("Documentation Browser");

        toolBar.setFloatable(false);
        toolBar.setRollover(true);

        backButton.setText("Back");
        backButton.setToolTipText("Back");
        backButton.setFocusable(false);
        backButton.setVerticalTextPosition(javax.swing.SwingConstants.BOTTOM);
        backButton.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                backButtonActionPerformed(evt);
            }
        });
        toolBar.add(backButton);

        forwardButton.setText("Forward");
        forwardButton.setToolTipText("Forward");
        forwardButton.setFocusable(false);
        forwardButton.setVerticalTextPosition(javax.swing.SwingConstants.BOTTOM);
        forwardButton.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                forwardButtonActionPerformed(evt);
            }
        });
        toolBar.add(forwardButton);
        toolBar.add(jSeparator1);

        previousButton.setText("Previous Topic");
        previousButton.setToolTipText("Previous Topic");
        previousButton.setFocusable(false);
        previousButton.setVerticalTextPosition(javax.swing.SwingConstants.BOTTOM);
        previousButton.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                previousButtonActionPerformed(evt);
            }
        });
        toolBar.add(previousButton);

        nextButton.setText("Next Topic");
        nextButton.setToolTipText("Next Topic");
        nextButton.setFocusable(false);
        nextButton.setVerticalTextPosition(javax.swing.SwingConstants.BOTTOM);
        nextButton.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                nextButtonActionPerformed(evt);
            }
        });
        toolBar.add(nextButton);

        splitPane.setDividerSize(4);
        splitPane.setResizeWeight(0.2);
        splitPane.setAutoscrolls(true);

        javax.swing.tree.DefaultMutableTreeNode treeNode1 = new javax.swing.tree.DefaultMutableTreeNode("root");
        indexTree.setModel(new javax.swing.tree.DefaultTreeModel(treeNode1));
        indexTree.setRootVisible(false);
        indexTree.setSelectionModel(null);
        indexTree.setShowsRootHandles(true);
        jScrollPaneForTree.setViewportView(indexTree);

        splitPane.setLeftComponent(jScrollPaneForTree);

        jScrollPaneForPage.setViewportView(pageTextPane);

        splitPane.setRightComponent(jScrollPaneForPage);

        closeButton.setText("Close");
        closeButton.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                closeButtonActionPerformed(evt);
            }
        });

        javax.swing.GroupLayout layout = new javax.swing.GroupLayout(getContentPane());
        getContentPane().setLayout(layout);
        layout.setHorizontalGroup(
            layout.createParallelGroup(javax.swing.GroupLayout.Alignment.LEADING)
            .addGroup(javax.swing.GroupLayout.Alignment.TRAILING, layout.createSequentialGroup()
                .addContainerGap()
                .addComponent(closeButton))
            .addComponent(toolBar, javax.swing.GroupLayout.DEFAULT_SIZE, 643, Short.MAX_VALUE)
            .addComponent(splitPane, javax.swing.GroupLayout.DEFAULT_SIZE, 643, Short.MAX_VALUE)
        );
        layout.setVerticalGroup(
            layout.createParallelGroup(javax.swing.GroupLayout.Alignment.LEADING)
            .addGroup(layout.createSequentialGroup()
                .addComponent(toolBar, javax.swing.GroupLayout.PREFERRED_SIZE, 35, javax.swing.GroupLayout.PREFERRED_SIZE)
                .addPreferredGap(javax.swing.LayoutStyle.ComponentPlacement.RELATED)
                .addComponent(splitPane, javax.swing.GroupLayout.DEFAULT_SIZE, 352, Short.MAX_VALUE)
                .addPreferredGap(javax.swing.LayoutStyle.ComponentPlacement.RELATED)
                .addComponent(closeButton))
        );

        pack();
    }// </editor-fold>//GEN-END:initComponents

    private void closeButtonActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_closeButtonActionPerformed
        switch (this.getDefaultCloseOperation()) {
            case SimpleDocBrowser.HIDE_ON_CLOSE:
                this.setVisible(false);
                break;
            case SimpleDocBrowser.DISPOSE_ON_CLOSE:
                this.setVisible(false);
                this.dispose();
                break;
            case SimpleDocBrowser.EXIT_ON_CLOSE:
                System.exit(0);
                break;
            case SimpleDocBrowser.DO_NOTHING_ON_CLOSE:
                break;
            default:
                assert false;
        }
    }//GEN-LAST:event_closeButtonActionPerformed

    private void backButtonActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_backButtonActionPerformed
        this.goBack();
    }//GEN-LAST:event_backButtonActionPerformed

    private void forwardButtonActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_forwardButtonActionPerformed
        this.goForward();
    }//GEN-LAST:event_forwardButtonActionPerformed

    private void previousButtonActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_previousButtonActionPerformed
        if (this.indexTree.getSelectionCount()==0) {
            return;
        }
        DefaultMutableTreeNode node = (DefaultMutableTreeNode)this.indexTree.getSelectionPath().getLastPathComponent();
        if (node.getParent()==null) {
            return;
        }
        if (node.getPreviousSibling()==null) {
            this.indexTree.setSelectionPath(this.indexTree.getSelectionPath().getParentPath());
        } else {
            this.indexTree.setSelectionPath(new TreePath(node.getPreviousSibling().getLastLeaf().getPath()));
        }

    }//GEN-LAST:event_previousButtonActionPerformed

    private void nextButtonActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_nextButtonActionPerformed
        if (this.indexTree.getSelectionCount()==0) {
            return;
        }
        DefaultMutableTreeNode node = (DefaultMutableTreeNode)this.indexTree.getSelectionPath().getLastPathComponent();
        if (node.isLeaf()==false) {
            this.indexTree.setSelectionPath(new TreePath(((DefaultMutableTreeNode)node.getFirstChild()).getPath()));
        } else while (true) {
            if (node.getNextSibling()==null) {
                node = (DefaultMutableTreeNode)node.getParent();
                if (node==null) {
                    return;
                }
            } else {
                this.indexTree.setSelectionPath(new TreePath(node.getNextSibling().getPath()));
                return;
            }
        }

    }//GEN-LAST:event_nextButtonActionPerformed

    private boolean hasPrevious(TreePath path) {
        if (path==null) {
            return false;
        }
        DefaultMutableTreeNode node = (DefaultMutableTreeNode)path.getLastPathComponent();
        DefaultMutableTreeNode parent = (DefaultMutableTreeNode)node.getParent();
        if (node.getParent()==null) {
            return false;
        }
        if (parent.equals(this.indexTree.getModel().getRoot()) && parent.getIndex(node)==0) {
            return false;
        }
        return true;

    }

    private boolean hasNext(TreePath path) {
        if (path==null) {
            return false;
        }
        DefaultMutableTreeNode node = (DefaultMutableTreeNode)path.getLastPathComponent();
        if (node.isLeaf()==false) {
            return true;
        }
        DefaultMutableTreeNode parent = (DefaultMutableTreeNode)node.getParent();
        int i = node.getParent().getIndex(node);
        while (true) {
            if (parent.getChildCount()>i+1) {
                return true;
            } else {
                if (parent.getParent()==null) {
                    return false;
                } else {
                    i = parent.getParent().getIndex(parent);
                    parent = (DefaultMutableTreeNode)parent.getParent();
                }             
            }
        }
    }


    // Variables declaration - do not modify//GEN-BEGIN:variables
    private javax.swing.JButton backButton;
    private javax.swing.JButton closeButton;
    private javax.swing.JButton forwardButton;
    private javax.swing.JTree indexTree;
    private javax.swing.JScrollPane jScrollPaneForPage;
    private javax.swing.JScrollPane jScrollPaneForTree;
    private javax.swing.JToolBar.Separator jSeparator1;
    private javax.swing.JButton nextButton;
    private javax.swing.JTextPane pageTextPane;
    private javax.swing.JButton previousButton;
    private javax.swing.JSplitPane splitPane;
    private javax.swing.JToolBar toolBar;
    // End of variables declaration//GEN-END:variables

}
