/*
 * Copyright 2010 Lari Natri <lari.natri@gmail.com>
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Lesser General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package net.aavalla.utils.gui.fontchooser;

import java.awt.Font;
import java.awt.GraphicsEnvironment;
import java.lang.reflect.InvocationTargetException;
import javax.swing.DefaultListModel;
import javax.swing.SwingUtilities;


/**
 *
 * @author Lari Natri <lari.natri@gmail.com>
 */
public class FontChooserPanel extends javax.swing.JPanel {

    private final Object lock = new Object();
    private Font chosenFont = null;

    /** Creates new form FontChooserPanel */
    public FontChooserPanel() {
        initComponents();

        // Get all font family names
        GraphicsEnvironment ge = GraphicsEnvironment.getLocalGraphicsEnvironment();
        String fontNames[] = ge.getAvailableFontFamilyNames();
        DefaultListModel listModel = new DefaultListModel();
        listModel.clear();
        for (String fontName : fontNames) {
            listModel.addElement(fontName);
        }
        this.fontNameList.setModel(listModel);

        this.updateUIfromChosenFont();

    }

    public FontChooserPanel(Font initialFont, boolean allowSizeEdit) {
        this();

        this.setAllowSizeEdit(allowSizeEdit);
        this.chosenFont = initialFont;
        this.updateUIfromChosenFont();

    }

    public void setChosenFont(Font f) {
        synchronized(lock) {
            this.chosenFont = f;
        }
        this.updateUIfromChosenFont();
    }

    public final void setAllowSizeEdit(final boolean allowSizeEdit) {
        Runnable r = new Runnable() {
            public void run() {
                synchronized(lock) {
                    FontChooserPanel.this.sizeLabel.setVisible(allowSizeEdit);
                    FontChooserPanel.this.sizeSpinner.setVisible(allowSizeEdit);
                }
            }
        };
        runOnEDTNoBlock(r);
    }

    public boolean isAllowSizeEdit() {
        synchronized (lock) {
            return this.sizeSpinner.isVisible();
        }
    }

    public Font getChosenFont() {
        synchronized(lock) {
            return this.chosenFont;
        }
    }

    private void updateUIfromChosenFont() {
        Runnable r = new Runnable() {
            public void run() {
                synchronized(lock) {
                    if (chosenFont==null) {
                        styleList.setSelectedIndex(0);
                        styleField.setText(styleList.getSelectedValue().toString());
                        fontNameList.setSelectedIndex(-1);
                        fontNameField.setText("");
                        return;
                    }
                    fontNameList.setSelectedValue(chosenFont.getFamily(),true);
                    fontNameField.setText(fontNameList.getSelectedValue().toString());
                    sizeSpinner.getModel().setValue(chosenFont.getSize());
                    if (chosenFont.isBold() && chosenFont.isItalic()) {
                        styleList.setSelectedIndex(3);
                    } else if (chosenFont.isItalic()) {
                        styleList.setSelectedIndex(2);
                    } else if (chosenFont.isBold()) {
                        styleList.setSelectedIndex(1);
                    } else {
                        styleList.setSelectedIndex(0);
                    }
                    styleField.setText(styleList.getSelectedValue().toString());
                    previewPane.setFont(chosenFont);
                }
            }
        };
        runOnEDTNoBlock(r);
    }

    private void updateChosenFontFromUI() {
        synchronized(lock) {
            if (fontNameList.getSelectedIndex() < 0) {
                this.chosenFont = null;
            } else {
                int style = 0;
                switch(this.styleList.getSelectedIndex()) {
                    case 3:
                        style = Font.BOLD | Font.ITALIC;
                        break;
                    case 2:
                        style = Font.ITALIC;
                        break;
                    case 1:
                        style = Font.BOLD;
                        break;
                    case 0:
                    default:
                        style = Font.PLAIN;
                        break;
                }
                this.chosenFont = new Font(this.fontNameList.getSelectedValue().toString(),style,(Integer)this.sizeSpinner.getModel().getValue());
            }
        }
        this.updateUIfromChosenFont();
    }

    private static void runOnEDTNoBlock(Runnable r) {
        try {
            runOnEDT(r,false);
        } catch (InvocationTargetException ex) {
            // Doesn't happen
        } catch (InterruptedException ex) {
            // Doesn't happen
        }
    }

    private static void runOnEDT(Runnable r, boolean block) throws InvocationTargetException, InterruptedException {
        boolean isEDT = SwingUtilities.isEventDispatchThread();
        if (isEDT && block) {
            r.run();
        } else if (!isEDT && block) {
            SwingUtilities.invokeAndWait(r);
        } else {
            SwingUtilities.invokeLater(r);
        }

    }

    /** This method is called from within the constructor to
     * initialize the form.
     * WARNING: Do NOT modify this code. The content of this method is
     * always regenerated by the Form Editor.
     */
    @SuppressWarnings("unchecked")
    // <editor-fold defaultstate="collapsed" desc="Generated Code">//GEN-BEGIN:initComponents
    private void initComponents() {

        fontNameLabel = new javax.swing.JLabel();
        styleLabel = new javax.swing.JLabel();
        sizeLabel = new javax.swing.JLabel();
        fontNameField = new javax.swing.JTextField();
        styleField = new javax.swing.JTextField();
        sizeSpinner = new javax.swing.JSpinner();
        jScrollPane1 = new javax.swing.JScrollPane();
        fontNameList = new javax.swing.JList();
        jScrollPane2 = new javax.swing.JScrollPane();
        styleList = new javax.swing.JList();
        jScrollPane3 = new javax.swing.JScrollPane();
        previewPane = new javax.swing.JTextPane();

        fontNameLabel.setLabelFor(fontNameField);
        fontNameLabel.setText("Font:");

        styleLabel.setLabelFor(styleField);
        styleLabel.setText("Style:");

        sizeLabel.setLabelFor(sizeSpinner);
        sizeLabel.setText("Size:");

        fontNameField.setEditable(false);

        styleField.setEditable(false);

        sizeSpinner.setModel(new javax.swing.SpinnerNumberModel(12, 6, 72, 1));
        sizeSpinner.addChangeListener(new javax.swing.event.ChangeListener() {
            public void stateChanged(javax.swing.event.ChangeEvent evt) {
                sizeSpinnerStateChanged(evt);
            }
        });

        fontNameList.setModel(new javax.swing.AbstractListModel() {
            String[] strings = { "Item 1", "Item 2", "Item 3", "Item 4", "Item 5" };
            public int getSize() { return strings.length; }
            public Object getElementAt(int i) { return strings[i]; }
        });
        fontNameList.setSelectionMode(javax.swing.ListSelectionModel.SINGLE_SELECTION);
        fontNameList.addListSelectionListener(new javax.swing.event.ListSelectionListener() {
            public void valueChanged(javax.swing.event.ListSelectionEvent evt) {
                fontNameListValueChanged(evt);
            }
        });
        jScrollPane1.setViewportView(fontNameList);

        styleList.setModel(new javax.swing.AbstractListModel() {
            String[] strings = { "Plain", "Bold", "Italic", "Bold + Italic" };
            public int getSize() { return strings.length; }
            public Object getElementAt(int i) { return strings[i]; }
        });
        styleList.setSelectionMode(javax.swing.ListSelectionModel.SINGLE_SELECTION);
        styleList.addListSelectionListener(new javax.swing.event.ListSelectionListener() {
            public void valueChanged(javax.swing.event.ListSelectionEvent evt) {
                styleListValueChanged(evt);
            }
        });
        jScrollPane2.setViewportView(styleList);

        previewPane.setEditable(false);
        previewPane.setText("Preview. AaBbCcDdEeFf 0123456789");
        previewPane.setToolTipText("Preview of the chosen font is displayed here");
        jScrollPane3.setViewportView(previewPane);

        javax.swing.GroupLayout layout = new javax.swing.GroupLayout(this);
        this.setLayout(layout);
        layout.setHorizontalGroup(
            layout.createParallelGroup(javax.swing.GroupLayout.Alignment.LEADING)
            .addGroup(layout.createSequentialGroup()
                .addGroup(layout.createParallelGroup(javax.swing.GroupLayout.Alignment.LEADING)
                    .addComponent(fontNameLabel, javax.swing.GroupLayout.PREFERRED_SIZE, 188, javax.swing.GroupLayout.PREFERRED_SIZE)
                    .addComponent(jScrollPane1, javax.swing.GroupLayout.DEFAULT_SIZE, 194, Short.MAX_VALUE)
                    .addComponent(fontNameField, javax.swing.GroupLayout.DEFAULT_SIZE, 194, Short.MAX_VALUE))
                .addPreferredGap(javax.swing.LayoutStyle.ComponentPlacement.RELATED)
                .addGroup(layout.createParallelGroup(javax.swing.GroupLayout.Alignment.LEADING)
                    .addComponent(jScrollPane2, javax.swing.GroupLayout.DEFAULT_SIZE, 113, Short.MAX_VALUE)
                    .addComponent(styleField, javax.swing.GroupLayout.DEFAULT_SIZE, 113, Short.MAX_VALUE)
                    .addComponent(styleLabel, javax.swing.GroupLayout.DEFAULT_SIZE, 113, Short.MAX_VALUE))
                .addPreferredGap(javax.swing.LayoutStyle.ComponentPlacement.RELATED)
                .addGroup(layout.createParallelGroup(javax.swing.GroupLayout.Alignment.LEADING, false)
                    .addComponent(sizeLabel, javax.swing.GroupLayout.DEFAULT_SIZE, 87, Short.MAX_VALUE)
                    .addComponent(sizeSpinner)))
            .addComponent(jScrollPane3, javax.swing.GroupLayout.Alignment.TRAILING, javax.swing.GroupLayout.DEFAULT_SIZE, 418, Short.MAX_VALUE)
        );
        layout.setVerticalGroup(
            layout.createParallelGroup(javax.swing.GroupLayout.Alignment.LEADING)
            .addGroup(layout.createSequentialGroup()
                .addGroup(layout.createParallelGroup(javax.swing.GroupLayout.Alignment.BASELINE)
                    .addComponent(fontNameLabel)
                    .addComponent(sizeLabel)
                    .addComponent(styleLabel))
                .addPreferredGap(javax.swing.LayoutStyle.ComponentPlacement.RELATED)
                .addGroup(layout.createParallelGroup(javax.swing.GroupLayout.Alignment.BASELINE)
                    .addComponent(fontNameField, javax.swing.GroupLayout.PREFERRED_SIZE, javax.swing.GroupLayout.DEFAULT_SIZE, javax.swing.GroupLayout.PREFERRED_SIZE)
                    .addComponent(styleField, javax.swing.GroupLayout.PREFERRED_SIZE, javax.swing.GroupLayout.DEFAULT_SIZE, javax.swing.GroupLayout.PREFERRED_SIZE)
                    .addComponent(sizeSpinner, javax.swing.GroupLayout.PREFERRED_SIZE, javax.swing.GroupLayout.DEFAULT_SIZE, javax.swing.GroupLayout.PREFERRED_SIZE))
                .addPreferredGap(javax.swing.LayoutStyle.ComponentPlacement.RELATED)
                .addGroup(layout.createParallelGroup(javax.swing.GroupLayout.Alignment.LEADING)
                    .addComponent(jScrollPane2, javax.swing.GroupLayout.DEFAULT_SIZE, 164, Short.MAX_VALUE)
                    .addComponent(jScrollPane1, javax.swing.GroupLayout.DEFAULT_SIZE, 164, Short.MAX_VALUE))
                .addPreferredGap(javax.swing.LayoutStyle.ComponentPlacement.RELATED)
                .addComponent(jScrollPane3, javax.swing.GroupLayout.DEFAULT_SIZE, 83, Short.MAX_VALUE))
        );
    }// </editor-fold>//GEN-END:initComponents

    private void fontNameListValueChanged(javax.swing.event.ListSelectionEvent evt) {//GEN-FIRST:event_fontNameListValueChanged
        if (evt.getValueIsAdjusting() == false) {
            updateChosenFontFromUI();
        }
    }//GEN-LAST:event_fontNameListValueChanged

    private void styleListValueChanged(javax.swing.event.ListSelectionEvent evt) {//GEN-FIRST:event_styleListValueChanged
        if (evt.getValueIsAdjusting() == false) {
            updateChosenFontFromUI();
        }
    }//GEN-LAST:event_styleListValueChanged

    private void sizeSpinnerStateChanged(javax.swing.event.ChangeEvent evt) {//GEN-FIRST:event_sizeSpinnerStateChanged
        updateChosenFontFromUI();
    }//GEN-LAST:event_sizeSpinnerStateChanged


    // Variables declaration - do not modify//GEN-BEGIN:variables
    private javax.swing.JTextField fontNameField;
    private javax.swing.JLabel fontNameLabel;
    private javax.swing.JList fontNameList;
    private javax.swing.JScrollPane jScrollPane1;
    private javax.swing.JScrollPane jScrollPane2;
    private javax.swing.JScrollPane jScrollPane3;
    private javax.swing.JTextPane previewPane;
    private javax.swing.JLabel sizeLabel;
    private javax.swing.JSpinner sizeSpinner;
    private javax.swing.JTextField styleField;
    private javax.swing.JLabel styleLabel;
    private javax.swing.JList styleList;
    // End of variables declaration//GEN-END:variables

}
