/*
 *  Copyright 2010 Lari Natri <lari.natri@gmail.com>.
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Lesser General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package net.aavalla.utils.gui.twovenn;

import net.aavalla.utils.general.StrUtils;
import java.awt.BasicStroke;
import java.awt.Color;
import java.awt.Font;
import java.awt.FontMetrics;
import java.awt.Graphics2D;
import java.awt.RenderingHints;
import java.awt.image.BufferedImage;

/**
 *
 * @author Lari Natri <lari.natri@gmail.com>
 */
public class VennCreator {

    public static VennDrawData calculateVennDrawData(final VennParameters p, int width, int height) {

        final int margin = p.getMarginImage();
        final int marginBetween = p.getMarginBetween();
        final double aTotal = p.getaTotal();
        final double abIntersect = p.getAbIntersect();
        final double bTotal = p.getbTotal();

        final int x_o = width/2;  // x of the middle point on canvas
        final int y_o = height/2; // y of the middle point on canvas
        int x_1 = x_o;       // x of the center of circle 1
        final int y_1 = y_o; // y of the center of circle 1
        int x_2 = x_o;       // x of the center of circle 2
        final int y_2 = y_o; // y of the center of circle 2

        final int r_big = Math.min((height-2*margin)/2, (width-2*margin-marginBetween)/4); // radius of bigger circle
        final double a_big = Math.PI*r_big*r_big; // area for bigger circle in pixels
        final double a_small = Math.min(aTotal, bTotal)*a_big/Math.max(aTotal, bTotal); // area for smaller circle in pixels
        final int r_small = (int)Math.round(Math.sqrt(a_small/Math.PI)); // radius of smaller circle

        final int r_1; // radius of circle 1
        final int r_2; // radius of circle 2

        // set radiuses and move circles so that edges touch
        if (aTotal >= bTotal) { // set a (circle 1) is bigger or they are same size
            r_1 = r_big;
            r_2 = r_small;
            // moves circle 2 (smaller) to the right so that the edges touch
            x_2 = x_1 + r_1 - r_2;
        } else { // set b (circle 2) is bigger
            r_1 = r_small;
            r_2 = r_big;
            // moves circle 1 (smaller) to the right so that the edges touch
            x_1 = x_2 - r_2 + r_1;
        }

        if (abIntersect <= 0) {
            // Case 1: The sets do not intersect
            //         ==> Separate the two circles

            int neededSeparation = 2*Math.min(r_1,r_2)+marginBetween;
            x_1 -= neededSeparation / 2;
            x_2 += neededSeparation / 2;
            if (neededSeparation % 2 == 1) {
                x_2 += 1;
            }

        } else if (abIntersect<aTotal && abIntersect<bTotal) {
            // Case 2: The sets partly intersect (neither set is totally
            //         contained within the other)
            //         ==> Move circles to right positions

            double optRatio = abIntersect / (aTotal+bTotal);
            double optArea = optRatio * ((a_big)+(a_small)); // optimal area for intersection

            int x_old1 = x_1;
            int x_old2 = x_2;

            x_1 -= 1; // move circle 1 one pixel left
            boolean lastMoved1 = true; // if true, last move was made for circle 1
            double oldAreaDiff = Double.MAX_VALUE; // holds the difference between optimal intersection area and the one obtained in last iteration

            while(x_1!=x_old1 || x_2!=x_old2) { // iterate while solution is not getting worse

		double dx = x_1 - x_2; // circle centers delta in x; always non-zero, because x_1 was moved one pixel left before while loop
		double dy = y_1 - y_2; // circle centers delta in y
		double d = Math.sqrt(dx*dx + dy*dy);  // distance between circle centers; sqrt is safe, because dx*dx>0 && dy*dy>=0
                // Note: circles intersect in two points, iff
                // (d <= r_1+r_2 && d >= Math.abs(r_1-r_2)) (always so here)
                double alpha = 2 * Math.acos((d*d + r_1*r_1 - r_2*r_2) / (2 * r_1 * d));
                double beta  = 2 * Math.acos((d*d + r_2*r_2 - r_1*r_1) / (2 * r_2 * d));
                double area = (0.5 * r_1*r_1 * (alpha - Math.sin(alpha)) +
                               0.5 * r_2*r_2 * (beta  - Math.sin(beta))); // intersection area
                double areaDiff = Math.abs(optArea-area); // difference between optimal and current intersection area

                // System.out.println("DEBUG: Diff: "+areaDiff+"Int area: "+area+", Opt area: "+optArea);

                if (areaDiff==Double.NaN || areaDiff <= oldAreaDiff) {
                    // diff is not getting worse. move circles further apart
                    // (and reiterate)
                    oldAreaDiff = areaDiff;
                    x_old1 = x_1;
                    x_old2 = x_2;
                    if (lastMoved1) {
                        x_2 += 1; // move circle 2 to the right
                    } else {
                        x_1 -= 1; // move circle 1 to the left
                    }
                    lastMoved1 = !lastMoved1;
                } else {
                    // area diff is getting worse: stop here and set old value
                    // as the solution.
                    x_1 = x_old1;
                    x_2 = x_old2;
                }

            } // while

        } else {
            // Case 3: One set is totally contained within the other
            //         ==> Nothing further to do
        }

        int aArCenterX  = ((x_1-r_1) + (x_2-r_2)) / 2;
        int aArCenterY  = y_1;
        int bArCenterX  = ((x_1+r_1) + (x_2+r_2)) / 2;
        int bArCenterY  = y_2;
        int abArCenterX = ((x_1+r_1) + (x_2-r_2)) / 2;
        int abArCenterY = (y_1+y_2) /2;

        return new VennDrawData(width, height,
                                x_1, y_1, r_1,
                                x_2, y_2, r_2,
                                aArCenterX, aArCenterY,
                                bArCenterX, bArCenterY,
                                abArCenterX, abArCenterY);
    }

    public static BufferedImage createVennImage(VennParameters p, VennDrawData dData, boolean isArgbImage) {
        int imageType = isArgbImage ? BufferedImage.TYPE_INT_ARGB : BufferedImage.TYPE_INT_RGB;
        BufferedImage image = new BufferedImage(dData.canvasSizeX, dData.canvasSizeY, imageType);

        Graphics2D ig = image.createGraphics();

        Color bgColor = p.getCanvasColorBG();
        if (isArgbImage) {
            bgColor = new Color(p.getCanvasColorBG().getRed(), p.getCanvasColorBG().getGreen(), p.getCanvasColorBG().getBlue(),0);
        }

        ig.setBackground(bgColor);

        // set background
        ig.setColor(bgColor);
        ig.fillRect(0, 0, image.getWidth(), image.getHeight());

        String labelA = p.getaLabel().trim();
        String labelB = p.getbLabel().trim();
        String labelAB = p.getAbLabel().trim();

        double intersect = p.getAbIntersect();
        double aUnique = p.getaTotal()-intersect;
        double bUnique = p.getbTotal()-intersect;
        double total = aUnique+bUnique+intersect;
        double aUniquePercent = 100*(aUnique/total);
        double bUniquePercent = 100*(bUnique/total);
        double intersectPercent = 100*(intersect/total);
        String aUniqueStr = StrUtils.asString(aUnique);
        String bUniqueStr = StrUtils.asString(bUnique);
        String isectStr = StrUtils.asString(intersect);
        String aUniquePercStr = StrUtils.asString(aUniquePercent,0,3);
        String bUniquePercStr = StrUtils.asString(bUniquePercent,0,3);
        String isectPercStr = StrUtils.asString(intersectPercent,0,3);

        switch (p.getEmbeddedLabelType()) {
            case PREFIXONLY:
                labelA  = " "+labelA+" ";  //NOI18N
                labelB  = " "+labelB+" ";  //NOI18N
                labelAB = " "+labelAB+" "; //NOI18N
                break;
            case PREFIXANDNUMBER:
                labelA = " "+labelA+(labelA.length()>0 ? ": " : "")+aUniqueStr+" ";
                labelB = " "+labelB+(labelB.length()>0 ? ": " : "")+bUniqueStr+" ";
                labelAB = " "+labelAB+(labelAB.length()>0 ? ": " : "")+isectStr+" ";
                break;
            case PREFIXANDPERCENTAGE:
                labelA  = " "+labelA+(labelA.length()>0 ? ": " : "")+aUniquePercStr+" % ";  //NOI18N
                labelB  = " "+labelB+(labelB.length()>0 ? ": " : "")+bUniquePercStr+" % ";  //NOI18N
                labelAB = " "+labelAB+(labelAB.length()>0 ? ": " : "")+isectPercStr+" % ";  //NOI18N
                break;
            case PREFIXANDNUMBERPERCENTAGE:
                labelA = " "+labelA+(labelA.length()>0 ? ": " : "")+aUniqueStr+" ("+aUniquePercStr+" %) ";
                labelB = " "+labelB+(labelB.length()>0 ? ": " : "")+bUniqueStr+" ("+bUniquePercStr+" %) ";
                labelAB = " "+labelAB+(labelAB.length()>0 ? ": " : "")+isectStr+" ("+isectPercStr+" %) ";
                break;
            case NONE:
            default:
                labelA=""; //NOI18N
                labelB=""; //NOI18N
                labelAB=""; //NOI18N
                break;
        }


        // do the actual drawing

        if (p.isAntialiase()) {
            ig.setRenderingHint(RenderingHints.KEY_ANTIALIASING, RenderingHints.VALUE_ANTIALIAS_ON);
        } else {
            ig.setRenderingHint(RenderingHints.KEY_ANTIALIASING, RenderingHints.VALUE_ANTIALIAS_OFF);
        }

        ig.setStroke(new BasicStroke(p.getLineThickness()));

        ig.setColor(p.getaCircleOutlineColor());
        ig.drawOval(dData.aCenterX-dData.aRadius, dData.aCenterY-dData.aRadius,dData.aRadius*2,dData.aRadius*2);
        ig.setColor(p.getbCircleOutlineColor());
        ig.drawOval(dData.bCenterX-dData.bRadius, dData.bCenterY-dData.bRadius,dData.bRadius*2,dData.bRadius*2);

        Font f = p.getLabelFont();

        FontMetrics fm;
        int labStartA;
        int labEndA;
        int labStartAB;
        int labEndAB;
        int labStartB;
        int labEndB;
        int labAY;
        int labABY;
        int labBY;

        if (f.getSize()<p.getMinFontSize()) {
            f = new Font(f.getFontName(),f.getStyle(),p.getMinFontSize());
        } else if (f.getSize()>p.getMaxFontSize()) {
            f = new Font(f.getFontName(),f.getStyle(),p.getMaxFontSize());
        }

        boolean reiterate;
        do { // calculate font size and label locations
            reiterate = false;
            fm = ig.getFontMetrics(f);
            labStartA  = dData.aArCenterX-(fm.stringWidth(labelA)/2);
            labEndA    = dData.aArCenterX+(fm.stringWidth(labelA)/2);
            labStartAB = dData.abArCenterX-(fm.stringWidth(labelAB)/2);
            labEndAB   = dData.abArCenterX+(fm.stringWidth(labelAB)/2);
            labStartB  = dData.bArCenterX-(fm.stringWidth(labelB)/2);
            labEndB    = dData.bArCenterX+(fm.stringWidth(labelB)/2);
            labAY      = dData.aArCenterY  + (fm.getAscent()+fm.getDescent())/2;
            labABY     = dData.abArCenterY + (fm.getAscent()+fm.getDescent())/2;
            labBY      = dData.bArCenterY  + (fm.getAscent()+fm.getDescent())/2;
            int fSize = f.getSize();
            if (fSize > p.getMinFontSize() && (labStartA < 0 || labEndB > dData.canvasSizeX)) {
                f = new Font(f.getFontName(),f.getStyle(),fSize-1);
                reiterate = true;
            } else if (fSize > p.getMinFontSize() && labStartA < dData.aCenterX - dData.aRadius + 5
                                 && labEndB > dData.bCenterX+dData.bRadius-5
                                 /*&& labStartAB < dData.bCenterX-dData.bRadius+2*/) {
                f = new Font(f.getFontName(),f.getStyle(),fSize-1);
                reiterate = true;
            } else if (fSize < p.getMaxFontSize() && labStartA > dData.aCenterX - dData.aRadius + 5
                                  && labEndB < dData.bCenterX+dData.bRadius-5
                                  /*&& labStartAB > dData.bCenterX-dData.bRadius+2*/) {
                f = new Font(f.getFontName(),f.getStyle(),fSize+1);
                reiterate = true;
            } else {
                boolean abyFixed = false;
                int h = fm.getAscent()+fm.getDescent()+fm.getLeading();
                if (labEndA>labStartAB || labEndAB>labStartB) {
                    labAY-=h/2;
                    labBY-=h/2;
                    labABY+=h/2;
                    abyFixed = true;
                }
                if (labEndA>labStartB) {
                    if (abyFixed) {
                        labBY-=h;
                    } else {
                        labAY-=h/2;
                        labBY+=h/2;
                    }
                }
            } // else

        } while(reiterate);

        ig.setFont(f);
        ig.setColor(p.getLabelAColor());
        ig.drawString(labelA, labStartA, labAY);
        ig.setColor(p.getLabelABColor());
        ig.drawString(labelAB, labStartAB, labABY);
        ig.setColor(p.getLabelBColor());
        ig.drawString(labelB, labStartB, labBY);

        // update the font size (if changed) in parameters
        if (p.getLabelFont().getSize()!=f.getSize()) {
            p.setLabelFontSize(f.getSize());
        }

        return image;
    }


}
