/*
 *  Copyright 2010 Lari Natri <lari.natri@gmail.com>.
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Lesser General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package net.aavalla.utils.gui.twovenn;

import net.aavalla.utils.general.Disposeable;
import net.aavalla.utils.gui.twovenn.VennExportParameters.ImageType;
import net.aavalla.utils.jobpool.JobFinishedListener;
import net.aavalla.utils.jobpool.JobPool;
import net.aavalla.utils.jobpool.JobPriority;
import net.aavalla.utils.jobpool.JobResultEvent;
import net.aavalla.utils.jobpool.ParameterizedJob;
import net.aavalla.utils.jobpool.parameters.gui.ParameterDialog;
import java.awt.image.BufferedImage;
import java.io.IOException;
import java.net.URL;
import javax.imageio.ImageIO;
import javax.swing.ImageIcon;
import javax.swing.JOptionPane;
import javax.swing.SwingUtilities;

/**
 *
 * @author Lari Natri <lari.natri@gmail.com>
 */
public class VennDiagram extends javax.swing.JPanel implements Disposeable, JobFinishedListener {

    private final boolean useIcons = true;
    private JobPool pool = null;
    private ParameterDialog pDialog = null;

    /** Creates new form VennDiagram */
    public VennDiagram() {
        initComponents();

        if (useIcons) { // load icons
            SwingUtilities.invokeLater(new Runnable() {
                @Override public void run() {
                    URL iUrl = getClass().getResource("resources/configure.png"); //NOI18N
                    if (iUrl != null) {
                        settingsButton.setIcon(new ImageIcon(iUrl));
                    }
                    iUrl = getClass().getResource("resources/document-export.png"); //NOI18N
                    if (iUrl != null) {
                        exportButton.setIcon(new ImageIcon(iUrl));
                    }

                }
            });
        }
    }

    public VennDiagram(JobPool pool) {
        this();
        this.pool = pool;
    }

    public VennDiagram(VennParameters params) {
        this();
        this.vennPanel.setParameters(params);
    }

    public VennDiagram(VennParameters params, JobPool pool) {
        this(params);
        this.pool = pool;
    }

    public VennDiagram(JobPool pool, boolean openSettingsDialog) {
        this(pool);
        if (openSettingsDialog) {
            SwingUtilities.invokeLater(new Runnable() {
                public void run() {
                    VennDiagram.this.settingsButtonActionPerformed(null);
                }
            });
        }
    }


    public void dispose() {
        this.pDialog = null;
        this.pool = null;
        this.vennPanel.dispose();
    }

    public void jobFinished(final JobResultEvent event) {
        if (event.cancelled()) {
            return;
        }
        if (event.failed()) {
            Throwable error = (Throwable) event.getFailReason();
            String message = error.getMessage();
            JOptionPane.showMessageDialog(this,
                event.getSource().getName() + " failed with error: \n" + message,
                "A Job has failed",
                 JOptionPane.ERROR_MESSAGE);
        } else {
            Object result = event.getResult();
            if (result instanceof String) {
                JOptionPane.showMessageDialog(this,
                    result.toString(), event.getSource().getName()+" has finished",
                    JOptionPane.INFORMATION_MESSAGE);
            }
        }
    }





    /** This method is called from within the constructor to
     * initialize the form.
     * WARNING: Do NOT modify this code. The content of this method is
     * always regenerated by the Form Editor.
     */
    @SuppressWarnings("unchecked")
    // <editor-fold defaultstate="collapsed" desc="Generated Code">//GEN-BEGIN:initComponents
    private void initComponents() {

        vennPanel = new net.aavalla.utils.gui.twovenn.VennPanel();
        jToolBar1 = new javax.swing.JToolBar();
        settingsButton = new javax.swing.JButton();
        exportButton = new javax.swing.JButton();

        javax.swing.GroupLayout vennPanelLayout = new javax.swing.GroupLayout(vennPanel);
        vennPanel.setLayout(vennPanelLayout);
        vennPanelLayout.setHorizontalGroup(
            vennPanelLayout.createParallelGroup(javax.swing.GroupLayout.Alignment.LEADING)
            .addGap(0, 400, Short.MAX_VALUE)
        );
        vennPanelLayout.setVerticalGroup(
            vennPanelLayout.createParallelGroup(javax.swing.GroupLayout.Alignment.LEADING)
            .addGap(0, 295, Short.MAX_VALUE)
        );

        jToolBar1.setFloatable(false);
        jToolBar1.setRollover(true);

        settingsButton.setText("Settings");
        settingsButton.setToolTipText("Configure Venn diagram");
        settingsButton.setFocusable(false);
        settingsButton.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                settingsButtonActionPerformed(evt);
            }
        });
        jToolBar1.add(settingsButton);

        exportButton.setText("Export image");
        exportButton.setToolTipText("Export diagram to an image file");
        exportButton.setFocusable(false);
        exportButton.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                exportButtonActionPerformed(evt);
            }
        });
        jToolBar1.add(exportButton);

        javax.swing.GroupLayout layout = new javax.swing.GroupLayout(this);
        this.setLayout(layout);
        layout.setHorizontalGroup(
            layout.createParallelGroup(javax.swing.GroupLayout.Alignment.LEADING)
            .addComponent(jToolBar1, javax.swing.GroupLayout.DEFAULT_SIZE, 400, Short.MAX_VALUE)
            .addComponent(vennPanel, javax.swing.GroupLayout.DEFAULT_SIZE, javax.swing.GroupLayout.DEFAULT_SIZE, Short.MAX_VALUE)
        );
        layout.setVerticalGroup(
            layout.createParallelGroup(javax.swing.GroupLayout.Alignment.LEADING)
            .addGroup(javax.swing.GroupLayout.Alignment.TRAILING, layout.createSequentialGroup()
                .addComponent(vennPanel, javax.swing.GroupLayout.DEFAULT_SIZE, javax.swing.GroupLayout.DEFAULT_SIZE, Short.MAX_VALUE)
                .addPreferredGap(javax.swing.LayoutStyle.ComponentPlacement.RELATED)
                .addComponent(jToolBar1, javax.swing.GroupLayout.PREFERRED_SIZE, 25, javax.swing.GroupLayout.PREFERRED_SIZE))
        );
    }// </editor-fold>//GEN-END:initComponents

    private void settingsButtonActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_settingsButtonActionPerformed
        if (this.pool==null) {
            this.pool = new JobPool();
        }
        if (this.pDialog==null || !this.pDialog.isEnabled() || !this.pDialog.isVisible()) {
            this.pDialog = new ParameterDialog(null,false,new UpdateJob(),this.pool);
        }
        this.pDialog.setVisible(true);
    }//GEN-LAST:event_settingsButtonActionPerformed

    private void exportButtonActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_exportButtonActionPerformed

        if (this.vennPanel.getParameters()==null) {
            JOptionPane.showMessageDialog(null,
                "Venn Diagram image must first be constructed.\n"
                + "Click 'Settings'.",
                "Can not export image",
                 JOptionPane.INFORMATION_MESSAGE);            
            return;
        }
        if (this.pool==null) {
            this.pool = new JobPool();
        }
        ExportVennImageJob j = new ExportVennImageJob();
        j.addJobFinishedListenerSwing(this);
        ParameterDialog dialog = new ParameterDialog(null,false,j,this.pool);
        dialog.setVisible(true);
    }//GEN-LAST:event_exportButtonActionPerformed


    // Variables declaration - do not modify//GEN-BEGIN:variables
    private javax.swing.JButton exportButton;
    private javax.swing.JToolBar jToolBar1;
    private javax.swing.JButton settingsButton;
    private net.aavalla.utils.gui.twovenn.VennPanel vennPanel;
    // End of variables declaration//GEN-END:variables


    public class UpdateJob extends ParameterizedJob<Object,VennParameters> {

        public UpdateJob() {
            super();
            this.setName("Venn Diagram Creator");
            this.setPriority(JobPriority.IMMEDIATE);
            this.setSilent(true);
        }

        @Override
        protected VennParameters getNonfinalEditableParameters() {
            VennParameters p = (VennDiagram.this.vennPanel.getParameters());
            if (p==null) {
                p = new VennParameters();
            } else {
                p = p.clone();
            }
            return p;
        }

        @Override
        protected Object work(VennParameters params) throws Throwable {
            VennDiagram.this.vennPanel.setParameters(params);
            return null;
        }
    }

    public class ExportVennImageJob extends ParameterizedJob<String,VennExportParameters> {

        public ExportVennImageJob() {
            super();
            this.setName("Venn diagram exporter");
            this.setPriority(JobPriority.IMMEDIATE);
            this.setSilent(true);
        }

        @Override
        protected VennExportParameters getNonfinalEditableParameters() {
            VennExportParameters pex = new VennExportParameters();
            VennParameters p = (VennDiagram.this.vennPanel.getParameters());
            pex.setParams(p, VennDiagram.this.vennPanel.getWidth(), VennDiagram.this.vennPanel.getHeight());
            return pex;
        }

        @Override
        protected String work(VennExportParameters p) throws Throwable, IOException {
            System.out.println("DEBUG: Venn exporter started");
            if (p.getVennParameters()==null) {
                throw new Exception("There is no Venn diagram to export.");
            }
            VennParameters vp = p.getVennParameters().clone();

            VennDrawData dData = VennCreator.calculateVennDrawData(vp,p.getX(),p.getY());
            BufferedImage im = VennCreator.createVennImage(vp, dData, p.getType().equals(ImageType.PNG));

            String imageType = p.getType().equals(ImageType.JPG) ? "jpg" : "png";
            ImageIO.write(im, imageType, p.getFile());

            return "Venn diagram image saved.";
        }
    }
}
