/*
 *  Copyright 2010 Lari Natri <lari.natri@gmail.com>.
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Lesser General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package net.aavalla.utils.gui.twovenn;

import net.aavalla.utils.general.Disposeable;
import java.awt.Graphics;
import java.awt.Graphics2D;
import javax.swing.SwingUtilities;

/**
 *
 * @author Lari Natri <lari.natri@gmail.com>
 */
public class VennPanel extends javax.swing.JPanel implements Disposeable {

    private final Object lock = new Object(); // used for synchronization
    private VennDrawData oldDrawData = null;

    private VennParameters params = null;

    /** Creates new form VennPanel */
    public VennPanel() {
        initComponents();
    }

    public VennPanel(VennParameters newParams) {
        this();
        this.setParameters(newParams);
    }

    public void dispose() {
        if (this.params!=null) {
            this.params.removeParameterChangedListener(this);
            this.params = null;
        }
    }

    public final void setParameters(VennParameters newParams) {
        System.out.println("DEBUG: setting Venn params");
        synchronized(this.lock) {
            if (this.params!=null) {
                this.params.removeParameterChangedListener(this);
            }
            this.params = newParams;
            if (this.params!=null) {
                this.params.addParameterChangedListener(this);
            }
        }
        this.parametersUpdated();
    }

    public final void parametersUpdated() {
        synchronized(this.lock) {
            this.oldDrawData = null;
        }
        SwingUtilities.invokeLater(new Runnable() {
            public void run() {
                VennPanel.this.repaint();
            }
        });
    }

    /**
     * Returns the parameters. It is safe to use parameter object's
     * setter methods. If you want to modify parameters by other
     * methods, please clone them first.
     * @return parameters if set, null otherwise
     */
    public final VennParameters getParameters() {
        return this.params;
    }


    @Override
    public void paint(Graphics g) {

        VennDrawData dData = null;

        synchronized(this.lock) {
            dData = this.oldDrawData;
            if (dData==null || dData.canvasSizeX != this.getSize().width || dData.canvasSizeY != this.getSize().height) {
                if (this.params==null) {
                    super.paint(g);
                    return;
                }
                dData = VennCreator.calculateVennDrawData(this.params,this.getWidth(),this.getHeight());
                this.oldDrawData = dData;
            }
        }

        if (dData==null) {
            super.paint(g);
            return;
        }

        Graphics2D g2d = (Graphics2D)g;
        g2d.setBackground(this.params.getCanvasColorBG());
        g2d.setColor(this.params.getCanvasColorBG());
        g2d.fillRect(0,0, this.getSize().width, this.getSize().height);
        g2d.drawImage(VennCreator.createVennImage(this.params, dData, true), 0,0, this);
    }

    


    /** This method is called from within the constructor to
     * initialize the form.
     * WARNING: Do NOT modify this code. The content of this method is
     * always regenerated by the Form Editor.
     */
    @SuppressWarnings("unchecked")
    // <editor-fold defaultstate="collapsed" desc="Generated Code">//GEN-BEGIN:initComponents
    private void initComponents() {

        setBackground(new java.awt.Color(255, 255, 255));
        setForeground(java.awt.Color.black);
        setMinimumSize(new java.awt.Dimension(26, 26));

        javax.swing.GroupLayout layout = new javax.swing.GroupLayout(this);
        this.setLayout(layout);
        layout.setHorizontalGroup(
            layout.createParallelGroup(javax.swing.GroupLayout.Alignment.LEADING)
            .addGap(0, 400, Short.MAX_VALUE)
        );
        layout.setVerticalGroup(
            layout.createParallelGroup(javax.swing.GroupLayout.Alignment.LEADING)
            .addGap(0, 300, Short.MAX_VALUE)
        );
    }// </editor-fold>//GEN-END:initComponents


    // Variables declaration - do not modify//GEN-BEGIN:variables
    // End of variables declaration//GEN-END:variables


}
