/*
 *  Copyright 2010 Lari Natri <lari.natri@gmail.com>.
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Lesser General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package net.aavalla.utils.gui.twovenn;

import net.aavalla.utils.jobpool.parameters.FieldAnnotations;
import net.aavalla.utils.jobpool.parameters.ParameterCollection;
import net.aavalla.utils.jobpool.parameters.validators.DoubleMinMaxValidator;
import net.aavalla.utils.jobpool.parameters.validators.IntegerMinMaxValidator;
import net.aavalla.utils.jobpool.parameters.validators.StringValidator;
import java.awt.Color;
import java.awt.Font;
import java.util.HashSet;

/**
 *
 * @author Lari Natri <lari.natri@gmail.com>
 */
public class VennParameters extends ParameterCollection implements Cloneable {

    public enum LabelType {
        PREFIXONLY,
        PREFIXANDNUMBER,
        PREFIXANDPERCENTAGE,
        PREFIXANDNUMBERPERCENTAGE,
        NONE;

        @Override public String toString() {
            if (this.equals(PREFIXONLY)) {
                return "Prefix only";
            }
            if (this.equals(PREFIXANDNUMBER)) {
                return "Prefix and amount";
            }
            if (this.equals(PREFIXANDPERCENTAGE)) {
                return "Prefix and %";
            }
            if (this.equals(PREFIXANDNUMBERPERCENTAGE)) {
                return "Prefix, amount and %";
            }
            if (this.equals(NONE)) {
                return "No labels";
            }
            return this.name();
        }
    }

    @FieldAnnotations(
        name="A: total size",
        description="Total size of A: A-(A\u2229B)",
        validator=DoubleMinMaxValidator.class,
        validatorSettings={"0.0","1000000000.0"})
    protected double aTotal = 75;
    @FieldAnnotations(
        name="A\u2229B: size",
        description="The size of the intersection A-(A\u2229B)",
        validator=DoubleMinMaxValidator.class,
        validatorSettings={"0.0","1000000000.0"})
    protected double abIntersect = 50;
   @FieldAnnotations(
        name="B: total size",
        description="Total size of B: B-(A\u2229B)",
        validator=DoubleMinMaxValidator.class,
        validatorSettings={"0.0","1000000000.0"})
    protected double bTotal = 100;
    @FieldAnnotations(name="Label A prefix",
        description="Label prefix for circle A",
        validator=StringValidator.class,
        validatorSettings={"0","32",FieldAnnotations.NULL})
    private String aLabel = "A-(A\u2229B)";
    @FieldAnnotations(name="Label A\u2229B prefix",
        description="Label prefix for intersection area A\u2229B",
        validator=StringValidator.class,
        validatorSettings={"0","32",FieldAnnotations.NULL})
    private String abLabel = "A\u2229B";
    @FieldAnnotations(name="Label B prefix",
        description="Label prefix for circle B",
        validator=StringValidator.class,
        validatorSettings={"0","32",FieldAnnotations.NULL})
    private String bLabel = "B-(A\u2229B)";
    @FieldAnnotations(name="Label type",
        description="Label type")
    private LabelType embeddedLabelType = LabelType.PREFIXANDNUMBER;
    private LabelType externalLabelType = LabelType.NONE;
    @FieldAnnotations(name="Font", description="Font used for labels")
    private Font labelFont = new Font("Dialog",Font.PLAIN,16);
    @FieldAnnotations(name="Min font size",
        description="Minimum size of the font",
        validator=IntegerMinMaxValidator.class,
        validatorSettings={"6","72"})
    private int minFontSize = 10;
    @FieldAnnotations(name="Max font size",
        description="Maximum size of the font",
        validator=IntegerMinMaxValidator.class,
        validatorSettings={"6","72"})
    private int maxFontSize = 72;
    @FieldAnnotations(name="Margin",
        description="Minimum size of margin in the edges of the image",
        validator=IntegerMinMaxValidator.class,
        validatorSettings={"1","100"})
    private int marginImage = 5;
    private int marginBetween = 5;
    private int marginLabel = 4;
    @FieldAnnotations(name="Line Thickness",
        description="Thickness of the circles in pixels",
        validator=IntegerMinMaxValidator.class,
        validatorSettings={"1","100"})
    private int lineThickness = 2;
    @FieldAnnotations(name="Background color",
        description="Has effect only when displaying or exporting to JPG.\n"
        + "PNG's will be exported with transparent background.")
    private Color canvasColorBG = Color.WHITE;
    @FieldAnnotations(name="Text Color for Label A",
        description="Color for the text in Label A")
    private Color aLabelColor = Color.BLACK;
    @FieldAnnotations(name="Text Color for Label A\u2229B",
        description="Color for the text in Label A\u2229B")
    private Color abLabelColor = Color.BLACK;
    @FieldAnnotations(name="Text Color for Label B",
        description="Color for the text in Label B")
    private Color bLabelColor = Color.BLACK;
    @FieldAnnotations(name="Circle Outline Color for A",
        description="Color for the circle A's outline")
    private Color aCircleOutlineColor = Color.BLACK;
    @FieldAnnotations(name="Circle Outline Color for B",
        description="Color for the circle B's outline")
    private Color bCircleOutlineColor = Color.BLACK;
    @FieldAnnotations(name="Antialiasing",
        description="Use antialiasing to boost image quality (recommended)")
    private boolean antialiase = true;



    private HashSet<VennPanel> parameterChangedListeners = new HashSet<VennPanel>(1,1);


    public VennParameters() {
        this.setDescription("Parameters for a visual two-way Venn diagram");
    }

    public VennParameters(double aTotal, double bTotal, double abIntersect) {
        if (aTotal<0 || bTotal<0 || abIntersect<0) {
            throw new IllegalArgumentException("Parameter values for Venn Diagram must be >=0");
        }
        if (abIntersect>aTotal || abIntersect>bTotal) {
            throw new IllegalArgumentException("Intersection can not be greater than a nor b");
        }

        this.aTotal = aTotal;
        this.bTotal = bTotal;
        this.abIntersect = abIntersect;
    }


    public synchronized int getMaxFontSize() {
        return maxFontSize;
    }

    public synchronized void setMaxFontSize(int maxFontSize) {
        if (maxFontSize < 6) {
            this.maxFontSize = 6;
        } else if (maxFontSize > 72) {
            this.maxFontSize = 72;
        } else {
            this.maxFontSize = maxFontSize;
        }
        this.parametersChanged();
    }

    public synchronized int getMinFontSize() {
        return minFontSize;
    }

    public synchronized void setMinFontSize(int minFontSize) {
        if (minFontSize > 72) {
            this.minFontSize = 72;
        } else if (minFontSize < 6) {
            this.minFontSize = 6;
        } else {
            this.minFontSize = minFontSize;
        }
        this.parametersChanged();
    }


    public synchronized Color getLabelAColor() {
        return aLabelColor;
    }

    public synchronized void setLabelAColor(Color aColorFG) {
        this.aLabelColor = aColorFG;
        this.parametersChanged();
    }

    public synchronized String getaLabel() {
        return aLabel;
    }

    public synchronized void setaLabel(String aLabel) {
        this.aLabel = aLabel;
        this.parametersChanged();
    }

    public synchronized double getaTotal() {
        return aTotal;
    }

    public synchronized void setaTotal(double aTotal) {
        this.aTotal = aTotal;
        this.parametersChanged();
    }

    public synchronized Color getLabelABColor() {
        return abLabelColor;
    }

    public synchronized void setLabelABColor(Color abColorFG) {
        this.abLabelColor = abColorFG;
        this.parametersChanged();
    }

    public synchronized double getAbIntersect() {
        return abIntersect;
    }

    public synchronized void setAbIntersect(double abIntersect) {
        this.abIntersect = abIntersect;
        this.parametersChanged();
    }

    public synchronized String getAbLabel() {
        return abLabel;
    }

    public synchronized void setAbLabel(String abLabel) {
        this.abLabel = abLabel;
        this.parametersChanged();
    }

    public synchronized boolean isAntialiase() {
        return antialiase;
    }

    public synchronized void setAntialiase(boolean antialiase) {
        this.antialiase = antialiase;
        this.parametersChanged();
    }

    public synchronized Color getLabelBColor() {
        return bLabelColor;
    }

    public synchronized void getLabelBColor(Color bColorFG) {
        this.bLabelColor = bColorFG;
        this.parametersChanged();
    }

    public synchronized Color getaCircleOutlineColor() {
        return aCircleOutlineColor;
    }

    public synchronized void setaCircleOutlineColor(Color aCircleOutlineColor) {
        this.aCircleOutlineColor = aCircleOutlineColor;
        this.parametersChanged();
    }

    public synchronized Color getbCircleOutlineColor() {
        return bCircleOutlineColor;
    }

    public synchronized void setbCircleOutlineColor(Color bCircleOutlineColor) {
        this.bCircleOutlineColor = bCircleOutlineColor;
        this.parametersChanged();
    }

    public synchronized String getbLabel() {
        return bLabel;
    }

    public synchronized void setbLabel(String bLabel) {
        this.bLabel = bLabel;
        this.parametersChanged();
    }

    public synchronized double getbTotal() {
        return bTotal;
    }

    public synchronized void setbTotal(double bTotal) {
        this.bTotal = bTotal;
        this.parametersChanged();
    }

    public synchronized Color getCanvasColorBG() {
        return canvasColorBG;
    }

    public synchronized void setCanvasColorBG(Color canvasColorBG) {
        this.canvasColorBG = canvasColorBG;
        this.parametersChanged();
    }

    public synchronized LabelType getEmbeddedLabelType() {
        return embeddedLabelType;
    }

    public synchronized void setEmbeddedLabelType(LabelType embeddedLabelType) {
        this.embeddedLabelType = embeddedLabelType;
        this.parametersChanged();
    }

    public synchronized LabelType getExternalLabelType() {
        return externalLabelType;
    }

    public synchronized void setExternalLabelType(LabelType externalLabelType) {
        this.externalLabelType = externalLabelType;
        this.parametersChanged();
    }

    public synchronized int getLineThickness() {
        return lineThickness;
    }

    public synchronized void setLineThickness(int lineThickness) {
        this.lineThickness = lineThickness;
        this.parametersChanged();
    }

    public synchronized int getMarginBetween() {
        return marginBetween;
    }

    public synchronized void setMarginBetween(int marginBetween) {
        this.marginBetween = marginBetween;
        this.parametersChanged();
    }

    public synchronized int getMarginImage() {
        return marginImage;
    }

    public synchronized void setMarginImage(int marginImage) {
        this.marginImage = marginImage;
        this.parametersChanged();
    }

    public synchronized int getMarginLabel() {
        return marginLabel;
    }

    public synchronized void setMarginLabel(int marginLabel) {
        this.marginLabel = marginLabel;
        this.parametersChanged();
    }

    public synchronized Font getLabelFont() {
        return labelFont;
    }

    public synchronized void setLabelFont(Font labelFont) {
        this.labelFont = labelFont;

        if (this.labelFont.getSize()<this.getMinFontSize()) {
            this.labelFont = new Font(this.labelFont.getFontName(),this.labelFont.getStyle(),this.getMinFontSize());
        }
        if (this.labelFont.getSize()>this.getMaxFontSize()) {
            this.labelFont = new Font(this.labelFont.getFontName(),this.labelFont.getStyle(),this.getMaxFontSize());
        }
        
        this.parametersChanged();
    }

    /**
     * Sets a new size for the font (replaces it). Does not fire
     * parametersChanged(). To be used only in VennCreator.CreateVennImage
     * @param size new size for the font
     */
    synchronized void setLabelFontSize(int size) {
        this.labelFont = new Font(this.labelFont.getFontName(),this.labelFont.getStyle(),size);

        if (this.labelFont.getSize()<this.getMinFontSize()) {
            this.labelFont = new Font(this.labelFont.getFontName(),this.labelFont.getStyle(),this.getMinFontSize());
        }
        if (this.labelFont.getSize()>this.getMaxFontSize()) {
            this.labelFont = new Font(this.labelFont.getFontName(),this.labelFont.getStyle(),this.getMaxFontSize());
        }
    }




    public synchronized void addParameterChangedListener(VennPanel p) {
        this.parameterChangedListeners.add(p);
    }

    public synchronized boolean removeParameterChangedListener(VennPanel p) {
        return this.parameterChangedListeners.remove(p);
    }

    @Override
    public synchronized String validateFieldRelations() {
        if (this.aTotal < this.abIntersect) {
            return "Intersect can't be greater than A";
        }
        if (this.bTotal < this.abIntersect) {
            return "Intersect can't be greater than B";
        }
        if (this.maxFontSize < this.minFontSize) {
            return "Min font size can't be greater than max font size";
        }
        return null;
    }

    @Override
    public synchronized VennParameters clone() {
        VennParameters p = new VennParameters();
        p.setaTotal(this.getaTotal());
        p.setbTotal(this.getbTotal());
        p.setAbIntersect(this.getAbIntersect());
        copySecondaryParamData(this,p);
        return p;
    }

    private static void copySecondaryParamData(VennParameters pFrom, VennParameters pTo) {
        pTo.setLabelABColor(pFrom.getLabelABColor());
        pTo.setAbLabel(pFrom.getAbLabel());
        pTo.setAntialiase(pFrom.isAntialiase());
        pTo.setCanvasColorBG(pFrom.getCanvasColorBG());
        pTo.setEmbeddedLabelType(pFrom.getEmbeddedLabelType());
        pTo.setExternalLabelType(pFrom.getExternalLabelType());
        pTo.setLineThickness(pFrom.getLineThickness());
        pTo.setMarginBetween(pFrom.getMarginBetween());
        pTo.setMarginImage(pFrom.getMarginImage());
        pTo.setMarginLabel(pFrom.getMarginLabel());
        pTo.setMaxFontSize(pFrom.getMaxFontSize());
        pTo.setMinFontSize(pFrom.getMinFontSize());
        pTo.setLabelAColor(pFrom.getLabelAColor());
        pTo.setaLabel(pFrom.getaLabel());
        pTo.getLabelBColor(pFrom.getLabelBColor());
        pTo.setbLabel(pFrom.getbLabel());
        pTo.setLabelFont(pFrom.getLabelFont());
        pTo.copyHiddenFieldsSetFrom(pFrom);
    }
    
    private synchronized void parametersChanged() {
        for (VennPanel p : this.parameterChangedListeners) {
            p.parametersUpdated();
        }
    }
}
