/*
 * Copyright 2010 Lari Natri <lari.natri@gmail.com>
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Lesser General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package net.aavalla.utils.jobpool;


import net.aavalla.utils.general.Timer;
import java.util.EnumSet;
import java.util.HashMap;
import java.util.LinkedHashSet;
import java.util.Set;
import java.util.TreeSet;
import javax.swing.SwingUtilities; // for calling Swing listeners in EDT

/**
 *
 * @author Lari Natri
 */
public abstract class Job<R, P> implements Comparable<Job> {

    /**
     * Derived classes' work() methods should return or throw this object,
     * if they are cancelled. Jobs that support cancellation should
     * call setSupportsCancellation(true) in their constructors and in the
     * work() method they should repeatedly check cancelPending().
     */
    static protected final Throwable CANCELLEDRESULT = new Throwable("CANCEL");

    // finish listeners to be called in current thread. Used also as a
    // synchronization lock for operating with finished listeners.
    private final LinkedHashSet<JobFinishedListener> finishedListeners;
    // finished listeners to be called in EDT
    private final LinkedHashSet<JobFinishedListener> finishedListenersSwing;

    private final HashMap<Job,JobResultEvent> subjobResults;
    private final TreeSet<Job> subjobs;
    private final Object subWaitLock; // lock for waiting
    private final Timer timer;

    // set by constructor or by ParameterizedJob.getParametersForModification(),
    // accessed by execute() and passed to work()
    P parameter = null;

    private Job waitedJob = null; // the subjob currently waited for
    private JobState state = JobState.CREATED;
    private JobPriority priority = JobPriority.NORMAL;
    private Integer additionOrder = null; // set by register()
    private JobPool pool = null;          // set by register()
    private Job parent = null;            // set by parent's addSubjob()
    private EnumSet<JobCategory> categories = EnumSet.noneOf(JobCategory.class);
    private EnumSet<JobCategory> conflictingCategories = EnumSet.noneOf(JobCategory.class);
    private String name = "Job";
    private int progress = 0;
    private String statusMessage = "";
    private boolean supportsCancellation = false;
    private boolean recursiveCancel = true;
    private boolean cancelPending = false;
    private boolean silent = false;
    






    /**** PUBLIC ****/


    /**
     * Constructs a job and sets it's parameter
     * @param p parameter (will be passed to {@link #execute()}).
     * @see ParameterizedJob
     */
    public Job(final P p) {
        this.finishedListeners = new LinkedHashSet<JobFinishedListener>();
        this.finishedListenersSwing = new LinkedHashSet<JobFinishedListener>();
        this.subjobs = new TreeSet<Job>();
        this.subjobResults = new HashMap<Job,JobResultEvent>();
        this.subWaitLock = new Object();
        this.timer = new Timer();
        this.parameter = p;
    }

    /**
     * Constructs a job and sets it's parameter (which will be passed to
     * {@link #work(java.lang.Object)] to null.
     */
    public Job() {
        this(null);
    }

    /**
     * Returns the state this job is currently in.
     * @return state of this job
     * @see JobState
     */
    public final JobState getState() {
        synchronized(this) {
            return this.state;
        }
    }

   /**
     * Returns the name and state of this job as string.
    *  @return job's name and state
     */
    @Override
    public synchronized String toString() {
        return this.name + " ["+this.getState()+"]";
    }

    /**
     * Returns the name of this job. Never null.
     * @return job's name
     * @see #setName(java.lang.String)
     */
    public synchronized final String getName() {
        return this.name;
    }

    /**
     * Returns the progress of this job as a number {-1, [0..100]}. -1 signifies
     * unknown progress.
     * @return current progress of this job
     */
    public synchronized final int getProgress() {
        return this.progress;
    }

    /**
     * Returns the last job's status message reported in the Job's
     * {@link #execute()} method by calling {@link #reportStatus(String)}.
     * Not null.
     * @return status message
     */
    public synchronized final String getStatus() {
        return this.statusMessage;
    }

    /**
     * Returns the priority of this job. Guaranteed not to change after this
     * job has been added to a {@link JobPool}.
     * @return priority
     * @see JobPriority
     */
    public synchronized final JobPriority getPriority() {
        return this.priority;
    }

    /**
     * Returns execution time in nanoseconds. Time waited for other threads
     * is not included in the execution time.
     * @return time in nanoseconds
     */
    public final long getExecutionTime() {
        return this.timer.getExecutionTime();
    }

    /**
     * Returns the time spent waiting for other threads for data. Time spent
     * in the queue is not included.
     * @return time waited for subjobs, in nanoseconds
     */
    public final long getWaitingTime() {
        return this.timer.getPausedTime();
    }

    /**
     * Returns the categories this job is a part of. Categories should be set in
     * the job's constructor and are guaranteed not to change after the job has
     * been added to a {@link JobPool}. Can be empty, but not null.
     * @return categories
     * @see #getConflictingCategories()
     */
    public synchronized final EnumSet<JobCategory> getCategories() {
        return EnumSet.copyOf(this.categories);
    }

    /**
     * Returns the categories this job conflicts with. A Job will not be started
     * while any jobs whose categories it conflicts with are running, not
     * counting the ancestors of the would-be-started job. Guaranteed not to
     * change after the job's been added to a {@link JobPool}. Can be empty, but
     * not null.
     * @return categories this job conflicts with
     * @see #getCategories()
     */
    public synchronized final EnumSet<JobCategory> getConflictingCategories() {
        return EnumSet.copyOf(this.conflictingCategories);
    }

    /**
     * Returns true, if this job is set to be silent ie. it will not be shown
     * in the {@link JobPool}-created UI. Note that even silent jobs are
     * counted in {@link JobPool#getRunningThreadsCount()} and
     * s{@link JobPool#getUnfinishedRootJobsCount()}.
     * @return true if job is silent, false otherwise
     */
    public synchronized boolean isSilent() {
        return this.silent;
    }

    /**
     * Returns true if this job is an ancestor of the job given as parameter
     * ie. if parameter job is a subjob of this job; else returns false. If
     * the jobs are the same job, return value is false.
     * @param j
     * @return true if argument job is a subjob of this job
     * @see #sharesAncestorWith(net.aavalla.utils.jobpool.Job)
     */
    public boolean isAncestorOf(Job j) {
        if (j==this) { return false; }
        Job p = j.getParent();
        while (p!=null) {
            if (p==this) {
                return true;
            }
            p = p.getParent();
        }
        return false;
    }

    /**
     * Returns true, if this job and the job given as parameter share an
     * ancestor ie. if they are part of the same subtree of jobs.
     * @param j job
     * @return true if jobs share an ancestor, false otherwise
     * @see #isAncestorOf(net.aavalla.utils.jobpool.Job)
     */
    public boolean sharesAncestorWith(Job j) {
        Job p = this.getParent();
        while (p!=null) {
            if (p==j || p.isAncestorOf(j)) {
                return true;
            }
            p = p.getParent();
        }
        return false;
    }

    /**
     * Returns true, if this Job has been set to support cancellation. There
     * are no guarantees that cancelling actually works. Derived classes'
     * {@link #work(java.lang.Object)} methods ought to check
     * {@link #cancelPending()} periodically.
     * @return true if cancelling is supported, false otherwise
     */
    public synchronized final boolean supportsCancellation() {
        return this.supportsCancellation;
    }

    public synchronized final boolean getUseRecursiveCancel() {
        return this.recursiveCancel;
    }


    /**
     * Returns true, if this job has a cancel pending ie. it supports
     * cancellation and it has been instructed to be cancelled by
     * {@link #setCancelPending(boolean)}. Jobs that support cancellation,
     * ought to check this periodically in their {@link #work(java.lang.Object)}
     * method.
     * @return true if the job has been set to be cancelled, false otherwise
     */
    public synchronized final boolean cancelPending() {
        return (this.cancelPending && this.supportsCancellation);
    }

    /**
     * Sets the name for this job. Must be called before adding this job to
     * a pool, best place for doing so is in the derived class' constructor.
     * @param name new name for the job
     * @throws IllegalStateException Thrown, if called after the job has been
     *         added to a {@link JobPool} ie when
     *         {@code getState()!=JobState.CREATED}
     */
    public synchronized final void setName(String name) 
            throws IllegalStateException
    {
        if (this.state!=JobState.CREATED) {
            throw new IllegalStateException("Trying to change the name of " +
                    "the job after it's been added to a pool.");
        }
        this.name = name==null ? "" : name;
    }

    public synchronized final void setCancelPending(boolean cancel) {
        if (this.getState().equals(JobState.CREATED) || this.getState().equals(JobState.READY)) {
            synchronized(this) {
                this.cancelPending = cancel;
            }
            if (cancel) {
                this.finalizeJob(new JobResultEvent<R>(null,CANCELLEDRESULT,true,this));
            }
        } else if (this.supportsCancellation()) {
            synchronized(this) {
                this.cancelPending = cancel;
                if (this.recursiveCancel) {
                    synchronized(this.subWaitLock) {
                        for (Job sub : this.subjobs) {
                            sub.setCancelPending(cancel);
                        }
                    }
                }
            }
        }
    }

    public synchronized final void setUseRecursiveCancel(boolean recCancel) {
        this.recursiveCancel = recCancel;
    }


    /**
     * Adds a listener to be notified when this job has ended. Listener will be
     * run in the same thread as the actual Job.work(). Each listener is only
     * added once and thus executed only once. Listeners must be added before
     * the job finishes (or fails), otherwise an exception is thrown. To play
     * it safe, listeners should be added before adding the job to a pool.
     * @param l listener to add
     * @return true if listener didn't exist before
     * @throws IllegalStateException Thrown if called after this Job has already
     *         finished or failed.
     */
    public final boolean addJobFinishedListener(JobFinishedListener l) {
        synchronized(this) {
            if (this.state.equals(JobState.FINISHED)
                    || this.state.equals(JobState.FAILED)
                    || this.state.equals(JobState.CANCELLED))
            {
                throw new IllegalStateException("Trying to add JobFinishedListener" +
                        "to a job which has already ended");
            }
        }
        synchronized(this.finishedListeners) {
            if (this.finishedListenersSwing.contains(l)) {
                return false;
            }
            return this.finishedListeners.add(l);
        }
    }

    /**
     * Adds a listener to be notified when this job has ended. Listener will be
     * run in the same thread as the actual Job.work(). Each listener is only
     * added once and thus executed only once. Listeners added with this method
     * are run by the Swing's Event Dispatcher Thread. Meant to be used for
     * GUI updating. Listeners must be added before the job finishes (or fails),
     * otherwise an exception is thrown. To play it safe, listeners should be
     * added before adding the job to a pool.
     * @param l listener to add
     * @return true if listener didn't exist before
     * @throws IllegalStateException Thrown if called after this Job has already
     *         finished or failed.
     */
    public final boolean addJobFinishedListenerSwing(JobFinishedListener l) {
        synchronized(this) {
            if (this.state.equals(JobState.FINISHED)
                    || this.state.equals(JobState.FAILED)
                    || this.state.equals(JobState.CANCELLED))
            {
                throw new IllegalStateException("Trying to add JobFinishedListener" +
                        "to a job which has already finished, failed or cancelled");
            }
        }
        synchronized(this.finishedListeners) {
            if (this.finishedListeners.contains(l)) {
                return false;
            }
            return this.finishedListenersSwing.add(l);
        }
    }

    /**
     * Removes a listener. It will not be called after the Job finishes.
     * Listeners will be cleared automatically after Job finishes so there is
     * no need to call this method in the actual listener.
     * @param l listener to remove
     * @return true if the listener actually existed
     */
    public final boolean removeJobFinishedListener(JobFinishedListener l) {
        synchronized(this.finishedListeners) {
            if (this.finishedListenersSwing.remove(l)) {
                return true;
            }
            return this.finishedListeners.remove(l);
        }
    }



    /**** IMPLEMENTATIONS ****/

    /**
     * Compares Jobs by priorities. If this Job would be run before parameter
     * job, the value is negative, otherwise positive. 0 will be returned only
     * if objects are the same object. Consistent with {@code equals()}. Used
     * by {@link JobPool} to determine the order of starting jobs.
     * @param o other job to compare to
     * @return negative if this job is to be run before o job.
     */
    @Override
    public final int compareTo(Job o) {
        if (this==o) {
            return 0;
        }
        synchronized(this) {
            if (this.priority.ordinal() < o.priority.ordinal()) {
                return Integer.MIN_VALUE;
            }
            if (this.priority.ordinal() > o.priority.ordinal()) {
                return Integer.MAX_VALUE;
            }
            if (this.additionOrder==null) {
                if (o.additionOrder==null) {
                    return this.hashCode() - o.hashCode(); // last resort
                }
                return Integer.MAX_VALUE;
            }
            if (o.additionOrder==null) {
                return Integer.MIN_VALUE;
            }
            return this.additionOrder - o.additionOrder;
        }
    }


    /**** PROTECTED ****/


    /**
     * Sets the priority for this job. Must be called before adding this job
     * to a pool. Recommended place is in a derived class' constructor. Also
     * PoolThread uses this for new sub jobs to set sub job's priority to the
     * same as it's parent job.
     * @param newPriority priority for this job
     * @throws IllegalStateException Thrown if trying to set priority when
     *         this job's state is anything else than CREATED.
     */
    protected synchronized final void setPriority(JobPriority newPriority) 
            throws IllegalStateException
    {
        if (this.state!=JobState.CREATED) {
            throw new IllegalStateException("Trying to change job priority " +
                    "after it has been added to a pool");
        }
        this.priority = newPriority;
    }

    protected synchronized final boolean addCategory(JobCategory cat)
            throws IllegalStateException
    {
        if (this.state!=JobState.CREATED) {
            throw new IllegalStateException("Trying to add a category " +
                    "for a job after it's been added to a pool");
        }
        return this.categories.add(cat);
    }


    protected synchronized final void setCategories(Set<JobCategory> catset)
            throws IllegalStateException
    {
        if (this.state!=JobState.CREATED) {
            throw new IllegalStateException("Trying to add a category " +
                    "for a job after it's been added to a pool");
        }
        this.categories = EnumSet.copyOf(catset);
    }

    protected synchronized final boolean addConflictingCategory(JobCategory cat)
            throws IllegalStateException
    {
        if (this.state!=JobState.CREATED) {
            throw new IllegalStateException("Trying to add a conflicting " +
                    "category for a job after it's been added to a pool");
        }
        return this.conflictingCategories.add(cat);
    }


    protected synchronized final void setConflictingCategories(Set<JobCategory> catset)
            throws IllegalStateException
    {
        if (this.state!=JobState.CREATED) {
            throw new IllegalStateException("Trying to add a conflicting " +
                    "category for a job after it's been added to a pool");
        }
        this.conflictingCategories = EnumSet.copyOf(catset);
    }


    protected synchronized final void setSupportsCancellation(boolean supCan) {
        this.supportsCancellation = supCan;
    }

    protected synchronized final void setSilent(boolean silent)
            throws IllegalStateException
    {
        if (this.state!=JobState.CREATED) {
            throw new IllegalStateException("Trying to change job's silence " +
                    "status after it's been added to a pool");
        }
        this.silent = silent;
    }

    /**
     * Reports the progress of this job to listeners including the pool.
     * Should be called from within derived classes' work() method, but is also
     * called from PoolThread just before (setting it to -1) and after
     * (setting it to 100) calling work(). The progress should be in range
     * [0..100] or -1, which signifies unknown progress.
     * @param newProgress
     */
    protected final void reportProgress(int newProgress) {
        if (newProgress < -1) {
            newProgress = -1;
        }
        if (newProgress > 100) {
            newProgress = 100;
        }
        boolean changed = false;
        synchronized(this) {
            if (this.progress!=newProgress) {
                this.progress = newProgress;
                changed = true;
            }
        }
        if (changed) {
            this.pool.progressChanged(this);
        }
    }

    protected final void reportStatus(String newStatus) {
        if (newStatus==null) {
            newStatus = "";
        }
        boolean changed = false;
        synchronized(this) {
            if (!this.statusMessage.equals(newStatus)) {
                this.statusMessage = newStatus;
                changed = true;
            }
        }
        if (changed) {
            this.pool.statusChanged(this, newStatus);
        }
    }
    
    protected final void addSubjob(Job sub) {
        boolean added;
        synchronized(this) {
            if (this.state!=JobState.RUNNING) {
                throw new IllegalStateException("Job doesn't have a running " +
                        "thread object. Probable reason: a function which is " +
                        "supposed to be called only from within Job.work() is " +
                        "called from elsewhere.");
            }
        }
        synchronized(this.subWaitLock) {
            if (!this.subjobs.contains(sub) && sub.getState()!=JobState.CREATED) {
                throw new IllegalStateException("Subjob has been added to a pool" +
                        " before");
            }
            added = this.subjobs.add(sub);
            if (added) {
                synchronized(sub) {
                    sub.parent = this;
                }
            }
        }
        if (added) {
            pool.addSubjob(this, sub);
        }
    }

    @SuppressWarnings("unchecked")
    protected final <S> S waitSubJob(Job<S,?> sub) throws Throwable {
        this.addSubjob(sub);
        JobResultEvent res = this.waitSubJobActual(sub);
        if (res.failed()) {
            throw res.getFailReason();
        } else {
            return (S)res.getResult();
        }
    }


    /**** ABSTRACT PROTECTED ****/


    abstract protected R work(P p) throws Throwable;



    /**** PACKAGE ****/


    final Job getParent() {
        synchronized(this) {
            return this.parent;
        }
    }

    final void register(JobPool pool, int additionOrder) {
        synchronized(this) {
            this.pool = pool;
            this.additionOrder = additionOrder;
        }
    }

    /**
     * Sets the state for this. Called only from JobPool when changing
     * states between RUNNING, READY and WAITING. Beginning state (CREATED)
     * is set directly by Job in its constructor and ending states
     * (FINISHED, FAILED, CANCELLED) by PoolThread.run() after the job finishes.
     * @param state new state
     */
    final void setState(JobState state) {
        synchronized(this) {
            this.state = state;
        }
    }

    /**
     * Executes this job by calling Job.work(). After work returns, waits for
     * all subjobs to end, calls finalizeJob(), lets parent continue and
     * finally notifies the JobPool. Called only from PoolThread.run().
     */
    @SuppressWarnings("unchecked")
    void execute() {
        JobResultEvent result = null;

        try {
            this.timer.start();
            this.reportProgress(-1);
            if (this.cancelPending()) {
                throw Job.CANCELLEDRESULT;
            }
            
            R r = this.work(parameter); // call user's working method

            if (r==Job.CANCELLEDRESULT) {
                result = new JobResultEvent<R>(null,(Throwable)r,true,this);
            } else {
                result = new JobResultEvent<R>(r,null,false,this);
            }

        } catch (Throwable ex) { // job failed
            if (ex==Job.CANCELLEDRESULT) {
                result = new JobResultEvent<R>(null,ex,true,this);
            } else {
                result = new JobResultEvent<R>(null,ex,false,this);
            }
        } finally {
            
            synchronized(this.subWaitLock) {
                if (result.cancelled() && this.getUseRecursiveCancel()) {
                    for (Job sub : this.subjobs) {
                        sub.setCancelPending(true);
                    }
                }
                while (this.subjobs.size()>0) {
                    Job sub = this.subjobs.first();
                    this.waitSubJobActual(sub);
                }
            }
//            this.reportProgress(100);
//            Job par = null;
//            synchronized(this) {
//                par = this.parent;
//            }
//            // sets state, notify listeners, frees memory
//            this.finalizeJob(result);
//            if (par!=null) { // notifies parent if one exists
//                par.subEnded(this,result);
//            }
//            this.timer.stop();
//            this.pool.jobEnded(this,result,par); // notifies JobPool
            this.finalizeJob(result);
        }
    }



    /**** PRIVATE ****/


    private final JobResultEvent waitSubJobActual(Job sub)
            throws IllegalStateException
    {
        JobResultEvent res = null;
        synchronized(this.subWaitLock) {
            this.addSubjob(sub); // adds if needed
            if (this.subjobResults.containsKey(sub)) {
                res = this.subjobResults.remove(sub);
                this.subjobs.remove(sub);
            } else {
                this.pool.startedWaitingSub(this);
                this.waitedJob = sub;
                this.timer.pause();
                try {
                    this.subWaitLock.wait();
                } catch (InterruptedException ex) {}
                this.timer.unPause();
                this.waitedJob = null;
                res = this.subjobResults.remove(sub);
                this.subjobs.remove(sub);
                this.pool.finishedWaitingSub(this);
            }
        }
        return res;
    }

    /**
     * Records subjob's end subResult and if this Job is currently waiting for
     * the subjob, allows this to continue by calling notify() on the waiting
     * lock object. Called only from subjob's execute() after said job has
     * ended.
     * @param endedSub subjob that has ended
     * @param subResult the ended subjob's result
     */
    private final void subEnded(Job endedSub, JobResultEvent subResult) {
        synchronized(this.subWaitLock) {
            
            this.subjobResults.put(endedSub, subResult);
            if (waitedJob==endedSub) {
                subWaitLock.notify();
            }
            
        }
    }


    /**
     * Called only from Job.execute() after the job has ended. Sets the final
     * state, calls the listeners and frees memory.
     * @param subResult Job's subResult
     */
    private final void finalizeJob(final JobResultEvent result) {

        this.reportProgress(100);
        Job par = null;
        synchronized(this) {
            par = this.parent;
        }
        if (par!=null) { // notifies parent if one exists
            par.subEnded(this,result);
        }

        synchronized(this) {
            if (result.cancelled()) {
                this.state = JobState.CANCELLED;
            } else if (result.failed()) {
                this.state = JobState.FAILED;
            } else {
                this.state = JobState.FINISHED;
            }
            this.parent = null;
        }

        synchronized(this.finishedListeners) { // call EDT listeners
            for (final JobFinishedListener l : this.finishedListenersSwing) {
               SwingUtilities.invokeLater(new Runnable() {
                    @Override public void run() {
                        l.jobFinished(result);
                    } // run
                });
            }
            this.finishedListenersSwing.clear();
        }

        synchronized(this.finishedListeners) { // call other listeners
            for (JobFinishedListener l : this.finishedListeners) {
                try {
                    l.jobFinished(result);
                } catch (Exception ex) {}
            }
            this.finishedListeners.clear();
        }

        synchronized(this) { // asserts
            assert(this.subjobs.size()==0);
            synchronized(this.subWaitLock) {
                assert(this.subjobResults.size()==0);
            }
            assert(this.waitedJob==null);
            assert(this.state==JobState.FINISHED
                    || this.state==JobState.CANCELLED
                    || this.state==JobState.FAILED);
            assert(this.parent==null);
        }

        this.timer.stop();
        this.pool.jobEnded(this,result,par); // notifies JobPool
    }


}
