/*
 * Copyright 2010 Lari Natri <lari.natri@gmail.com>
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Lesser General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */


package net.aavalla.utils.jobpool;

import net.aavalla.utils.general.Disposeable;
import net.aavalla.utils.jobpool.gui.JobPoolForm;
import java.text.DecimalFormat;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedHashSet;
import java.util.LinkedList;
import java.util.Set;
import java.util.TreeSet;
import javax.swing.SwingUtilities;
import javax.swing.tree.DefaultMutableTreeNode;
import javax.swing.tree.DefaultTreeModel;
import javax.swing.tree.TreeModel;

/**
 *
 * @author Lari Natri
 */
public class JobPool implements Disposeable {

   
    private static int additionOrderNow = Integer.MIN_VALUE;

    private static final DecimalFormat twoDig = new DecimalFormat("######0.00");

    
    private int maxThreads = JobPool.getAvailableProcessors();
    private int maxRootJobs = maxThreads;
    private int threadPoolSize = Math.min(maxThreads, 8);
    private boolean runImmediateImmediately = true;
    private boolean betterPriorityJobsMustEndFirst = true;
    private final TreeSet<Job> readyRootJobs;
    private final TreeSet<Job> readySubJobs;
    private final TreeSet<Job> runningRootJobs; // root jobs which are running or have running subjobs
    private final HashSet<PoolThread> activeThreads; // contains threads which have a job RUNNING on them
    private final LinkedList<PoolThread> idleJobPool;
    private final HashMap<Job,PoolThread> jobToThread; // contains all jobs which are RUNNING or WAITING, ie started but not ended
    // maps exclusivity types to the maximum amount of concurrent threads
    // with that type. null value means no maximum. 0 means no jobs of that
    // type are allowed to run at all; they'll stay in the queue forever or
    // until the max is increased (or set to null)
    private final HashMap<JobCategory,Integer> categoryToMaxConcurrent;
    private final GuiJobTreeModelHandler uiTree;
    private final LinkedHashSet<JobPoolMessageListener> poolMessageListeners;
    private final LinkedHashSet<JobPoolMessageListener> poolMessageListenersSwing;


    public JobPool() {
        this.readyRootJobs = new TreeSet<Job>();
        this.readySubJobs = new TreeSet<Job>();
        this.runningRootJobs = new TreeSet<Job>();
        this.activeThreads = new HashSet<PoolThread>(this.maxThreads,1);
        this.jobToThread = new HashMap<Job,PoolThread>();
        this.idleJobPool = new LinkedList<PoolThread>();

        this.categoryToMaxConcurrent = new HashMap<JobCategory,Integer>
                (JobCategory.values().length,1);
        for (JobCategory jet : JobCategory.values()) {
            this.categoryToMaxConcurrent.put(jet, null);
        }
        this.uiTree = new GuiJobTreeModelHandler();
        this.poolMessageListeners = new LinkedHashSet<JobPoolMessageListener>();
        this.poolMessageListenersSwing = new LinkedHashSet<JobPoolMessageListener>();
    }

    public static int getAvailableProcessors() {
        return Runtime.getRuntime().availableProcessors();
    }

    public JobPoolForm getNewJobTreeForm(boolean popupMenu) {
        return new JobPoolForm(this.uiTree.getTreeModel(), this.uiTree.getSubtreeRunning(), popupMenu);
    }


    public synchronized void addJob(Job j) {
        System.out.println("DEBUG: JobPool.addJob: adding root job "+j);

        this.readyRootJobs.add(j);
        j.setState(JobState.READY);
        j.register(this, additionOrderNow++);

        this.uiTree.addRootjob(j);

        this.readyToRunning();
    }



    public synchronized void setMaxRootJobs(int max) {
        if (max < 0) {
            throw new IllegalArgumentException("Trying to set a negative " +
                    "value for maximum root jobs amount");
        }
        int oldMax = this.maxRootJobs;
        if (max>this.maxThreads) {
            this.maxThreads = max;
        }
        this.maxRootJobs = max;
        if (max>oldMax) {
            this.readyToRunning();
        }
    }

    public synchronized final void setJobPoolSize(int size) {
        if (size < 0) {
            throw new IllegalArgumentException("Trying to set a negative " +
                    "value for thread pool size");
        }
        this.threadPoolSize = size;
        while (this.threadPoolSize < this.idleJobPool.size()) {
            this.idleJobPool.removeFirst();
        }
    }

    public synchronized void setMaxThreads(int max) {
        if (max < 0) {
            throw new IllegalArgumentException("Trying to set a negative " +
                    "value for maximum threads");
        }
        int oldMax = this.maxThreads;
        if (max < this.maxRootJobs) {
            this.maxRootJobs = max;
        }
        this.maxThreads = max;
        if (max > oldMax) {
            this.readyToRunning();
        }
    }

    public synchronized final void setRunImmediateImmediately(boolean immediately) {
        boolean old = this.runImmediateImmediately;
        this.runImmediateImmediately = immediately;
        if (old==false && immediately==true) {
            this.readyToRunning();
        }
    }

    public synchronized final void setBetterPriorityJobsMustEndFirst(boolean betterFirst) {
        boolean old = this.betterPriorityJobsMustEndFirst;
        this.betterPriorityJobsMustEndFirst = betterFirst;
        if (old==true && betterFirst==false) {
            this.readyToRunning();
        }
    }

    public synchronized final boolean getBetterPriorityJobsMustEndFirst() {
        return this.betterPriorityJobsMustEndFirst;
    }

    public synchronized final boolean getRunImmediateImmediately() {
        return this.runImmediateImmediately;
    }

    public synchronized final Integer getMaxConcurrencyForCategory(JobCategory cat) {
        return this.categoryToMaxConcurrent.get(cat);
    }


    public synchronized final Integer setMaxConcurrencyForCategory(JobCategory cat, Integer max)
            throws IllegalArgumentException
    {
        if (max!=null && max<0) {
            throw new IllegalArgumentException("Trying to set negative value " +
                    "to maximum threads for a job category");
        }
        Integer old = this.categoryToMaxConcurrent.get(cat);
        this.categoryToMaxConcurrent.put(cat, max);
        if (old!=null && (max==null || max>old)) {
            this.readyToRunning();
        }
        return old;
    }

    public synchronized final void setMaxThreadsForAllExclusivityTypes(Integer max)
            throws IllegalArgumentException
    {
        if (max!=null && max<0) {
            throw new IllegalArgumentException("Trying to set negative value " +
                    "to maximum threads for an exclusivity type");
        }
        this.categoryToMaxConcurrent.clear();
        for (JobCategory jet : JobCategory.values()) {
            this.categoryToMaxConcurrent.put(jet, max);
        }
        this.readyToRunning();
    }

    public synchronized final void clearMaxThreadsForAllExclusivityTypes()
            throws IllegalArgumentException
    {
        this.setMaxThreadsForAllExclusivityTypes(null);
    }

    @Override
    public synchronized final void dispose() {
        this.setJobPoolSize(0);
        this.readyRootJobs.clear();
    }

    public synchronized int getUnfinishedRootJobsCount() {
        return this.runningRootJobs.size() + this.readyRootJobs.size();
    }

    public synchronized int getRunningThreadsCount() {
        return this.activeThreads.size();
    }

    public void addJobPoolMessageListenerSwing(JobPoolMessageListener l) {
        synchronized(this.poolMessageListeners) {
            if (!this.poolMessageListeners.contains(l)) {
                this.poolMessageListenersSwing.add(l);
            }
        }
    }

    public void addJobPoolMessageListener(JobPoolMessageListener l) {
        synchronized(this.poolMessageListeners) {
            if (!this.poolMessageListenersSwing.contains(l)) {
                this.poolMessageListeners.add(l);
            }
        }
    }

    public void removeJobPoolMessageListener(JobPoolMessageListener l) {
        synchronized(this.poolMessageListeners) {
            this.poolMessageListeners.remove(l);
            this.poolMessageListenersSwing.remove(l);
        }
    }

    /**** PACKAGE METHODS ****/



    /**
     * Called from this.threadEnded() only.
     * @param endedJob
     * @param result
     */
    synchronized final void jobEnded(Job endedJob,
                                         JobResultEvent result,
                                         Job parent)
    {

        this.runningRootJobs.remove(endedJob);
        this.jobToThread.remove(endedJob);

        // for cancellation before running
        this.readyRootJobs.remove(endedJob);
        this.readySubJobs.remove(endedJob);

        this.uiTree.jobEnded(endedJob,parent);

    }

    final void addSubjob(Job parent, Job sub) {
        synchronized(this) {
            this.readySubJobs.add(sub);
            sub.setState(JobState.READY);
            sub.register(this, additionOrderNow++);
        }

        this.uiTree.addSubjob(parent, sub);

        this.readyToRunning();
    }

    final void startedWaitingSub(Job parent) {
        synchronized(this) {
            parent.setState(JobState.WAITING);
            this.activeThreads.remove(this.jobToThread.get(parent));
        }

        this.uiTree.jobChanged(parent);
        this.readyToRunning();
    }

    void finishedWaitingSub(Job parent) {
        synchronized(this) {
            parent.setState(JobState.RUNNING);
            this.activeThreads.add(this.jobToThread.get(parent));
        }
        this.uiTree.jobChanged(parent);
    }

    final void progressChanged(Job j) {
        this.uiTree.jobChanged(j);
    }

    final void statusChanged(Job j, String status) {
        this.firePoolMessageEvent(j.getName()+": "+status);
    }

    /**
     * Called from PoolThread object to notify that a job has ended and the
     * job's runningThread is set to null.
     * @param thread the thread the ended endedJob was running on
     * @param result result of the ended Job
     */
    final void threadEnded(PoolThread t) {

        synchronized(this) {
            this.activeThreads.remove(t);
            if (this.idleJobPool.size()<this.threadPoolSize) {
                t.setIdle();
                this.idleJobPool.add(t);
            } else {
                t.dispose();
            }
        }

        this.firePoolMessageEvent(null);

        this.readyToRunning();

    }


    
    /**** PRIVATE METHODS ****/


    /**
     * Sets ready jobs running (starts them) as long as there are such ready
     * jobs available, which pass all the required criteria.
     * 1. j = Subjob with highest priority not skipped by category, if any ready
     * 2. If j is not of priority IMMEDIATE and there is a root level job ready
     *    with IMMEDIATE priority not skipped before by category, replace j with it
     * 3. If j is not yet found, j = ready root job with highest priority not
     *    skipped by category
     * 4. If j not found, abort
     * 5. If too many threads running, abort (unless runImmediateImmediately
     *    is set and j is of priority IMMEDIATE)
     * 6. If j root job and too many root jobs running, abort (unless
     *    runImmediateImmediately is set and j is of priority IMMEDIATE)
     * 7. If there is any waiting or running job of a category conflicting
     *    with j and which is not an ancestor of j, skip j by category and go
     *    back to 1.
     * 8. If there are maximum number of _such_ jobs, which are of a shared
     *    category with j and are not j's ancestors, skip j by category and go
     *    back to 1. Note that each non-ancestor of j (sub)job sharing a
     *    category will be counted. (Perhaps should be changed, mm...)
     * 9. If betterPriorityJobsMustEndFirst is set and there is a job with
     *    better priority than j running or waiting, abort
     * 10. Start j!
     */
    @SuppressWarnings("unchecked") // for EnumSet casting
    private synchronized void readyToRunning() {

        HashSet<Job> skip = new HashSet<Job>();

        while (true) {
            boolean sub = true;
            Job j = null;
            if (!this.readySubJobs.isEmpty()) { // try to get subjob first
                for (Job r : this.readySubJobs) {
                    if (!skip.contains(r)) {
                        j = r;
                        break;
                    }
                }
            }
            if (j!=null && j.getPriority()!=JobPriority.IMMEDIATE) {
                if (!this.readyRootJobs.isEmpty()) {
                    Job j2 = this.readyRootJobs.first();
                    if (!skip.contains(j2) && j2.getPriority()==JobPriority.IMMEDIATE) {
                        j = null; // allow immediate root jobs to go before non-immediate subjobs
                    }
                }
            }
            if (j==null && !this.readyRootJobs.isEmpty()) { // then rootjob
                sub = false;
                for (Job r : this.readyRootJobs) {
                    if (!skip.contains(r)) {
                        j = r;
                        break;
                    }
                }
            }

            if (j==null) { break; } // no sub or root jobs ready: break

            boolean ok = false;
            if (this.runImmediateImmediately && j.getPriority()==JobPriority.IMMEDIATE) {
                ok = true;
            } else if (sub && this.activeThreads.size() < this.maxThreads) {
                ok = true;
            } else if (!sub && this.runningRootJobs.size() < this.maxRootJobs) {
                ok = true;
            }
            if (!ok) { break; }

            if (!j.getCategories().isEmpty() || !j.getConflictingCategories().isEmpty()) { // check categories
                System.out.println("DEBUG: Checking exc");
                Set<JobCategory> jCats = j.getCategories();
                Set<JobCategory> jConfCats = j.getConflictingCategories();
                HashMap<JobCategory,Integer> actCats = new HashMap<JobCategory,Integer>(JobCategory.values().length,1); // categories for active jobs with counts
                for (Job actJob : this.jobToThread.keySet()) {
                    if (actJob.isAncestorOf(j)) { // ignore if active job is an ancestor
                        System.out.println("DEBUG: EXC, is ancestor");
                        continue;
                    }
                    for (JobCategory cat : (Set<JobCategory>)actJob.getCategories()) {
                        if (jConfCats.contains(cat)) { // check for conflict
                            ok = false;
                            break;
                        }
                        if (actCats.containsKey(cat)) { // +1 concurrency count for category
                            actCats.put(cat, actCats.get(cat)+1);
                        } else { // first 
                            actCats.put(cat, 1);
                        }
                    }
                    if (!ok) { break; } // there was a conflict
                }

                if (ok) { // still ok: there was no conflict, check max concurrency per category
                    for (JobCategory cat : jCats) {
                        Integer maxForCat = this.categoryToMaxConcurrent.get(cat);
                        if (maxForCat!=null) {
                            System.out.println("DEBUG: EXC max not null");
                            System.out.println("DEBUG: "+cat+": max"+maxForCat+", current: "+actCats.get(cat));
                            if (maxForCat==0 || (actCats.containsKey(cat) && actCats.get(cat)>=maxForCat)) {
                                ok = false;
                                break;
                            }
                        }
                    }
                }
            }
            if (!ok) { // !ok: category conflict / max concurrency ==> add j to skiplist
                skip.add(j);
                System.out.println("DEBUG: exclusivity skipping");
                continue;
            }


            if (!this.runningRootJobs.isEmpty()
                    && this.betterPriorityJobsMustEndFirst
                    && this.runningRootJobs.first().getPriority().ordinal() < j.getPriority().ordinal())
            {
                ok = false;
            }
            if (!ok) { break; }

            if (sub) {
                this.readySubJobs.remove(j);
            } else {
                this.readyRootJobs.remove(j);
                this.runningRootJobs.add(j);
            }

            PoolThread t = idleJobPool.pollFirst();
            if (t==null) {
                t = new PoolThread(this);
            }
            j.setState(JobState.RUNNING);
            System.out.println("DEBUG: JobPool.readyToRunning: starting job "+j);
            this.activeThreads.add(t);

            jobToThread.put(j, t);
            t.startJob(j);

            this.uiTree.setRunning(j);

        }
    }

    private void firePoolMessageEvent(String message) {

        final JobPoolMessageEvent event;
        synchronized(this) {
            event = new JobPoolMessageEvent(message,this.activeThreads.size(),
                    this.runningRootJobs.size(),this.readyRootJobs.size());
        }

        synchronized(this.poolMessageListeners) { // call EDT listeners
            for (final JobPoolMessageListener l : this.poolMessageListenersSwing) {
               SwingUtilities.invokeLater(new Runnable() {
                    @Override public void run() {
                        l.jobPoolMessageReceived(event);
                    } // run
                });
            }
        }

        synchronized(this.poolMessageListeners) { // call other listeners
            for (JobPoolMessageListener l : this.poolMessageListeners) {
                try {
                    l.jobPoolMessageReceived(event);
                } catch (Exception ex) {}
            }
        }
    }

    
    /**** PRIVATE INNER CLASSES ****/


    public final class GuiJobTreeModelHandler {

        private final DefaultTreeModel tree;
        private final DefaultMutableTreeNode treeRoot;
        private final DefaultMutableTreeNode treeRunning;
        private final DefaultMutableTreeNode treeReady;
        private final DefaultMutableTreeNode treeFinished;
        private final DefaultMutableTreeNode treeFailed;
        private final HashMap<Job,DefaultMutableTreeNode> jobToNode;

        private GuiJobTreeModelHandler() {
            this.treeRunning = new DefaultMutableTreeNode("<html><b>Running</b> (0 jobs)</html>");
            this.treeReady = new DefaultMutableTreeNode("<html><b>Queued</b> (0 jobs)</html>");
            this.treeFinished = new DefaultMutableTreeNode("<html><b>Finished</b> (0 jobs)</html>");
            this.treeFailed = new DefaultMutableTreeNode("<html><b>Failed</b> (0 jobs)</html>");
            this.treeRoot = new DefaultMutableTreeNode("<html><b>Jobs</b> (0 jobs)</html>");
            this.treeRoot.add(treeRunning);
            this.treeRoot.add(treeReady);
            this.treeRoot.add(treeFinished);
            this.treeRoot.add(treeFailed);
            this.tree = new DefaultTreeModel(treeRoot);
            this.jobToNode = new HashMap<Job,DefaultMutableTreeNode>();
        }

        private final TreeModel getTreeModel() {
            return this.tree;
        }

        private final DefaultMutableTreeNode getSubtreeRunning() {
            synchronized(this) {
                return this.treeRunning;
            }
        }

        private final void jobChanged(final Job j) {
            final boolean silent = j.isSilent();
            if (silent) return;
            final DefaultMutableTreeNode node;
            synchronized(this) {
                node = this.jobToNode.get(j);
            }
            SwingUtilities.invokeLater(new Runnable() {
                @Override public void run() {
                    synchronized(GuiJobTreeModelHandler.this.treeRoot) {
                        tree.nodeChanged(node);
                    }
                }
            });
            firePoolMessageEvent(null);
        }

        private final void addRootjob(final Job j) {
            final boolean silent = j.isSilent();
            final DefaultMutableTreeNode node = new DefaultMutableTreeNode(j);
            synchronized(this) {
                this.jobToNode.put(j, node);
            }
            if (silent) {
                firePoolMessageEvent(null);
            } else {
                SwingUtilities.invokeLater(new Runnable() {
                    @Override public void run() {
                        synchronized(GuiJobTreeModelHandler.this.treeRoot) {
                            GuiJobTreeModelHandler.this.tree.insertNodeInto(node, treeReady, treeReady.getChildCount());
                            updateSubtreeCounts(false,true,false,false);
                        }
                    }
                });
                firePoolMessageEvent("Queued: "+j.getName());
            }
        }

        private final void addSubjob(final Job parent, final Job child) {
            final boolean silent = parent.isSilent() || child.isSilent();
            final DefaultMutableTreeNode childNode = new DefaultMutableTreeNode(child);
            final DefaultMutableTreeNode parentNode;
            synchronized(this) {
                this.jobToNode.put(child, childNode);
                parentNode = this.jobToNode.get(parent);
            }
            if (!silent) {
                SwingUtilities.invokeLater(new Runnable() {
                    @Override public void run() {
                        synchronized(GuiJobTreeModelHandler.this.treeRoot) {
                            GuiJobTreeModelHandler.this.tree.insertNodeInto(childNode, parentNode, parentNode.getChildCount());
                        }
                    }
                });
            }
            firePoolMessageEvent(null);
        }

        private final void setRunning(final Job j) {
            final boolean silent = j.isSilent();
            final DefaultMutableTreeNode node;
            final boolean isSub;
            synchronized(this) {
                node = this.jobToNode.get(j);
                if (j.getParent()!=null) {
                    isSub = true;
                } else {
                    isSub = false;
                }
            }
            if (silent) {
                firePoolMessageEvent(null);
            } else {
                SwingUtilities.invokeLater(new Runnable() {
                    @Override public void run() {
                        synchronized(GuiJobTreeModelHandler.this.treeRoot) {
                            if (isSub) {
                                tree.nodeChanged(node);
                            } else {
                                tree.removeNodeFromParent(node);
                                tree.insertNodeInto(node, treeRunning, treeRunning.getChildCount());
                                updateSubtreeCounts(true,true,false,false);
                            }
                        }
                    }
                });
                if (!isSub) {
                    firePoolMessageEvent("Started: "+j.getName());
                } else {
                    firePoolMessageEvent(null);
                }
            }

        }

        private final void jobEnded(Job endedJob, Job parent) {
            final boolean silent = endedJob.isSilent();
            final DefaultMutableTreeNode node;
            final JobState state;
            final String jobStr;
            final boolean isSub;
            synchronized(this) {
                node = jobToNode.remove(endedJob);
                jobStr = endedJob.toString()
                        + " (run for: "
                        + twoDig.format(((endedJob.getExecutionTime()+endedJob.getWaitingTime())/60000000000d))
                        + " min)";
                state = endedJob.getState();
                if (parent!=null) {
                    isSub = true;
                } else {
                    isSub = false;
                }
            }
            if (silent) {
                firePoolMessageEvent(null);
            } else {
                SwingUtilities.invokeLater(new Runnable() {
                    @Override public void run() {
                        synchronized(GuiJobTreeModelHandler.this.treeRoot) {
                            tree.removeNodeFromParent(node);
                            if (isSub) {
                                return;
                            }
                            switch(state) {
                                case FINISHED:
                                    tree.insertNodeInto(new DefaultMutableTreeNode(jobStr), treeFinished, treeFinished.getChildCount());
                                    updateSubtreeCounts(true,true,true,false);
                                    break;
                                case FAILED:
                                case CANCELLED:
                                    tree.insertNodeInto(new DefaultMutableTreeNode(jobStr), treeFailed, treeFailed.getChildCount());
                                    updateSubtreeCounts(true,true,false,true);
                                    break;
                            }
                        }
                    }
                });
                if (!isSub) {
                    firePoolMessageEvent("Ended: "+endedJob.getName());
                } else {
                    firePoolMessageEvent(null);
                }

            }
        }

        private final void updateSubtreeCounts(final boolean running,
                                               final boolean ready,
                                               final boolean finished,
                                               final boolean failed) {
            SwingUtilities.invokeLater(new Runnable() {
                @Override public void run() {
                    synchronized(GuiJobTreeModelHandler.this.treeRoot) {
                        if (running) {
                            treeRunning.setUserObject("<html><b>Running</b> ("+treeRunning.getChildCount()+" jobs)</html>");
                            tree.nodeChanged(treeRunning);
                        }
                        if (ready) {
                            treeReady.setUserObject("<html><b>Queued</b> ("+treeReady.getChildCount()+" jobs)</html>");
                            tree.nodeChanged(treeReady);
                        }
                        if (finished) {
                            treeFinished.setUserObject("<html><b>Finished</b> ("+treeFinished.getChildCount()+" jobs)</html>");
                            tree.nodeChanged(treeFinished);
                        }
                        if (failed) {
                            treeFailed.setUserObject("<html><b>Failed</b> ("+treeFailed.getChildCount()+" jobs)</html>");
                            tree.nodeChanged(treeFailed);
                        }

                    }
                }
            });
        }

    }

}
