/*
 * Copyright 2009 Lari Natri <lari.natri@gmail.com>
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Lesser General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */


package net.aavalla.utils.jobpool;

import net.aavalla.utils.jobpool.parameters.ParameterCollection;


/**
 *
 * @author Lari Natri
 */
public abstract class ParameterizedJob<R, P extends ParameterCollection> extends Job<R,P> {


    /**
     * Returns a ParameterCollection object to be edited before starting the
     * Job with it.
     * @return
     */
    public synchronized final ParameterCollection getParametersForModification() {
        if (this.parameter==null) { // set the parameter object only once
            this.parameter = this.getNonfinalEditableParameters();
        }
        return this.parameter;
    }

    /**
     * Returns the ParameterCollection for displaying. If parameter is not set
     * yet, or it is not yet finalized, null is returned.
     * @return parameters used by this job or null, if not set final yet
     */
    public synchronized final ParameterCollection getFinalParametersForDisplay() {
        if (this.parameter!=null && this.parameter.isFinal()) {
            return this.parameter;
        } else {
            return null;
        }
    }

    /**
     * This should be overridden in derived classes. The method should return
     * a non-final ParameterCollection object filled with default values and
     * annotated with FieldAnnotation annotations in such a way that those
     * parameters can be modified. That same object (with modifications) will
     * later be passed to work() method as a parameter.
     * @return parameters object which is not final yet
     */
    protected abstract P getNonfinalEditableParameters();
    
    @Override
    protected abstract R work(final P params) throws Throwable;

    /**
     * Creates and starts a new thread running this job, if it's not already
     * running. Called from PoolThread class. Parameter is set final.
     */
    @Override
    final void execute() {
        if (this.parameter==null) {
            this.getParametersForModification();
        }
        // TODO: do something if any of the obligatory fields are still null -- an exception perhaps?
        // TODO: validate fields -- do something if they don't pass the tests
        this.parameter.makeFinal();
        super.execute();
    }
}
