/*
 * Copyright 2010 Lari Natri <lari.natri@gmail.com>
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Lesser General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */


package net.aavalla.utils.jobpool;

import net.aavalla.utils.general.Disposeable;


/**
 *
 * @author Lari Natri
 */
final class PoolThread extends Thread implements Disposeable {

    private final JobPool pool;             // JobPool which created this thread
    private final Object jobLock = new Object(); // for sync and wait
    private Job currentJob = null;          // lock jobLock before accessing
    private boolean disposePending = false; // lock jobLock before accessing

    PoolThread(JobPool pool) {
        super("net.aavalla.utils.jobpool.PoolThread"); // NOI18N
        this.pool = pool;
        this.setDaemon(false);
        this.start();
    }

    final boolean isActive() {
        synchronized(this.jobLock) {
            return this.currentJob!=null;
        }
    }

    final boolean isDisposed() {
        synchronized(this.jobLock) {
            return this.disposePending;
        }
    }

    /**
     * Called from JobPool to start running a new job.
     * @param j
     */
    final void startJob(Job j) {
        if (j==null) {
            throw new NullPointerException("Trying to run null job");
        }

        synchronized(this.jobLock) {
            if (this.isActive()) { // thread already active
                throw new IllegalStateException("Trying to run a Job with a " +
                        "thread which is already active");
            }
            if (this.isDisposed()) {
                throw new IllegalStateException("Can not add Jobs to an " +
                        "already disposed thread");
            }
            this.currentJob = j;
            this.setName("PoolThread running "+j);
            this.jobLock.notify();
        }
    }


    final void setIdle() {
        synchronized(this.jobLock) {
            this.currentJob = null;
            this.setName("PoolThread [idle]");
        }
    }

    final Job getCurrentJob() {
        synchronized(this.jobLock) {
            return this.currentJob;
        }
    }

    /**
     * Stops this thread (after currently possibly running Job ends). Called
     * from JobPool only.
     */
    @Override
    public void dispose() {
        synchronized(this.jobLock) {
            this.disposePending = true;
            this.jobLock.notify();
        }
    }



    @Override
    public final void run() {

        while(true) {
            if (this.isDisposed()) {
                break;
            }
            boolean hasJob;
            synchronized(this.jobLock) {
                hasJob = this.currentJob!=null;
            }

            if (hasJob) { // if job available, execute it
                this.currentJob.execute();
                this.pool.threadEnded(this);
            } else {
                synchronized(this.jobLock) {
                    try {
                        if (this.currentJob==null) {
                            this.jobLock.wait();
                        }
                    } catch(InterruptedException ex) {}
                }
            }
            
        } // while true

        // disposed
        
        synchronized(this.jobLock) {
            this.currentJob = null;
        }

        return;
    }

}
