/*
 * Copyright 2010 Lari Natri <lari.natri@gmail.com>
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Lesser General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */


package net.aavalla.utils.jobpool.gui;

import net.aavalla.utils.jobpool.Job;
import net.aavalla.utils.jobpool.JobState;
import java.awt.Color;
import java.awt.Component;
import java.awt.Rectangle;
import java.text.DecimalFormat;
import javax.swing.JPanel;
import javax.swing.JTree;
import javax.swing.UIManager;
import javax.swing.tree.DefaultMutableTreeNode;
import javax.swing.tree.TreeCellRenderer;

/**
 *
 * @author Lari Natri
 */
public class JobTreeCellRenderer extends JPanel 
                                 implements TreeCellRenderer
{

    public static final DecimalFormat twoDig = new DecimalFormat("######0.00");

    private Color nonselFG;
    private Color nonselBG;
    private Color selFG;
    private Color selBG;

    /** Creates new panel JobTreeCellRenderer */
    public JobTreeCellRenderer() {
        initComponents();

        nonselFG = UIManager.getColor("Tree.textForeground"); //NOI18N
        nonselBG = UIManager.getColor("Tree.textBackground"); //NOI18N
        selFG = UIManager.getColor("Tree.selectionForeground"); //NOI18N
        selBG = UIManager.getColor("Tree.selectionBackground"); //NOI18N
        if (nonselFG==null) nonselFG = Color.BLACK;
        if (nonselBG==null) nonselBG = Color.WHITE;
        if (selFG==null) selFG = nonselBG;
        if (selBG==null) selBG = nonselFG;

    }


    @Override
    public Component getTreeCellRendererComponent(JTree tree,
                                                  Object value,
                                                  boolean selected,
                                                  boolean expanded,
                                                  boolean leaf,
                                                  int row,
                                                  boolean hasFocus)
    {
        Color fg = selected ? selFG : nonselFG;
        Color bg = selected ? selBG : nonselBG;
        this.setForeground(fg);
        this.label.setForeground(fg);
        this.stateLabel.setForeground(fg);
        this.setBackground(bg);
        this.label.setBackground(bg);
        this.stateLabel.setBackground(bg);


        Object userObj;
        if (value instanceof DefaultMutableTreeNode) {
            userObj = ((DefaultMutableTreeNode)value).getUserObject();
        } else {
            userObj = tree.convertValueToText(value, selected, expanded, leaf, row, hasFocus);
        }

        boolean enabled = tree.isEnabled();

        this.setEnabled(enabled);
        this.label.setEnabled(enabled);
        this.detProgressBar.setEnabled(enabled);
        this.indetProgressBar.setEnabled(enabled);

        if (userObj instanceof Job) {
            final Job j = (Job)userObj;
            JobState state = j.getState();
            int progress = j.getProgress();
            if (state==JobState.RUNNING || state==JobState.WAITING) {
                if (progress < 0) {
                    this.detProgressBar.setVisible(false);
                    this.indetProgressBar.setVisible(true);
                } else {
                    this.detProgressBar.setValue(progress);
                    this.detProgressBar.setString(null);
                    this.detProgressBar.setVisible(true);
                    this.indetProgressBar.setVisible(false);
                }
            } else {
                this.detProgressBar.setVisible(false);
                this.indetProgressBar.setVisible(false);
            }
            this.stateLabel.setVisible(true);
            this.label.setText(j.getName());
            StringBuilder sb = new StringBuilder("<html><b><u>"+j.getName()+"</u></b><br/><br/>");
            sb.append("State: ").append(j.getState()).append("<br/>");
            String status = j.getStatus();
            if (status!=null && !status.isEmpty()) {
                sb.append("Status: ").append(status).append("<br/>");
            }
            sb.append("Running time: ").append(twoDig.format((j.getExecutionTime() + j.getWaitingTime()) / 60000000000d)).append(" min<br/>");
            sb.append("Supports cancellation: ").append(j.supportsCancellation() ? " yes" : "no").append("<br/>");
            if (j.cancelPending()) {
                this.stateLabel.setText("[C|"+state.toString()+"]");
                sb.append("<b>Cancel is pending...</b><br/>");
            } else {
                this.stateLabel.setText("["+state.toString()+"]");
            }
            sb.append("<br/><i>Click Job with secondary button for options</i></html>");
            this.setToolTipText(sb.toString());

        } else {
            this.label.setText(userObj.toString());
            this.setToolTipText(this.label.getText());
            this.stateLabel.setText("");
            this.stateLabel.setVisible(false);
            this.detProgressBar.setVisible(false);
            this.indetProgressBar.setVisible(false);
        }

        this.setComponentOrientation(tree.getComponentOrientation());

        return this;
    }

    //////// OVERRIDDEN UNNEEDED METHODS
    //////// If more features are added, these overrides must be checked

    @Override
    public void revalidate() {}

    @Override
    public void repaint(long tm, int x, int y, int width, int height) {}

    @Override
    public void repaint(Rectangle r) {}

    @Override
    public void repaint() {}

    @Override
    protected void firePropertyChange(String propertyName, Object oldValue, Object newValue) {}

    @Override
    public void firePropertyChange(String propertyName, byte oldValue, byte newValue) {}

    @Override
    public void firePropertyChange(String propertyName, char oldValue, char newValue) {}

    @Override
    public void firePropertyChange(String propertyName, short oldValue, short newValue) {}

    @Override
    public void firePropertyChange(String propertyName, int oldValue, int newValue) {}

    @Override
    public void firePropertyChange(String propertyName, long oldValue, long newValue) {}

    @Override
    public void firePropertyChange(String propertyName, float oldValue, float newValue) {}

    @Override
    public void firePropertyChange(String propertyName, double oldValue, double newValue) {}

    @Override
    public void firePropertyChange(String propertyName, boolean oldValue, boolean newValue) {}




    /** This method is called from within the constructor to
     * initialize the form.
     * WARNING: Do NOT modify this code. The content of this method is
     * always regenerated by the Form Editor.
     */
    @SuppressWarnings("unchecked")
    // <editor-fold defaultstate="collapsed" desc="Generated Code">//GEN-BEGIN:initComponents
    private void initComponents() {

        label = new javax.swing.JLabel();
        stateLabel = new javax.swing.JLabel();
        detProgressBar = new javax.swing.JProgressBar();
        indetProgressBar = new net.aavalla.utils.jobpool.gui.DisplayableProgressBar();

        setPreferredSize(new java.awt.Dimension(370, 15));

        label.setFont(new java.awt.Font("Dialog", 0, 12)); // NOI18N
        label.setText("name");

        stateLabel.setFont(new java.awt.Font("Dialog", 0, 10)); // NOI18N
        stateLabel.setText("state");
        stateLabel.setToolTipText("Current State");

        detProgressBar.setFont(new java.awt.Font("Dialog", 0, 8)); // NOI18N
        detProgressBar.setToolTipText("Current Progress");
        detProgressBar.setMinimumSize(new java.awt.Dimension(10, 12));
        detProgressBar.setPreferredSize(new java.awt.Dimension(148, 12));
        detProgressBar.setStringPainted(true);

        indetProgressBar.setFont(new java.awt.Font("Dialog", 0, 10));
        indetProgressBar.setMinimumSize(new java.awt.Dimension(10, 12));
        indetProgressBar.setPreferredSize(new java.awt.Dimension(148, 12));

        javax.swing.GroupLayout layout = new javax.swing.GroupLayout(this);
        this.setLayout(layout);
        layout.setHorizontalGroup(
            layout.createParallelGroup(javax.swing.GroupLayout.Alignment.LEADING)
            .addGroup(layout.createSequentialGroup()
                .addComponent(label, javax.swing.GroupLayout.PREFERRED_SIZE, 179, javax.swing.GroupLayout.PREFERRED_SIZE)
                .addPreferredGap(javax.swing.LayoutStyle.ComponentPlacement.RELATED)
                .addComponent(indetProgressBar, javax.swing.GroupLayout.PREFERRED_SIZE, 50, javax.swing.GroupLayout.PREFERRED_SIZE)
                .addPreferredGap(javax.swing.LayoutStyle.ComponentPlacement.RELATED)
                .addComponent(detProgressBar, javax.swing.GroupLayout.PREFERRED_SIZE, 50, javax.swing.GroupLayout.PREFERRED_SIZE)
                .addPreferredGap(javax.swing.LayoutStyle.ComponentPlacement.RELATED)
                .addComponent(stateLabel, javax.swing.GroupLayout.PREFERRED_SIZE, 181, javax.swing.GroupLayout.PREFERRED_SIZE))
        );
        layout.setVerticalGroup(
            layout.createParallelGroup(javax.swing.GroupLayout.Alignment.LEADING)
            .addComponent(label, javax.swing.GroupLayout.DEFAULT_SIZE, javax.swing.GroupLayout.DEFAULT_SIZE, Short.MAX_VALUE)
            .addComponent(indetProgressBar, javax.swing.GroupLayout.DEFAULT_SIZE, 15, Short.MAX_VALUE)
            .addComponent(detProgressBar, javax.swing.GroupLayout.DEFAULT_SIZE, 15, Short.MAX_VALUE)
            .addComponent(stateLabel, javax.swing.GroupLayout.DEFAULT_SIZE, 15, Short.MAX_VALUE)
        );
    }// </editor-fold>//GEN-END:initComponents


    // Variables declaration - do not modify//GEN-BEGIN:variables
    private javax.swing.JProgressBar detProgressBar;
    private net.aavalla.utils.jobpool.gui.DisplayableProgressBar indetProgressBar;
    private javax.swing.JLabel label;
    private javax.swing.JLabel stateLabel;
    // End of variables declaration//GEN-END:variables

}
