/*
 * Copyright 2010 Lari Natri <lari.natri@gmail.com>
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Lesser General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */


package net.aavalla.utils.jobpool.gui;



import net.aavalla.utils.general.Disposeable;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import javax.swing.JTree;
import javax.swing.Timer;
import javax.swing.event.TreeExpansionEvent;
import javax.swing.event.TreeExpansionListener;
import javax.swing.event.TreeModelEvent;
import javax.swing.event.TreeModelListener;
import javax.swing.event.TreeWillExpandListener;
import javax.swing.tree.DefaultTreeModel;
import javax.swing.tree.ExpandVetoException;
import javax.swing.tree.MutableTreeNode;
import javax.swing.tree.TreeNode;
import javax.swing.tree.TreePath;

/**
 *
 * @author Lari Natri
 */
public class JobTreeRepainter implements Disposeable, TreeExpansionListener,
                                         TreeWillExpandListener, TreeModelListener
{

    

    private final class Performer implements ActionListener {
        @Override
        public final void actionPerformed(ActionEvent e) {
            if (tree.getVisibleRowCount()<minVisibleRowsForUpdate) {
                return;
            }
            tree.repaint();
        }
    }

    private final JTree tree;
    private final ActionListener performer;
    private final javax.swing.Timer timer;
    private final TreePath consideredSubtree;
    private final int minVisibleRowsForUpdate;

    public JobTreeRepainter(JTree tree,  MutableTreeNode consideredSubtree,
                            int millis, int minVisibleRows)
    {
        this.tree = tree;

        this.performer = new Performer();
        this.timer = new Timer(millis,this.performer);
        this.consideredSubtree = new TreePath(((DefaultTreeModel)tree
                .getModel()).getPathToRoot(consideredSubtree));
        tree.getModel().addTreeModelListener(this);
        this.tree.addTreeWillExpandListener(this);
        this.tree.addTreeExpansionListener(this);
        this.minVisibleRowsForUpdate = minVisibleRows;
        this.maybeStop();
        this.maybeRestart();
    }

    @Override
    public void dispose() {
        this.tree.removeTreeWillExpandListener(this);
        this.tree.removeTreeExpansionListener(this);
        this.timer.stop();
    }

    @Override
    public void treeCollapsed(TreeExpansionEvent event) {
        this.maybeStop();
    }

    @Override
    public void treeExpanded(TreeExpansionEvent event) {
        this.maybeRestart();
    }

    @Override
    public void treeWillCollapse(TreeExpansionEvent event) throws ExpandVetoException {
        return;
    }

    @Override
    public void treeWillExpand(TreeExpansionEvent event) throws ExpandVetoException {
        return;
    }

    @Override
    public void treeNodesChanged(TreeModelEvent e) {
        return;
    }

    @Override
    public void treeNodesInserted(TreeModelEvent e) {
        this.maybeRestart();
    }

    @Override
    public void treeNodesRemoved(TreeModelEvent e) {
        this.maybeStop();
    }

    @Override
    public void treeStructureChanged(TreeModelEvent e) {
        this.maybeStop();
        this.maybeRestart();
    }

    public void stop() {
        if (!this.timer.isRunning()) {
            return;
        }
        this.timer.stop();
    }

    public void maybeRestart() {
        if (this.timer.isRunning()) {
            return;
        }
        if (tree.isExpanded(consideredSubtree)
                && ((TreeNode)consideredSubtree.getLastPathComponent()).getChildCount()>0)
        {
            this.timer.restart();
        }
    }

    @Override
    protected void finalize() throws Throwable {
        this.dispose();
        super.finalize();
    }

    private void maybeStop() {
        if (this.timer.isRunning()==false) {
            return;
        }
        if (tree.isCollapsed(consideredSubtree) 
                || ((TreeNode)consideredSubtree.getLastPathComponent()).getChildCount()==0)
        {
            this.timer.stop();
        }
    }

}

