/*
 * Copyright 2009 Lari Natri <lari.natri@gmail.com>
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Lesser General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */


package net.aavalla.utils.jobpool.parameters;

import net.aavalla.utils.jobpool.parameters.gui.ValueComponent;
import net.aavalla.utils.jobpool.parameters.validators.ParameterValidator;
import java.lang.reflect.Field;
import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedList;
import javax.print.attribute.UnmodifiableSetException;
import javax.swing.SwingUtilities;

/**
 *
 * @author Lari Natri
 */
public class FieldInstance {

    private static final LinkedList<FieldTypeHandler> fieldTypeHandlers = new LinkedList<FieldTypeHandler>();
    private static final HashMap<Class,FieldTypeHandler> fieldTypeToHandlerMap = new HashMap<Class,FieldTypeHandler>(8);

    static {
        addFieldTypeHandler(new DefaultFieldTypeHandler());
    }


    private final ParameterCollection owner;
    private final String name;
    private final String descr;
    private final String fullHtmlDescription; // constructed in constructor
    private final boolean optional;
    private final boolean internal;
    private final ParameterValidator validator;
    private final HashMap<FieldListener,Boolean> fieldListeners;
    private final Class supportedFieldType;
    private final int lines;

    private Field field;
    
    public FieldInstance(ParameterCollection owner, String name, String descr,
                         int lines, boolean optional, boolean internal,
                         Class<? extends ParameterValidator> validator, String[] validatorParams, Field f)
    {
        this.owner = owner;
        this.name = name;
        this.descr = descr;
        this.lines = lines;
        this.optional = optional;
        this.internal = internal;
        ParameterValidator val = null;
        try {
            val = validator.getConstructor(FieldInstance.class,validatorParams.getClass()).newInstance(new Object[] { this, validatorParams} );
        } catch (Exception ex) {
            // no validator, ok
        }
        this.validator = val;

        StringBuilder htmlSB = new StringBuilder();
        htmlSB.append("<html><b>"); // NOI18N
        htmlSB.append(name);
        htmlSB.append("</b><br/><br/>"); // NOI18N
        htmlSB.append(descr.replaceAll("\n", "<br/>")); // NOI18N
        if (val!=null) {
            String valRules = val.uiValidationRules();
            if (valRules!=null && valRules.length()>0) {
                htmlSB.append("<br/><br/>"); // NOI18N
                htmlSB.append(valRules.replaceAll("\n", "<br/>")); // NOI18N
            }
        }
        htmlSB.append("</html>"); // NOI18N
        this.fullHtmlDescription = htmlSB.toString();


        this.fieldListeners = new HashMap<FieldListener, Boolean>(2);
        this.field = f;


        Class fieldType = this.field.getType();
        Class supportedType = null;

        synchronized(fieldTypeHandlers) {
            for (Iterator<FieldTypeHandler> it = fieldTypeHandlers.descendingIterator(); it.hasNext();) {
                FieldTypeHandler fth = it.next();
                supportedType = fth.getAsSupportedType(fieldType);
                if (supportedType!=null) {
                    fieldTypeToHandlerMap.put(supportedType, fth);
                    break; // found
                }
            }
        }
        this.supportedFieldType = supportedType;
        if (supportedType==null) {
            supportedType = fieldType;
        }

    }

    
    public ParameterCollection getOwner() { return this.owner; }
    
    public String getName() { return this.name; }

    public String getDescription() { return this.descr; }

    public String getFullHtmlDescription() { return this.fullHtmlDescription; }

    public int getLines() { return this.lines; }

    /**
     * If a field is optional, it's value should be set to null if user
     * chooses not to opt.
     */
    public boolean isOptional() { return this.optional; }

    /**
     * If isInternal() is true, the field should not be displayed in UI.
     */
    public boolean isInternal() { return this.internal; }

    /**
     * Returns the type of this field in a form that is most supported
     * @return
     */
    public Class getType() {
        return this.supportedFieldType;
    }

    public synchronized Object getValue() {
        this.field.setAccessible(true);
        Object value = null;
        try {
            value = this.field.get(this.owner);
        } catch (IllegalAccessException ex) {} // never happens
        return value;
    }

    public synchronized boolean setValue(Object value) throws UnmodifiableSetException {
        if (this.owner.isFinal()) {
            this.fireValueChangedEvent(this.getValue());
            return false;
        }

        if (value==null) {
            return this.actualSetValue(value);
        }

        synchronized(fieldTypeToHandlerMap) {
            value = fieldTypeToHandlerMap.get(this.getType()).convertType(value, value.getClass(), this.getType());
        }
        
        if (value.getClass().equals(this.getType())) {
            return this.actualSetValue(value);
        } else {
            throw new IllegalArgumentException("Trying to set a parameter with an incorrect type ("+value.getClass()+" -> "+this.field.getType());
        }

    }

    @SuppressWarnings("unchecked")
    private synchronized boolean actualSetValue(Object value) {
        try {
            if (this.validator!=null) {
                String valResult = this.validator.validate(value);
                if (valResult!=null) {
                    this.fireValidationFailed(valResult);
                    return false;
                }
            }
            this.field.setAccessible(true);
            this.field.set(this.owner, value);
            this.fireValueChangedEvent(this.getValue());

        } catch (IllegalAccessException ex) {} // never happens
        return true;
    }

    @SuppressWarnings("unchecked")
    public String validate(Object value) {
        if (this.validator==null) {
            return null;
        }
        return this.validator.validate(value);
    }

    public ValueComponent getValueComponent(boolean allowEdit) {
        synchronized(fieldTypeToHandlerMap) {
            return fieldTypeToHandlerMap.get(this.getType()).getValueComponentFor(this.getType(), this, allowEdit);
        } 
    }

    public String uiValidationRules() {
        if (this.validator==null) {
            return "";
        }
        return this.validator.uiValidationRules();
    }

    public void addFieldListener(FieldListener listener) {
        this.addFieldListener(listener, false);
    }

    public void addFieldListener(FieldListener listener, boolean isSwingListener) {
        synchronized(this.fieldListeners) {
            this.fieldListeners.put(listener, isSwingListener);
        }
    }

    public boolean removeFieldListener(FieldListener listener) {
        synchronized(this.fieldListeners) {
            return this.fieldListeners.remove(listener)!=null;
        }
    }

    @SuppressWarnings("unchecked")
    private void fireValueChangedEvent(final Object newValue) {
        synchronized(this.fieldListeners) {
            for (final FieldListener listener : this.fieldListeners.keySet()) {
                if (this.fieldListeners.get(listener).equals(Boolean.TRUE)) {
                    SwingUtilities.invokeLater(new Runnable() {

                        @Override public void run() {
                            listener.fieldValueChanged(FieldInstance.this, newValue);
                        }
                    });
                } else {
                    listener.fieldValueChanged(FieldInstance.this, newValue);
                }
            }
        }
    }

    public void fireValidationFailed(final String reason) {
        synchronized(this.fieldListeners) {
            for (final FieldListener listener : this.fieldListeners.keySet()) {
                if (this.fieldListeners.get(listener).equals(Boolean.TRUE)) {
                    SwingUtilities.invokeLater(new Runnable() {
                        @Override public void run() {
                            listener.validationFailed(FieldInstance.this, reason);
                        }
                    });
                } else {
                    listener.validationFailed(FieldInstance.this, reason);
                }
            }
        }
    }

    public static void addFieldTypeHandler(FieldTypeHandler fth) {
        synchronized(fieldTypeHandlers) {
            fieldTypeHandlers.add(fth);
        }
    }

    public static void removeFieldTypeHandler(FieldTypeHandler fth) {
        synchronized(fieldTypeHandlers) {
            fieldTypeHandlers.remove(fth);
        }
    }


}
