/*
 * Copyright 2009 Lari Natri <lari.natri@gmail.com>
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Lesser General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */


package net.aavalla.utils.jobpool.parameters;

import java.lang.reflect.Field;
import java.net.URL;
import java.util.HashSet;
import java.util.ListIterator;
import java.util.Vector;

/**
 *
 * @author Lari Natri
 */
public abstract class ParameterCollection implements Iterable<FieldInstance> {


    private Vector<FieldInstance> fields;
    private boolean isFinal = false;
    private String description = ""; // NOI18N
    private String helpID = null;
    private URL helpURL = null;
    private final HashSet<String> hiddenFields = new HashSet<String>(1,1);

    public ParameterCollection() {   }

    public final synchronized boolean makeFinal() {
        this.isFinal = true;
        return true;
    }

    public final synchronized boolean isFinal() {
        return this.isFinal;
    }

    /**
     * Returns the description of this ParameterCollection
     * @return Text to be displayed in the UI
     */
    public final synchronized String getDescription() {
        return description;
    }

    /**
     * This ought to validate the relations of all annotated fields. It should
     * be called from UI parameter handler after all individual fields have
     * been validated. For individual field validation, use classes derived
     * from <link>ParameterValidator</link>. Derived classes should overwrite
     * this method. The default implementation only returns null signifying
     * that parameter collection is valid.
     * @return null if validation ok, reason as a string if not
     */
    public synchronized String validateFieldRelations() {
        return null;
    }

    public final synchronized String getHelpID() {
        return this.helpID;
    }

    public final synchronized URL getHelpURL() {
        return this.helpURL;
    }
    
    @Override
    public final synchronized ListIterator<FieldInstance> iterator() {
        this.setFields();
        return this.fields.listIterator();
    }

    public final synchronized Vector<FieldInstance> parseFieldsWithAnnotations(Object obj) {

        Vector<FieldInstance> resultFields = new Vector<FieldInstance>();
        for(Class c=obj.getClass(); c!=null; c=c.getSuperclass()) {

            int additionIndex = 0;

            for(Field f : c.getDeclaredFields()) {
  
                FieldAnnotations fa = f.getAnnotation(FieldAnnotations.class);

                if (fa!=null && !this.hiddenFields.contains(fa.name())) {
                    f.setAccessible(true);
                    FieldInstance fi = new FieldInstance(this, fa.name(),
                                                         fa.description(),
                                                         fa.lines(),
                                                         fa.optional(),
                                                         fa.internal(),
                                                         fa.validator(),
                                                         fa.validatorSettings(),
                                                         f);
                    resultFields.add(additionIndex++,fi);
                }
            }
        }
        resultFields.trimToSize();

        return resultFields;
    }

    public final synchronized void setDescription(String description) {
        if (description==null) {
            description = ""; // NOI18N
        }
        this.description = description;
    }

    public final synchronized void setHelpID(String helpID) {
        this.helpID = helpID;
    }

    public final synchronized void setHelpURL(URL helpURL) {
        this.helpURL = helpURL;
    }

    /**
     * Hides a field from user editing, but does not remove it's value. Should
     * be called before passing this ParameterCollection to a ParameterDialog.
     * @param fName Name of the field as in annotation 'name'
     * @return true if field did not exist in hide set and is now added to it
     */
    public final synchronized boolean hideFieldByName(String fName) {
        if (fName==null) {
            return false;
        }
        boolean added = this.hiddenFields.add(fName);
        if (added) {
            this.fields = null;
        }
        return added;
    }

    /**
     * Unhides ie removes a field from hide set. Should be called before
     * passing this ParameterCollection to a ParameterDialog.
     * @param fName Name of the field as in annotation 'name'
     * @return true if field existed in hide set and is now removed
     */
    public final synchronized boolean unHideFieldByName(String fName) {
        if (fName==null) {
            return false;
        }
        boolean removed = this.hiddenFields.remove(fName);
        if (removed) {
            this.fields = null;
        }
        return removed;
    }

    /**
     * Unhides all fields ie clears the hide set.
     * @return true if some fields were hidden before calling this
     */
    public final synchronized boolean clearHiddenFields() {
        if (this.hiddenFields.isEmpty()) {
            return false;
        }
        this.hiddenFields.clear();
        this.fields = null;
        return true;
    }
    
    /**
     * Returns a clone of the hidden fields set.
     * @return
     */
    @SuppressWarnings("unchecked")
    public synchronized final HashSet<String> getHiddenFields() {
        return (HashSet<String>)this.hiddenFields.clone();
    }

    /**
     * Copies the hidden field set from the clone of the hidden fields set in
     * the ParameterCollection given as parameter. Overwrites current hidden
     * fields in this.
     * @param from source for copying
     * @return true if some change was made, false otherwise
     */
    public final synchronized boolean copyHiddenFieldsSetFrom(ParameterCollection from) {
        if (from==null) {
            return false;
        }
        HashSet<String> otherHiddenFields = from.getHiddenFields();
        if (otherHiddenFields.isEmpty() && this.hiddenFields.isEmpty()) {
            return false;
        }
        this.hiddenFields.clear();
        this.hiddenFields.addAll(otherHiddenFields);
        this.fields = null;
        return true;
    }

    private final synchronized Vector<FieldInstance> setFields() {
        if (this.fields==null) {
            this.fields = parseFieldsWithAnnotations(this);
        }
        return this.fields;
    }

}
