/*
 * Copyright 2010 Lari Natri <lari.natri@gmail.com>
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Lesser General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package net.aavalla.utils.jobpool.parameters;


import java.util.Collection;
import java.util.Collections;
import java.util.EnumSet;
import java.util.Iterator;
import java.util.Set;

/**
 * A synchronized wrapper for EnumSet. Allows for extracting element type
 * information even in complicated generics situations (hopefully).
 * @author Lari Natri
 */
public final class SyncEnumSet<E extends Enum<E>> implements Set<E> {

    private final Class<E> type;
    private final Set<E> set;

    private SyncEnumSet(Class<E> enumClass, Set<E> enumSet) {
        this.type = enumClass;
        this.set = Collections.synchronizedSet(enumSet);
    }

    /**
     * Returns the element type (as Class, not Type) for this set.
     * @return element type
     */
    public final Class<E> getElementType() {
        return this.type;
    }

    @Override
    public final SyncEnumSet<E> clone() {
        EnumSet<E> eSet = EnumSet.noneOf(this.type);
        eSet.addAll(this.set);
        return new SyncEnumSet<E>(this.type,eSet);
    }

    public static <E extends Enum<E>> SyncEnumSet<E> noneOf(Class<E> elementType)
            throws NullPointerException
    {
        return new SyncEnumSet<E>(elementType,EnumSet.noneOf(elementType));
    }

    public static <E extends Enum<E>> SyncEnumSet<E> allOf(Class<E> elementType)
            throws NullPointerException
    {
        return new SyncEnumSet<E>(elementType,EnumSet.allOf(elementType));
    }

    public static <E extends Enum<E>> SyncEnumSet<E> copyOf(SyncEnumSet<E> s)
        throws NullPointerException 
    {
        return s.clone();
    }

    public static <E extends Enum<E>> SyncEnumSet<E> copyOf(Collection<E> c)
            throws NullPointerException, IllegalArgumentException {
        if (c instanceof SyncEnumSet) {
            return SyncEnumSet.copyOf((EnumSet<E>)c);
        } else {
            if (c.isEmpty())
                throw new IllegalArgumentException("Collection is empty");
            Iterator<E> it = c.iterator();
            SyncEnumSet<E> result = SyncEnumSet.of(it.next());
            while (it.hasNext())
                result.add(it.next());
            return result;
        }
    }

    public static <E extends Enum<E>> SyncEnumSet<E> complementOf(SyncEnumSet<E> s)
            throws NullPointerException
    {
        SyncEnumSet<E> result = SyncEnumSet.allOf(s.type);
        synchronized(s) {
            for (E e : s) {
                result.remove(e);
            }
        }
        return result;
    }

    public static <E extends Enum<E>> SyncEnumSet<E> of(E e)
            throws NullPointerException
    {
        return new SyncEnumSet<E>(e.getDeclaringClass(),EnumSet.of(e));
    }

    public static <E extends Enum<E>> SyncEnumSet<E> of(E first,E... rest)
            throws NullPointerException
    {
        return new SyncEnumSet<E>(first.getDeclaringClass(), EnumSet.of(first,rest));
    }

    public static <E extends Enum<E>> SyncEnumSet<E> range(E from, E to)
            throws NullPointerException, IllegalArgumentException
    {
        return new SyncEnumSet<E>(from.getDeclaringClass(),EnumSet.range(from, to));
    }

    @Override
    public boolean add(E e) throws NullPointerException {
        if (e==null) throw new NullPointerException("Element is null");
        return this.set.add(e);
    }

    @Override
    public boolean addAll(Collection<? extends E> c) {
        return this.set.addAll(c);
    }

    @Override
    public void clear() {
        this.set.clear();
    }

    @Override
    public boolean contains(Object o) {
        return this.set.contains(o);
    }

    @Override
    public boolean containsAll(Collection<?> c) {
        return this.set.containsAll(c);
    }

    @Override
    public boolean isEmpty() {
        return this.set.isEmpty();
    }

    @Override
    public Iterator<E> iterator() {
        return this.set.iterator();
    }

    @Override
    public boolean remove(Object o) {
        return this.set.remove(o);
    }

    @Override
    public boolean removeAll(Collection<?> c) {
        return this.set.removeAll(c);
    }

    @Override
    public boolean retainAll(Collection<?> c) {
        return this.set.retainAll(c);
    }

    @Override
    public int size() {
        return this.set.size();
    }

    @Override
    public Object[] toArray() {
        return this.set.toArray();
    }

    @Override
    public <T> T[] toArray(T[] a) {
        return this.set.toArray(a);
    }

    @Override
    public final String toString() {
        StringBuilder res = new StringBuilder();
        int i = 0;
        for (E e: this.set) {
            if (i>0) {
                res.append("; "); //NOI18N
            }
            res.append(e);
            ++i;
        }
        return res.toString();
    }




}
