/*
 * Copyright 2009, 2010 Lari Natri <lari.natri@gmail.com>
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Lesser General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */


package net.aavalla.utils.jobpool.parameters.gui;

import net.aavalla.utils.gui.docbrowser.SimpleDocBrowser;
import net.aavalla.utils.jobpool.ParameterizedJob;
import net.aavalla.utils.jobpool.parameters.FieldInstance;
import net.aavalla.utils.jobpool.parameters.ParameterCollection;
import net.aavalla.utils.general.Disposeable;
import net.aavalla.utils.jobpool.JobPool;
import net.aavalla.utils.jobpool.parameters.FieldListener;
import java.awt.Color;
import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.Image;
import java.awt.Insets;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.util.HashMap;
import javax.swing.BorderFactory;
import javax.swing.JCheckBox;
import javax.swing.JFrame;
import javax.swing.JLabel;
import javax.swing.JTextPane;


/**
 *
 * @author Lari Natri
 */
public class ParameterDialog extends javax.swing.JDialog implements FieldListener, Disposeable {

    private static final Object docBrowserLock = new Object();
    private static SimpleDocBrowser docBrowser = null;

    private final JobPool pool;
    private final ParameterizedJob job;
    private final ParameterCollection params;
    private final HashMap<ValueComponent,FieldInstance> map;
    private final HashMap<JCheckBox,ValueComponent> optMap;
    private final HashMap<FieldInstance,String> invalidFields;
    private final boolean viewOnly;
    private final String windowTitle;

    public ParameterDialog(java.awt.Frame parent, boolean modal, ParameterCollection params) {
        this(parent,modal,params,null,null,true,null);
    }

    public ParameterDialog(java.awt.Frame parent, boolean modal, ParameterizedJob job, JobPool pool) {
        this(parent,modal,job.getParametersForModification(),job,pool,false,null);
    }

    public ParameterDialog(java.awt.Frame parent, boolean modal, ParameterCollection params, Image icon) {
        this(parent,modal,params,null,null,true,icon);
    }

    public ParameterDialog(java.awt.Frame parent, boolean modal, ParameterizedJob job, JobPool pool, Image icon) {
        this(parent,modal,job.getParametersForModification(),job,pool,false,icon);
    }

    /** Creates new form ParameterDialog */
    private ParameterDialog(java.awt.Frame parent, boolean modal, ParameterCollection params, ParameterizedJob job, JobPool pool, boolean viewOnly, Image icon) {
        super(parent, modal);

        initComponents();

        if (parent!=null) {
            this.setLocationRelativeTo(parent);
        }

        if (icon!=null) {
            this.setIconImage(icon);
            this.getOwner().setIconImage(icon);
        }

        if (job!=null && params==null) {
            if (viewOnly) {
                params = job.getFinalParametersForDisplay();
            } else {
                params = job.getParametersForModification();
            }
        }

        this.pool = pool;
        this.params = params;
        this.job = job;

        this.map = new HashMap<ValueComponent,FieldInstance>();
        this.optMap = new HashMap<JCheckBox,ValueComponent>();
        this.invalidFields = new HashMap<FieldInstance,String>(4);
        this.viewOnly = viewOnly;

        if (this.params.getHelpID()==null && this.params.getHelpURL()==null) {
            this.jToolBar1.setVisible(false);
        }

        if (job==null) {
            if (viewOnly) {
                this.windowTitle = "View parameters";
            } else {
                this.windowTitle = "Set parameters";
            }
        } else {
            this.windowTitle = "Set parameters for " + job.getName();
        }

        this.createParameterUI();

    }

    /**
     * Recreates the UI and loads parameter values from underlying model
     */
    private void createParameterUI() {

        this.setTitle(this.windowTitle);

        jPanel1.setLayout(new GridBagLayout());

        Insets ins = new Insets(5,5,5,5);

        GridBagConstraints cTop = new GridBagConstraints();
        cTop.fill = GridBagConstraints.BOTH;
        cTop.gridwidth = GridBagConstraints.REMAINDER;
        //cTop.gridheight = 2;
        cTop.insets = ins;
//        cTop.weightx = 0.1;
        cTop.gridx = 0;
        cTop.gridy = 0;
        //cTop.ipady = 20;
        cTop.anchor = GridBagConstraints.NORTH;


        GridBagConstraints cLabel = new GridBagConstraints();
        cLabel.fill = GridBagConstraints.HORIZONTAL;
        cLabel.insets = ins;
        cLabel.weightx = 0.1;
        cLabel.gridx = 0;
        cLabel.gridy = 0;
        cLabel.anchor = GridBagConstraints.NORTH;


        GridBagConstraints cOptional = new GridBagConstraints();
        cOptional.fill = GridBagConstraints.HORIZONTAL;
        cOptional.insets = ins;
        cOptional.weightx = 0;
        cOptional.gridx = 1;
        cOptional.gridy = 0;
        cOptional.anchor = GridBagConstraints.NORTH;

        GridBagConstraints cValue = new GridBagConstraints();
        cValue.fill = GridBagConstraints.HORIZONTAL;
        cValue.insets = ins;
        cValue.weightx = 1.0;
        cValue.gridx = 2;
        cValue.gridy = 0;
        cValue.anchor = GridBagConstraints.NORTH;
        
        String descr = this.params.getDescription();

        if (descr!=null && descr.length()>0) {
            final JTextPane descrText = new JTextPane();
            descrText.setEditorKit(JTextPane.createEditorKitForContentType("text/html")); // NOI18N

            Color bg = jPanel1.getBackground();
            descrText.setBackground(new Color(bg.getRed(),bg.getGreen(),bg.getBlue(),0));

            descrText.setForeground(jPanel1.getForeground());
            descrText.setEditable(false);
            descrText.setFocusable(false);
            descrText.setText(descr);
            descrText.setBorder(BorderFactory.createTitledBorder(BorderFactory.createEtchedBorder(),"Description"));

            jPanel1.add(descrText,cTop);

            cLabel.gridy+=3;
            cOptional.gridy+=3;
            cValue.gridy+=3;

        }


        for (final FieldInstance f : params) {
            if (f.isInternal()) {
                continue;
            }

            f.addFieldListener(this, true);

            JLabel paramLabel = new JLabel(f.getName());
            ValueComponent valueComponent = f.getValueComponent(!viewOnly);

            if (valueComponent!=null) { // add the component
                paramLabel.setLabelFor(valueComponent);
                paramLabel.setToolTipText(f.getFullHtmlDescription());

                JCheckBox optCheckBox = new JCheckBox();

                if (f.isOptional()) { // provide checkbox to signify choosing
                    optCheckBox.setVisible(true);
                    optCheckBox.setToolTipText("Check to set this parameter, which is optional");
                    optCheckBox.setSelected(f.getValue()==null ? false : true);
                    valueComponent.setVisible(f.getValue()==null ? false : true);
                    optMap.put(optCheckBox, valueComponent);
                    if (this.viewOnly) {
                        optCheckBox.setEnabled(false);
                    } else {
                        optCheckBox.setEnabled(true);
                        optCheckBox.addActionListener(new ActionListener() {
                            @Override public void actionPerformed(ActionEvent e) {
                                JCheckBox optCB = (JCheckBox)e.getSource();
                                ValueComponent comp = optMap.get(optCB);
                                FieldInstance f = map.get(comp);
                                if (optCB.isSelected()) {
                                    comp.setVisible(true);
                                    // TODO: empty UI or set f to reflect the state of ui
                                } else {
                                    comp.setVisible(false);
                                    f.setValue(null);
                                }
                                //ParameterDialog.this.checkAllParameters();
                            }
                        });
                    }
                } else {
                    optCheckBox.setVisible(false);
                    optCheckBox.setEnabled(false);
                }

                valueComponent.setEnabled(!this.viewOnly);

                jPanel1.add(paramLabel,cLabel);
                jPanel1.add(optCheckBox,cOptional);
                jPanel1.add(valueComponent,cValue);
                map.put(valueComponent, f);
            }

            ++cLabel.gridy;
            ++cOptional.gridy;
            ++cValue.gridy;

        }
        
        if (this.viewOnly) {
            this.jButtonExecute.setText("OK");
            this.jButtonExecute.setToolTipText("Closes the window");
        } else {
            this.jButtonExecute.setToolTipText("This will start the execution of this job in the background");
            this.jButtonExecute.setIcon(new javax.swing.ImageIcon(getClass().getResource("resources/quickopen.png")));
        }
        
        //this.checkAllParameters();

        jPanel1.validate();

    }


    @Override
    public void dispose() {
        for (FieldInstance fi : this.params) {
            fi.removeFieldListener(this);
        }
        for (ValueComponent vc : this.map.keySet()) {
            vc.dispose();
        }
        this.invalidFields.clear();
        this.map.clear();
        this.optMap.clear();

        super.dispose();
    }

    @Override
    public void fieldValueChanged(FieldInstance field, Object newValue) {
        if (this.invalidFields.remove(field)!=null) {
            for (String msg : invalidFields.values()) {
                this.setNotification(msg);
                return;
            }

        }
        if (this.invalidFields.isEmpty()) {
            this.setNotification(null);
        }
        
    }

    @Override
    public void validationFailed(FieldInstance field, String reason) {
        reason = field.getName() + ": "+reason;
        this.invalidFields.put(field, reason);
        this.setNotification(reason);
    }

    public static void setBrowser(SimpleDocBrowser browser) {
        synchronized(ParameterDialog.docBrowserLock) {
            ParameterDialog.docBrowser = browser;
        }
    }


    private void setNotification(String msg) {
        if (msg==null) { // no problems
            notificationLabel.setText("");
            jButtonExecute.setEnabled(true);
        } else {
            notificationLabel.setText(msg);
            jButtonExecute.setEnabled(false);
        }
    }



    /** This method is called from within the constructor to
     * initialize the form.
     * WARNING: Do NOT modify this code. The content of this method is
     * always regenerated by the Form Editor.
     */
    @SuppressWarnings("unchecked")
    // <editor-fold defaultstate="collapsed" desc="Generated Code">//GEN-BEGIN:initComponents
    private void initComponents() {

        jButtonExecute = new javax.swing.JButton();
        jButtonCancel = new javax.swing.JButton();
        jScrollPane1 = new javax.swing.JScrollPane();
        jPanel1 = new javax.swing.JPanel();
        notificationLabel = new javax.swing.JLabel();
        jToolBar1 = new javax.swing.JToolBar();
        helpButton = new javax.swing.JButton();

        setDefaultCloseOperation(javax.swing.WindowConstants.DISPOSE_ON_CLOSE);
        setTitle("Set parameters for a Job"); // NOI18N
        setName("Form"); // NOI18N

        jButtonExecute.setText("Execute"); // NOI18N
        jButtonExecute.setToolTipText("Execute the job with selected parameters"); // NOI18N
        jButtonExecute.setName("jButtonExecute"); // NOI18N
        jButtonExecute.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                jButtonExecuteActionPerformed(evt);
            }
        });

        jButtonCancel.setText("Cancel"); // NOI18N
        jButtonCancel.setName("jButtonCancel"); // NOI18N
        jButtonCancel.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                jButtonCancelActionPerformed(evt);
            }
        });

        jScrollPane1.setName("jScrollPane1"); // NOI18N

        jPanel1.setAutoscrolls(true);
        jPanel1.setName("jPanel1"); // NOI18N
        jPanel1.setLayout(new java.awt.BorderLayout(1, 1));
        jScrollPane1.setViewportView(jPanel1);

        notificationLabel.setForeground(new java.awt.Color(204, 0, 0));
        notificationLabel.setToolTipText("It there is a problem with a parameter, the reason is displayed here"); // NOI18N
        notificationLabel.setName("notificationLabel"); // NOI18N

        jToolBar1.setFloatable(false);
        jToolBar1.setRollover(true);
        jToolBar1.setName("jToolBar1"); // NOI18N

        helpButton.setIcon(new javax.swing.ImageIcon(getClass().getResource("/net/aavalla/utils/jobpool/parameters/gui/resources/system-help.png"))); // NOI18N
        helpButton.setText("Help");
        helpButton.setToolTipText("Show Help");
        helpButton.setFocusable(false);
        helpButton.setName("helpButton"); // NOI18N
        helpButton.setVerticalTextPosition(javax.swing.SwingConstants.BOTTOM);
        helpButton.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                helpButtonActionPerformed(evt);
            }
        });
        jToolBar1.add(helpButton);

        javax.swing.GroupLayout layout = new javax.swing.GroupLayout(getContentPane());
        getContentPane().setLayout(layout);
        layout.setHorizontalGroup(
            layout.createParallelGroup(javax.swing.GroupLayout.Alignment.LEADING)
            .addGroup(javax.swing.GroupLayout.Alignment.TRAILING, layout.createSequentialGroup()
                .addContainerGap()
                .addComponent(notificationLabel, javax.swing.GroupLayout.DEFAULT_SIZE, 495, Short.MAX_VALUE)
                .addPreferredGap(javax.swing.LayoutStyle.ComponentPlacement.RELATED)
                .addComponent(jButtonCancel)
                .addPreferredGap(javax.swing.LayoutStyle.ComponentPlacement.RELATED)
                .addComponent(jButtonExecute))
            .addComponent(jToolBar1, javax.swing.GroupLayout.DEFAULT_SIZE, 695, Short.MAX_VALUE)
            .addComponent(jScrollPane1, javax.swing.GroupLayout.DEFAULT_SIZE, 695, Short.MAX_VALUE)
        );
        layout.setVerticalGroup(
            layout.createParallelGroup(javax.swing.GroupLayout.Alignment.LEADING)
            .addGroup(javax.swing.GroupLayout.Alignment.TRAILING, layout.createSequentialGroup()
                .addComponent(jToolBar1, javax.swing.GroupLayout.PREFERRED_SIZE, 25, javax.swing.GroupLayout.PREFERRED_SIZE)
                .addPreferredGap(javax.swing.LayoutStyle.ComponentPlacement.RELATED)
                .addComponent(jScrollPane1, javax.swing.GroupLayout.DEFAULT_SIZE, 511, Short.MAX_VALUE)
                .addPreferredGap(javax.swing.LayoutStyle.ComponentPlacement.RELATED)
                .addGroup(layout.createParallelGroup(javax.swing.GroupLayout.Alignment.BASELINE)
                    .addComponent(jButtonExecute)
                    .addComponent(jButtonCancel)
                    .addComponent(notificationLabel, javax.swing.GroupLayout.PREFERRED_SIZE, 20, javax.swing.GroupLayout.PREFERRED_SIZE)))
        );

        pack();
    }// </editor-fold>//GEN-END:initComponents

    private void jButtonExecuteActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_jButtonExecuteActionPerformed
        String relVal = params.validateFieldRelations();
        if (relVal!=null) {
            this.setNotification(relVal);
            return;
        }
        if (this.job!=null && this.pool!=null) {
            this.params.makeFinal();
            pool.addJob(job);
        }
        this.dispose();
    }//GEN-LAST:event_jButtonExecuteActionPerformed

    private void jButtonCancelActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_jButtonCancelActionPerformed
        this.dispose();
    }//GEN-LAST:event_jButtonCancelActionPerformed

    private void helpButtonActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_helpButtonActionPerformed
        SimpleDocBrowser browser;
        boolean staticBrowser;
        synchronized(ParameterDialog.docBrowserLock) {
            if (ParameterDialog.docBrowser==null) {
                if (this.job!=null) {
                    browser = new SimpleDocBrowser("Help for "+this.job.getName());
                } else {
                    browser = new SimpleDocBrowser();
                }
                browser.setDefaultCloseOperation(JFrame.DISPOSE_ON_CLOSE);
                staticBrowser = false;
            } else {
                browser = ParameterDialog.docBrowser;
                staticBrowser = true;
            }
        }
        if (staticBrowser && this.params.getHelpID()!=null) {
            browser.showPageById(this.params.getHelpID());
            browser.setVisible(true);
        } else if (this.params.getHelpURL()!=null) {
            browser.showPage(this.params.getHelpURL());
            browser.setVisible(true);
        }

    }//GEN-LAST:event_helpButtonActionPerformed


    // Variables declaration - do not modify//GEN-BEGIN:variables
    private javax.swing.JButton helpButton;
    private javax.swing.JButton jButtonCancel;
    private javax.swing.JButton jButtonExecute;
    private javax.swing.JPanel jPanel1;
    private javax.swing.JScrollPane jScrollPane1;
    private javax.swing.JToolBar jToolBar1;
    private javax.swing.JLabel notificationLabel;
    // End of variables declaration//GEN-END:variables

}
