/*
 * Copyright 2009 Lari Natri <lari.natri@gmail.com>
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Lesser General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */


package net.aavalla.utils.jobpool.parameters.validators;

import net.aavalla.utils.jobpool.parameters.*;

/**
 *
 * @author Lari Natri
 */
public class StringValidator extends ParameterValidator<String> {

    private final Integer minLength;
    private final Integer maxLength;
    private final String regExp;
    private final String uiRules;

    public StringValidator(FieldInstance field, String[] settings) {
        super(field, settings);

        if (settings[0]==null || settings[0].equals(FieldAnnotations.NULL)) {
            minLength = null;
        } else {
            minLength = Integer.parseInt(settings[0].replaceAll("[^0123456789-]", ""));
        }
        if (settings[1]==null || settings[1].equals(FieldAnnotations.NULL)) {
            maxLength = null;
        } else {
            maxLength = Integer.parseInt(settings[1].replaceAll("[^0123456789-]", ""));
        }
        if (settings[2]==null || settings[2].equals(FieldAnnotations.NULL)) {
            regExp = null;
        } else {
            regExp = settings[2];
        }

        StringBuilder uiRulesBuilder = new StringBuilder("");
        if (minLength != null) {
            uiRulesBuilder.append("minimum length: ");
            uiRulesBuilder.append(minLength);
        }
        if (maxLength != null) {
            if (uiRulesBuilder.length()>0) uiRulesBuilder.append(", ");
            uiRulesBuilder.append("maximum length: ");
            uiRulesBuilder.append(maxLength);
        }
        if (regExp != null) {
            if (uiRulesBuilder.length() > 0) uiRulesBuilder.append(", ");
            uiRulesBuilder.append("matching expression: ");
            uiRulesBuilder.append(regExp);
        }
        uiRules = uiRulesBuilder.toString();
    }

    @Override
    public String uiValidationRules() {
        return uiRules;
    }

    @Override
    public String validate(String param) {
        if (param==null && minLength != null && minLength > 0) {
            return "Cannot be empty";
        }
        if (minLength != null && param != null && param.length() < minLength) {
            return "Minimum allowed length of text is "+minLength;
        }
        if (maxLength !=null && param != null && param.length() > maxLength) {
            return "Maximum allowed length of text is "+maxLength;
        }
        if ((regExp != null && param==null) || (regExp != null && !param.matches(regExp))) {
            return "Text should match regular expression "+regExp;
        }
        return null;
    }



}
